/*! \file azucar/utils/configurationReader.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef CONFIGURATIONREADER_HPP_
#define CONFIGURATIONREADER_HPP_

#include <string>

#include <boost/filesystem/path.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMErrorHandler.hpp>
#include <xercesc/util/XMLString.hpp>

#include "azucar/utils/exceptions.hpp"

namespace configuration {
  class ConfigurationReader {
  private:
    bool started_;
    xercesc::DOMBuilder *parser_;
    xercesc::DOMDocument *doc_;
    xercesc::DOMErrorHandler *errorHandler_;
  public:
    ConfigurationReader() : started_(false), parser_(0) {};
    void init(const boost::filesystem::path &configFile);
    bool isActive() const;
    void release();
    bool existsTextOption(const std::string &name) const;
    const std::string getTextOption(const std::string &name) const;
    ~ConfigurationReader();  
  };

// private namespace to avoid name clashing in redefinition
// of these templates in different object libraries.
namespace {
  template<class T>
  const T readOption(ConfigurationReader &reader,
            const std::string &option) {
    return boost::lexical_cast<T>(reader.getTextOption(option));
  }
  
  template<>
  const boost::filesystem::path 
            readOption<boost::filesystem::path>(ConfigurationReader &reader,
                      const std::string &option) {
    return boost::filesystem::path(reader.getTextOption(option));
  }
  
  template<>
  const std::string readOption<std::string>(ConfigurationReader &reader,
                        const std::string &option) {
    return reader.getTextOption(option);
  }
  
  template<>
  const bool readOption<bool>(ConfigurationReader &reader,
                        const std::string &option) {
    std::string value = reader.getTextOption(option);
    boost::algorithm::trim(value);
    if (boost::algorithm::iequals(value, "true")) {
      return true;
    } else if (boost::algorithm::iequals(value, "false")) {
      return false;
    } else {
      throw azucar::utils::BaseException("Value is not boolean.");
    }
  }
} // end private namespace
  
  bool checkAndLogPathIsDir(const boost::filesystem::path &p,
      const std::string &errStr);
  
  bool checkAndLogPathIsRegular(const boost::filesystem::path &p,
      const std::string &errStr);

  bool checkAndLogPathExists(const boost::filesystem::path &p,
      const std::string &errStr);

  bool checkAndLogPathIsExe(const boost::filesystem::path &p,
      const std::string &errStr);

  bool checkAndLogLeaf(const std::string &str, const std::string &errStr);
      
  bool checkAndLogPrefix(const std::string &str, const std::string &errStr);

  bool checkAndLogSuffix(const std::string &str, const std::string &errStr);

  bool checkAndLogIsExt(const std::string &ext, const std::string &errStr);

  bool checkAndLogStrIsRegular(const std::string &filename,
      const std::string &errStr);
  
} // end namespace configuration

#endif /*CONFIGURATIONREADER_HPP_*/
