/*! \file azucar/utils/database/framework.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef FRAMEWORK_HPP_
#define FRAMEWORK_HPP_

#include <string>
#include <vector>
#include <map>

#include <boost/shared_ptr.hpp>

#include "azucar/utils/exceptions.hpp"

namespace azucar {
namespace utils {

namespace database {

  const std::string escapeString(const std::string &str);
  
  /*! \brief Exceptions specific to database actions.
   * 
   * \sa BaseException
   */
  class DBException : public azucar::utils::BaseException {
  public:
    /*! \sa BaseException::BaseException()
     */
    DBException(const std::string &msg);
    
    /*! \sa BaseException::~BaseException()
     */
    ~DBException() throw();       
  };

  /*! \brief A Database connection. If it is queried, it must have 
   * the proper structure. 
   * 
   * \internal
   * \invariant It is always connected to an existing database.
   * If the connection goes down, it will automatically try 
   * to reconnect the next time it accesses the database.
   * If the failure is permanent, no further operations
   * on this object will be allowed.
   */
  class Database {
  private:
    /*! \brief Copy constructor disabled.
     */
    Database(const Database&); 
    /*! \brief Assignment disabled.
     */
    Database& operator=(const Database &); 
  protected:
    /*! \brief Constructor for use by derived classes.
     */
    Database() {};
  public:
  
    /*! \brief Connect to database dbName at host:port,
     * as user with passwd. Return the connection as a Database
     * object.
     * 
     * \param dbName Name of the database to connect to.
     * \param host Network host name of database manager to use
     *  for connecting to dbName.
     * \param user User name in the database manager used to connect
     *  to dbName.
     * \param passwd Password associated to user when connecting
     *  to dbName.
     * \param port Network port number where database manager listens.
     * 
     * \exception DBException If the connection cannot be established.
     */
    static boost::shared_ptr<Database> openDatabase(const std::string &dbName, const std::string &host,
      const std::string &user, const std::string &passwd,
      int port);
    
    /*! \brief Create a new database in the database manager currently used.
     * 
     * This object keeps connected to the current database, and not
     * to the new database.
     * 
     * \pre newDb does not exist in the database namespace
     *  managed by the database manager associated to this object.
     * 
     * \param newDb Name of new database to create. 
     * 
     * \return True if newDb is created. False if it is not.
     * 
     * \exception DBException If there is any problem in the
     *  database connection or execution of commands. This implies
     *  that newDb could not be created.
     */
    virtual bool createDatabase(const std::string &newDb) = 0;
    /*! \brief Change connection from current database to the one specified.
     * 
     * \pre newDb exists and is accessible from the same
     *  database manager as the current one.
     * \param newDb Database to connect to. If connection fails,
     *  this object remains connected to the current database before
     *  calling this method.
     */
    virtual bool changeDatabase(const std::string &newDb) = 0;

    
    /*! \brief Close connection to current database.
     * 
     * This object becomes unusable after the closing.
     */
    virtual void close() = 0;
    
    /*! \brief Destroy this object. 
     *  Close connection to current database if open.
     */
    virtual ~Database() {};
  };

  const std::vector<std::map<std::string, std::string> > 
        doRowQuery(boost::shared_ptr<Database> database, const std::string &query);

  void doQuery(boost::shared_ptr<Database> database, const std::string &query);

  int doQueryAndGetId(boost::shared_ptr<Database> database, const std::string &query);
        

  
} // end namespace
}
}

#endif /*FRAMEWORK_HPP_*/
