/*! \file azucar/utils/deleterSync.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef DELETERSYNC_HPP_
#define DELETERSYNC_HPP_

#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>
#include <boost/utility.hpp>

/*! \brief Condition synchronization. The condition is the
 *  deletion of a resource.
 */
class DeleterSync : boost::noncopyable {
public:
  /*! \brief Create a condition object.
   */
  DeleterSync();
  ~DeleterSync();
  /*! \brief Wait at most secondsToWait seconds for the
   *  condition to become true.
   * 
   * \return The value of the condition. If no deletion is
   * associated to this condition synchronization object, it
   * returns true as well.
   * 
   * If the condition becomes true before secondsToWait,
   * this method returns immediately.
   */
  bool waitDeleter(unsigned int secondsToWait);
  /*! \brief Notify that the deletion has been done.
   * 
   * \post Condition is true. Only the first thread waiting
   * at waitDeleter is notified.
   */
  void deleterDone();
  /*! \brief Activate this condition synchronization object
   *  (that is, a deletion action becomes associated to this
   *  object).
   */
  void deleterRunning();
private:
  bool deleterRunning_; //!< Determines whether a deleter thread
    //!< is running.
  boost::condition deleterDoneCondition_; //!< Condition variable
    //!< to signal deleter thread is not executing.
  mutable boost::mutex m_mutex; //!< Mutex to control concurrent
    //!< access to the condition variable.

};


#endif /*DELETERSYNC_HPP_*/
