/*! \file azucar/utils/extensionObjectPattern.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef EXTENSIONOBJECTPATTERN_HPP_
#define EXTENSIONOBJECTPATTERN_HPP_

#include <cassert>
#include <map>

#include "azucar/utils/access.hpp"

typedef std::string ExtensionID;

template<class Subject>
class ExtensionI {
private:
  // ban copy and copy constructor.
  ExtensionI & operator= (const ExtensionI &other);
  ExtensionI(const ExtensionI &other); 
protected:
  typename Access<Subject>::weak_Ptr subject_;
public:
  ExtensionI(typename Access<Subject>::Ptr subject) : subject_(subject) {};
  virtual typename Access<ExtensionI<Subject> >::Ptr 
    clone() const = 0;
  virtual ~ExtensionI() {};  
}; 

template<class Subject> 
class Extensible {
private:
  typedef ExtensionI<Subject> Ext;
  typedef std::map<ExtensionID, 
                   typename Access<Ext>::Ptr> ExtMap;
  ExtMap extensions_;
public:
  typename Access<Ext>::const_Ptr
  getExtension(const ExtensionID& id) const;
  
  typename Access<Ext>::Ptr
  getExtension(const ExtensionID& id);
  
  void addExtension(const ExtensionID& id, 
    typename Access<Ext>::Ptr extension);
  
  void removeExtension(const ExtensionID& id);
  
  bool hasExtension(const ExtensionID& id); 
};

template<class S>
typename Access<ExtensionI<S> >::const_Ptr
Extensible<S>::getExtension(const ExtensionID& id) const {
  typename Extensible<S>::ExtMap::const_iterator it 
    = extensions_.find(id);
  assert(it != extensions_.end());
  return it->second;
}

template<class S>
typename Access<ExtensionI<S> >::Ptr
Extensible<S>::getExtension(const ExtensionID& id) {
  typename Extensible<S>::ExtMap::iterator it 
    = extensions_.find(id);
  assert(it != extensions_.end());
  return it->second;
}

template<class S>
void
Extensible<S>::addExtension(const ExtensionID& id, 
                typename Access<ExtensionI<S> >::Ptr extension) {
  assert(extensions_.find(id) == extensions_.end());
  extensions_[id] = extension;
}

template<class S>
void 
Extensible<S>::removeExtension(const ExtensionID& id) {
  extensions_.erase(id);
}

template<class S>
bool
Extensible<S>::hasExtension(const ExtensionID& id) {
  return (extensions_.find(id) != extensions_.end());
}


#endif /*EXTENSIONOBJECTPATTERN_HPP_*/
