/*! \file azucar/utils/geometry.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef GEOMETRY_HPP_INCLUDED_
#define GEOMETRY_HPP_INCLUDED_

#include <cmath>
#include <cassert>
#include <gmpxx.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>


template<class T> class Vector {
public:
  T x;
  T y;
  T z;

  Vector() : x(), y(), z() {};
  Vector(T xval, T yval, T zval) : x(xval), y(yval), z(zval) {};
};

typedef Vector<double> DoubleVector;
typedef Vector<int> IntVector;
typedef Vector<mpf_class> MPFVector;

/*! \brief A rectangular prism with its edges parallel to the axes.
 * 
 * It is defined by its lower left corner and its upper right corner.
 */
class Box {
public:
  DoubleVector minPoint;
  DoubleVector maxPoint;
};

class MPFBox {
public:
  MPFVector minPoint;
  MPFVector maxPoint;  
};

/*! \brief Convert from radians to degrees.
 */
double radiansToDegrees(double rad);

/*! \brief Convert from degrees to radians.
 */
double degreesToRadians(double deg);

/*! \brief v1 * scalar
 * 
 * \param v1 Vector
 * \param scalar Scalar
 * 
 * \return scalar multiplication v1 * scalar
 */
template<class T> const Vector<T> operator*(const Vector<T> &v1, 
					    T scalar) {
  Vector<T> result;
  result.x = v1.x * scalar;
  result.y = v1.y * scalar;
  result.z = v1.z * scalar;

  return result;
}

/*! \brief v1 / scalar
 * 
 * \param v1 Vector
 * \param scalar Scalar
 * 
 * \return Scalar division v1 / scalar
 */
template <class T> const Vector<T> operator/(const Vector<T> &v1, 
					    T scalar) {
  Vector<T> result;
  result.x = v1.x / scalar;
  result.y = v1.y / scalar;
  result.z = v1.z / scalar;

  return result;

}

/*! \brief v1 + v2
 * 
 * \param v1 An operand vector.
 * \param v2 The other operand vector
 * 
 * \return v1 + v2 
 */
template<class T> const Vector<T> operator+(const Vector<T> &v1, 
					    const Vector<T> &v2) {
  Vector<T> result;
  result.x = v1.x + v2.x;
  result.y = v1.y + v2.y;
  result.z = v1.z + v2.z;

  return result;
}

/*! \brief v1 - v2
 * 
 * \param v1 Left operand.
 * \param v2 Right operand.
 * 
 * \return v1 - v2
 */
template<class T> const Vector<T> operator-(const Vector<T> &v1, 
					    const Vector<T> &v2) {
  Vector<T> result;
  result.x = v1.x - v2.x;
  result.y = v1.y - v2.y;
  result.z = v1.z - v2.z;

  return result;
}

/*! \brief Dot product v1 · v2
 * 
 * \param v1 An operand vector.
 * \param v2 The other operand vector.
 * 
 * \return Dot product between vectors v1 and v2.
 */
template<class T> T operator*(const Vector<T> &v1, const Vector<T> &v2) {
  return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

/*! \brief Scalar product a * v
 * 
 * \param a scalar
 * \param v vector
 * 
 * \return scalar product a v 
 */
template<class T> Vector<T> operator*(double a, const Vector<T> &v) {
  Vector<T> result;
  result.x = a * v.x;
  result.y = a * v.y;
  result.z = a * v.z;
  return result;
}

/*! \brief Obtain |v1|^2 (square modulus)
 * 
 * \param v1 Vector whose square modulus is calculated.
 * 
 * \return |v1|^2
 */
template<class T> double squareModulusVector(const Vector<T> &v1) {
  return v1.x * v1.x + v1.y * v1.y + v1.z * v1.z;
}

/*! \brief |v1|
 * 
 * \param v1 Vector whose modulus is calculated.
 * 
 * \return |v1|
 */
template<class T> double modulusVector(const Vector<T> &v1) {
  return std::sqrt(squareModulusVector(v1));
}

/*! \brief Obtain v / |v|
 * 
 * \param v Vector to normalize.
 * 
 * \return A new normalized vector, v / |v|
 */
template<class T> Vector<T> unitVector(const Vector<T> &v) {
  return v / modulusVector(v);
}


/*! \brief Cross product v1 * v2
 * 
 * \param v1 Left operand vector of the cross product.
 * \param v2 Right operand vector of the cross product.
 * 
 * \return Cross product vector v1 * v2.
 */
template<class T> Vector<T> crossProduct(const Vector<T> &v1, 
					 const Vector<T> &v2) {
  Vector<T> result;
  result.x = v1.y * v2.z - v1.z * v2.y;
  result.y = -(v1.x * v2.z - v2.x * v1.z);
  result.z = v1.x * v2.y - v1.y * v2.x;

  return result;
}

template<class T> double getAngle(const Vector<T> &v1, const Vector<T> &v2) {
  double dotproduct = v1 * v2; // point product
  double modv1 = modulusVector(v1);
  double modv2 = modulusVector(v2);

  assert( (dotproduct != 0) and (modv1 != 0) and (modv2 != 0) );


  double angle = acos(dotproduct/(modv1 * modv2));
  
  return angle;
}

/**
 * Torsion angle in radians considering a-b-c-d
 */
template<class T> double torsionAngle(const Vector<T> &a, const Vector<T> &b,
				      const Vector<T> &c, const Vector<T> &d) {
  Vector<T> ab;
  ab.x = a.x - b.x;
  ab.y = a.y - b.y;
  ab.z = a.z - b.z;

  Vector<T> cb;
  cb.x = c.x - b.x;
  cb.y = c.y - b.y;
  cb.z = c.z - b.z;

  Vector<T> cd;
  cd.x = c.x - d.x;
  cd.y = c.y - d.y;
  cd.z = c.z - d.z;

  Vector<T> nda = crossProduct(ab, cb); // normals to the planes
  Vector<T> nea = crossProduct(cb, cd);

  double bl = modulusVector(nda); // modulus
  double el = modulusVector(nea); // modulus

  double bel = nda * nea; // point product

  assert( (bl * el) != 0);

  bel /= (bl * el);

  if (bel > 1.0)
    {
      bel = 1;
    }
  else if (bel < -1.0)
    {
      bel = -1;
    }
  
  double acosbel = acos(bel);

  double signTest = cb * crossProduct(nda, nea);

  if (signTest < 0) {
    acosbel = -acosbel;
  }

//   acosbel = (acosbel > 0.0) ? M_PI - acosbel : M_PI + acosbel;

  return acosbel;
}

// Set a Point according to Natural Extension Reference Frame ([parsons:2005]).
DoubleVector setPoint(const DoubleVector &a, const DoubleVector &b,
  const DoubleVector &c, double distance, double angle, double torsion);
  
DoubleVector setZMatrixPoint(const DoubleVector &point1, double distance,
  const DoubleVector &point2, double angle, 
  const DoubleVector &point3, double torsion);  

/*! Convert a unit quaternion to a rotation matrix.
 * 
 * \param quaternion Must be a unit quaternion.
 * \param rotation Must be a 3x3 matrix.
 */
void quaternionToRotationMatrix(const gsl_vector *quaternion,
    gsl_matrix *rotation);


/*! Calculate RMSD between points in x and in y, without transformation.
 * 
 * \param x Matrix of points (size 3xN).
 * \param y Matrix of points (size 3xN).
 * 
 * \return RMSD between points in x and in y, without transformation.
 */
double RMSD(const gsl_matrix *x, const gsl_matrix *y);
  
/*! Calculate RMSD between points in x and in y after best transformation.
 * 
 * \param x Matrix of points (size 3xN)
 * \param y Matrix of points (size 3xN)
 * \param transform Transformation matrix such that matrix
 * multiplication of transform by quaternion [x 1] is the best approximation to y.
 * This matrix is 4x4, and its first three column represent the rotation of x,
 * and the last column its translation. Last row of transform is [0 1]
 * \param rotQuaternion The quaternion representing the rotation that must be
 * applied to x for the best approximation to y.
 * 
 * \return The RMSD between x and y after transformation of x.
 */
double doRMSD(const gsl_matrix *x, const gsl_matrix *y,
  gsl_matrix *transform, gsl_vector *rotQuaternion);

/*! \brief Obtain the square distance between the points defined
 * by v1 and v2.
 *
 * \param v1 One of the vectors defining a point.
 * \param v2 The other vector defining the other point.  
 * \return squareDistance = Sum(i=x,y,z) (v1.i - v2.i)^2
 */
double squareDistance(const DoubleVector &v1,
      const DoubleVector &v2);

/*! \brief v <= w
 * 
 * v <= w if v.x < w.x or, if equal, v.y < w.y or, if equal,
 * v.z <= w.z
 * 
 * \param v A vector to compare.
 * \param w The other vector to compare.
 * 
 * \return v <= w, as explained above.
 */
bool lessThanOrEqual(const DoubleVector &v, const DoubleVector &w);

/*! \brief v < w
 * 
 * Same as in \sa lessThanOrEqual but excluding the case
 * where v and w are vectors with the same coordinates. 
 */
bool lessThan(const DoubleVector &v, const DoubleVector &w);

/*! \brief v > w
 * 
 * Same as in \sa lessThanOrEqual but excluding the case
 * where v and w are vectors with the same coordinates, and
 * using > instead of < for the comparisons.
 */
bool greaterThan(const DoubleVector &v, const DoubleVector &w);

/*! \brief v <= w considering coordinate dimension.
 * 
 * dimension must be in [1,3] (that is, x, y, or z).
 * 
 * For dimension 1, first x are compared, if equal then y, if equal
 * then z.
 * 
 * For dimension 2, first y are compared, if equal then z, if equal
 * then x.
 * 
 * For dimension 3, first z are compared, if equal then x, if equal
 * then y.
 * 
 * \param dimension Main coordinate of comparison (x, y, or z). Must
 *  be in range [1,3].
 * \param v Vector to compare.
 * \param w The other vector to compare.
 * 
 * \result v <= w considering the main coordinate dimension, as
 *  explained before.
 */
bool lessThanOrEqualDim(int dimension, 
        const DoubleVector &v, const DoubleVector &w);

/*! \brief v < w considering coordinate dimension.
 * 
 * Same as \sa lessThanOrEqualDim but excluding the case
 * where v and w are exactly equal in all their coordinates. 
 * 
 */
bool lessThanDim(int dimension,
        const DoubleVector &v, const DoubleVector &w);        

/*! \brief v > w considering coordinate dimension.
 * 
 * Same as \sa lessThanOrEqualDim but excluding the case
 * where v and w are exactly equal in all their coordinates, and
 * changing the comparison from < to >. 
 * 
 */
bool greaterThanDim(int dimension,
        const DoubleVector &v, const DoubleVector &w);        
        
/*! \brief States whether all points in region are in container.
 */
bool contains(const Box &container, const Box &region);

/*! \brief States whether all points in region are in container.
 */
bool contains(const MPFBox &container, const MPFBox &region);

/*! \brief States whether point is in container.
 */
bool contains(const Box &container, const DoubleVector &point);

/*! \brief States whether point is in container.
 */
bool contains(const MPFBox &container, const MPFVector &point);

/*! \brief States whether at least one point in region2 is 
 *  also in region1.
 */
bool intersects(const Box &region1, const Box &region2);

/*! \brief States whether at least one point in region2 is 
 *  also in region1.
 */
bool intersects(const MPFBox &region1, const MPFBox &region2);


#endif // GEOMETRY_HPP_INCLUDED_
