/*! \file azucar/utils/init.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef INIT_HPP_
#define INIT_HPP_

#include <vector>

#include <boost/function.hpp>
#include <boost/filesystem/path.hpp>

namespace azucar {
namespace utils {


/*! \brief Represents the knowledge of initialization and
 *  finalization of a module.
 * 
 * Allows initialization and finalization of the module it
 * represents, through init() or initDep() and destroy()
 * methods. Multiple invocation of these methods is allowed,
 * since only the first right invocation has effect.
 * 
 * It allows for setting dependencies among modules
 * with respect to initialization and finalization.
 */
class Initializer {
private:
  boost::function<void (Initializer&)> init_; //!< External function
  //!< to call for initialization of the module represented by 
  //!< this object.
  boost::function<void (Initializer&)> destroy_; //!< External function
  //!< to call for finalization of the module represented by 
  //!< this object.
  boost::filesystem::path configFile_;
public:
  /*! \brief State (with respect to initialization) for a module. 
   */
  enum StatusType {
      NON_INITIALIZED, //!< Module is not initialized, but may 
      //!< be changed to that state by calling init(). 
      INITIALIZED,  //!< Module is initialized and can be used.
      INIT_ERROR, //!< Module initialization was tried, and failed.
      DESTROYED //!< Module was finalized (from an INITIALIZED state),
      //!< No further operations (nor even init()) are allowed
      //!< for this module.
      };
  StatusType status; //!< State of the module represented by this
  //!< object.
  
  std::vector<Initializer*> dependents_; //!< Container of 
  //!< initializers of modules that MUST be initialized after this
  //!< one, and that must be destroyed before this one.

  /*! \brief Create an object representing a given module, and 
   *  capable of initializing and finalizing that module.
   * 
   * \pre initFunc and destroyFunc are not null.
   * 
   * \param initFunc External function to call for initiailization of
   *  the module represented by this object.
   * \param destroyFunc External function to call for finalization of 
   *  the module represented by this object.
   */
  Initializer(const boost::function<void (Initializer&)> &initFunc,
    const boost::function<void (Initializer&)> &destroyFunc,
    const boost::filesystem::path &configFile = "");

  /*! \brief Initialize the module represented by this object.
   * 
   * \sa initDep(), in case a dependent module wants to 
   * initialize this module first.
   * 
   * If this module is in NON_INITIALIZED status, try to 
   * INITIALIZE it (and those modules it depends on, if needed).
   * New status is INITIALIZED.
   * Otherwise, do nothing.
   * 
   * \exception BaseException Initialization of this module
   *  (or modules it depends on) failed. New status is INIT_ERROR.
   */
  void init();
  /*! \brief Initialize the module represented by this object,
   * and add dependent as a module depending on this one.
   * 
   * \sa init() for behaviour of the method. 
   */
  void initDep(Initializer& dependent);
  
  /*! \brief Finalize the module represented by this object.
   * 
   * If this module is in INITIALIZED status, finalize it 
   * (and those modules dependent on this, first, if needed).
   * New status is DESTROYED.
   * 
   * Otherwise, do nothing.
   */
  void destroy();
  
  const boost::filesystem::path getConfigFile() const;
  void setConfigFile(const boost::filesystem::path &configFile);
};

extern Initializer globalInitializer; //!< Initializer for all
//!< modules of a program. It should be defined by the main program,
//!< and its init and destroy methods must call to the required
//!< initializers for the different modules.
//!< Main program should call the init() method at the beginning
//!< of it, and the destroy() method at the end.
//!< Behaviour is undefined if these methods are not called
//!< inside the execution of main().

}
}

#endif /*INIT_HPP_*/
