/*! \file azucar/utils/iteratorPIMPL.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef ITERATORPIMPL_HPP_INCLUDED_
#define ITERATORPIMPL_HPP_INCLUDED_

#include "azucar/utils/iteratorTempl.hpp"

#include <vector>

#include <boost/shared_ptr.hpp>


/*
 * This is one kind of iterator implementation. It depends on the container
 * to provide a vector of elements to iterate, as well as a vector of
 * different vectors for elementComposites. The iterator will go through the
 * vector of elements to iterate first, and then, it will go through the
 * different elementComposites in the vector of vectors.
 */

namespace IteratorPIMPLDefs {
  enum Status {ITERATING, DONE};
}

// Invariant: 
// ITERATING status: either v1Iter or containerIter point to a valid T item
// (if containerIter does, then v2Iter and v2AuxIter are also valid).
// DONE status: If none of v1Iter or containerIter point to a valid T item,
// then the GeneralIteratorPIMPL is set to point to the end.

template<class Traits>
class GeneralIteratorPIMPL : public Traits::Iterator::PIMPL {
  typedef typename Traits::Value T;
  typedef typename Traits::Tag TTag;
  typedef typename Traits::Iterator TIter;
  typedef typename Traits::Iterator::PIMPL TPIMPL;
  typedef typename Traits::Container TContainer;
public:
  /**
   * Create a new GeneralIteratorPIMPL. Should provide:
   * stat - the status intended for this iterator. It can be ITERATING
   * or DONE. In the first case, the constructor checks that there exist 
   * actually elements to iterate through. Otherwise, it changes status to
   * DONE.
   * vector1 - It is the first place to iterate through, and it contains
   * pointers to the elements. It can be void.
   * vector2 - It is the next place to iterate through. It goes, in order,
   * through the containers and iterates through them (they can be void).
   * Once vector1 and vector2 are exhausted, the iterator will become DONE.
   * Notice that iterators can only be created in two states: begin() and
   * end() (which corresponds to DONE). Other states can only be obtained
   * by cloning() or calls to advance().
   * Check Invariant definition above.
   */
  GeneralIteratorPIMPL(IteratorPIMPLDefs::Status stat,
      std::vector<boost::shared_ptr<T> > &vector1,
      std::vector< std::vector<boost::shared_ptr<TContainer> >* > &vector2)
    : status(stat), v1(vector1), v2(vector2) {

    // if not DONE, check that there exists really some element to return.
    // otherwise, change status to DONE.
    // If that element exists, initialize the internal pointers to that
    // element (v1Iter, v2Iter and v2AuxIter).
    if (status != IteratorPIMPLDefs::DONE) {
      v1Iter = v1.begin();
      if (v1Iter == v1.end()) {
	// find first element pointed by containerIter
	v2Iter = v2.begin();
	bool iterFound = false;
	while(not iterFound and (v2Iter != v2.end())) {
	  v2AuxIter = (*v2Iter)->begin();
	  while(not iterFound and (v2AuxIter != (*v2Iter)->end() )) {
	    containerIter = (*v2AuxIter)->begin(TTag());
	    if (containerIter != (*v2AuxIter)->end(TTag())) {
	      iterFound = true;
	    } else {
	      ++v2AuxIter;
	    }
	  }

	  if (not iterFound) {
	    ++v2Iter;
	  }
	}
	if (not iterFound) {
	  status = IteratorPIMPLDefs::DONE;
	}
      }
    }
  };

  T &current() {
    assert(status != IteratorPIMPLDefs::DONE);

    // Decide where to extract the data from (see Invariant).
    if (v1Iter == v1.end()) {
      return *containerIter;
    } else {
      return **v1Iter; // the T pointed by the shared_ptr v1Iter points to
    }
    
  };

  /**
   * Only allowed for iterators not in the DONE state.
   * If advance gets to the end, it automatically sets the DONE state.
   */
  void advance() {
    assert(status != IteratorPIMPLDefs::DONE);

    // Check whether we should advance in the v1 vector or in the v2
    if (v1Iter == v1.end()) {
      // we advance in the v2 vector. This implies that 
      // containerIter points to the last valid element. 
      // We should advance containerIter, if at end, we should advance
      // v2AuxIter, if at end, we should advance v2Iter,
      // if at end, we should set status to DONE
      ++containerIter;

      while(containerIter == (*v2AuxIter)->end(TTag())) {
	++v2AuxIter;
	while(v2AuxIter == (*v2Iter)->end()) {
	  ++v2Iter;
	  if (v2Iter == v2.end()) {
	    status = IteratorPIMPLDefs::DONE;
	    return;
	  }
	  v2AuxIter = (*v2Iter)->begin();
	}
	containerIter = (*v2AuxIter)->begin(TTag());
      }
    } else {
      ++v1Iter;
      // If we happened to have advanced to the end of v1, we must
      // set containerIter to point to a valid element, otherwise,
      // we must change state to DONE.
      if (v1Iter == v1.end()) {
	// find first element pointed by containerIter
	v2Iter = v2.begin();
	bool iterFound = false;
	while(not iterFound and (v2Iter != v2.end())) {
	  v2AuxIter = (*v2Iter)->begin();
	  while(not iterFound and (v2AuxIter != (*v2Iter)->end() )) {
	    containerIter = (*v2AuxIter)->begin(TTag());
	    if (containerIter != (*v2AuxIter)->end(TTag())) {
	      iterFound = true;
	    } else {
	      ++v2AuxIter;
	    }
	  }

	  if (not iterFound) {
	    ++v2Iter;
	  }
	}
	if (not iterFound) {
	  status = IteratorPIMPLDefs::DONE;
	}
      }
    }
  };

  TPIMPL *clone() {
    GeneralIteratorPIMPL *result = new GeneralIteratorPIMPL(*this);
    return result;
  };

  /**
   * Two iterators are equal if they:
   * - are of same type, and are both DONE
   * - are of same type, are both ITERATING, and their active iterators
   * (that is, v1Iter or (v2Iter, v2AuxIter, containerIter) are 
   * all the same.
   *
   * Perhaps this should be less implementation dependent and more general
   * in meaning.
   */
  bool operator==(const TPIMPL &a) const {
    if (typeid(*this) == typeid(a)) {
      const GeneralIteratorPIMPL &ac =
	dynamic_cast<const GeneralIteratorPIMPL &>(a);

      // ?Should we check that the vectors are identical first?

      if ( (status == IteratorPIMPLDefs::DONE) 
	   and (ac.status == IteratorPIMPLDefs::DONE)) {
	return true;
      }

      // check that iterators coincide, but no void iterator is compared
      // to another void iterator
      if ( (status == IteratorPIMPLDefs::ITERATING) 
	   and (ac.status == IteratorPIMPLDefs::ITERATING) ) {
	if (v1Iter == ac.v1Iter) {
	  if (v1Iter != v1.end()) {
	    return true;
	  }

	  if (v2Iter == ac.v2Iter) {
	    // We could assume the following check
	    if (v2Iter != v2.end()) {
	      if (v2AuxIter == ac.v2AuxIter) {
		// We could assume the following check
		if (v2AuxIter != (*v2Iter)->end() ) {

		  // We could check containerIter is not end, 
		  // but it is not really needed.
		  return (containerIter == ac.containerIter);
		} else {
		  assert("Iterating but nothing pointed" and 0);
		}
	      } else {
		return false;
	      }
	    } else {
	      assert("Iterating but nothing pointed" and 0);
	    }
	  } else {
	    return false;
	  }
	} else {
	  return false;
	}
      } else {
	return false;
      }
    } else {
      return false;
    }
//     return false;  // Not needed
  };

  bool operator!=(const TPIMPL &a) const {
    return ( not (*this == a));
  };

protected:
  /**
   * Copy is only allowed through cloning
   */
  GeneralIteratorPIMPL(const GeneralIteratorPIMPL &a) 
    : status(a.status), v1(a.v1), v2(a.v2), 
      v1Iter(a.v1Iter), v2Iter(a.v2Iter), v2AuxIter(a.v2AuxIter),
      containerIter(a.containerIter) {};

private:
  // copy not allowed (it cannot be anyway, since v1 is a reference).
  GeneralIteratorPIMPL &operator=(const GeneralIteratorPIMPL &);

  IteratorPIMPLDefs::Status status;
  std::vector<boost::shared_ptr<T> > &v1;
  // v2 is not a reference, since the vector passed is created explicitely
  // for this PIMPL, and, therefore, it is not stored somewhere else
  std::vector<std::vector<boost::shared_ptr<TContainer> >* > v2;

  // v1Iter is the iterator pointing to v1.
  typename std::vector<boost::shared_ptr<T> >::iterator v1Iter;

  // v2Iter is the interator pointing to v2
  typename std::vector< std::vector<boost::shared_ptr<TContainer> >* >::iterator v2Iter;

  // for each container in v2, v2AuxIter points to the corresponding container
  // in the current vector of containers pointed by v2Iter.
  typename std::vector<boost::shared_ptr<TContainer> >::iterator v2AuxIter;

  // containerIter points to the current element in the current container
  // (pointed by v2AuxIter).
  TIter containerIter;
};

#endif // ITERATORPIMPL_HPP_INCLUDED_
