/*! \file azucar/utils/iteratorTempl.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef ITERATORTEMPL_HPP_INCLUDED_
#define ITERATORTEMPL_HPP_INCLUDED_

/*
 * This is the base iterator framework. It consists of a common
 * interface for all iterators of a given type, but allows for 
 * different implementations (for example, residue and molecule will
 * implement differently the atomIterator). Implementations are 
 * accessed through the IteratorTemplPIMPL interface.
 * Concrete PIMPLs are created by instantiating IteratorTemplPIMPL.
 * Since a basic way to create iterators is by going through 
 * a vector of elements which can be recursively iterated
 * (for example, a molecule possesses a vector of residues; an
 * atomIterator for a molecule will recursively go through the
 * atomIterators of the residues). For such cases, there exists
 * a common implementation: GeneralIteratorPIMPL (see
 * iteratorPIMPL.hpp). This is ready to be instantiated for the exact
 * type of element to iterate (that is, the corresponding traits).
 * For traits, see iteratorTraits.hpp
 */

// All functions must be implemented or be pure virtual if the typeinfo 
// of RTTI is to be generated. Otherwise, linking errors will happen due to
// missing implementations of functions.

// This is a base class, and it has so protected constructors.
// This PIMPL allows for different implementations of iterators (one for
// residue, another for Protein, etc.) and sharing the same facade,
// therefore allowing for generic algorithms using AtomIterator

template<typename T> class IteratorTemplPIMPL {
public:
  virtual T& current() = 0;
  virtual void advance() = 0;
  virtual IteratorTemplPIMPL *clone() = 0;
  virtual bool operator==(const IteratorTemplPIMPL &a) const = 0;
  virtual bool operator!=(const IteratorTemplPIMPL &a) const = 0;
  virtual ~IteratorTemplPIMPL() = 0;
protected:
  IteratorTemplPIMPL() {};
  IteratorTemplPIMPL(const IteratorTemplPIMPL &a) {};
};

// We must provide a definition so that the compiler doesn't complain.
// even though it is a pure virtual mehtod. (see [Meyers, p.43])
template<typename T>
IteratorTemplPIMPL<T>::~IteratorTemplPIMPL() {};


template<typename T> class IteratorTempl {
public:
  typedef IteratorTemplPIMPL<T> PIMPL;
  IteratorTempl() : pimpl(0) {}; // Allows uninitialized iterators
  // All comparisons should give an error
  // Copy should be allowed

  IteratorTempl(IteratorTemplPIMPL<T> *state);
  IteratorTempl(const IteratorTempl &a);
  IteratorTempl& operator=(const IteratorTempl &a);
  T& operator*();
  IteratorTempl& operator ++();
  IteratorTempl operator ++(int);
  T* operator->();
  bool operator==(const IteratorTempl &a) const;
  bool operator!=(const IteratorTempl &a) const;
private:
  IteratorTemplPIMPL<T> *pimpl;
};


template<typename T> 
IteratorTempl<T>::IteratorTempl<T>(IteratorTemplPIMPL<T> *state) {
  pimpl = state->clone();
}

template<typename T>
IteratorTempl<T>::IteratorTempl<T>(const IteratorTempl<T> &a) {
  // allow copy of void iterators
  if (a.pimpl == 0) {
    pimpl = 0;
  } else {
    pimpl = a.pimpl->clone();
  }
}

template<typename T>
IteratorTempl<T>& IteratorTempl<T>::operator=(const IteratorTempl &a) {
  // allow copy of void iterators
  if (a.pimpl == 0) {
    pimpl = 0;
  } else {
    pimpl = a.pimpl->clone();
  }
  return *this;
}

template<typename T>
T& IteratorTempl<T>::operator*() {
  return pimpl->current();
}

template<typename T>
IteratorTempl<T>& IteratorTempl<T>::operator ++() {
  pimpl->advance();
  return *this;
}

template<typename T>
IteratorTempl<T> IteratorTempl<T>::operator ++(int) {
  IteratorTempl tmp = *this;
  ++(*this);
  return tmp;
}

template<typename T>
T* IteratorTempl<T>::operator->() {
  return &(pimpl->current());
}

template<typename T>
bool IteratorTempl<T>::operator==(const IteratorTempl<T> &a) const {
  assert( (this->pimpl != 0) and (a.pimpl != 0));
  return ( (*(this->pimpl)) == (*(a.pimpl)) );
}

template<typename T>
bool IteratorTempl<T>::operator!=(const IteratorTempl<T> &a) const {
  return (not (*this == a));
}

#endif // ITERATORTEMPL_HPP_INCLUDED_
