/*! \file azucar/utils/kdtree.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef KDTREE_HPP_
#define KDTREE_HPP_
#include <vector>
#include <list>
#include <cassert>
#include <cmath>

#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/numeric/conversion/cast.hpp>

#include "azucar/utils/geometry.hpp"


/*! 
 * This KDTree organizes objects of type T
 * 
 * This T objects must have the following method:
 * getCoordinates(), which returns a DoubleVector with
 * x, y, and z coordinates.
 * 
 * 1.- For each node, if it has an even number of leaves under it,
 *  its left and right subtrees both have the same number of leaves.
 *  For odd number of leaves, the right subtree has one leaf more
 *  than the left subtree.
 * 2.- For any node, the maximum point/leaf in the left subtree is <=(by that
 *  node dimension) than the split value of that node; and the
 *  minimum point in the right subtree is >(by that node dimension)
 *  than the split value of that node.
 * 
 * 
 */
template <class T> class KDTreeNode 
  : public boost::enable_shared_from_this<KDTreeNode<T> > {
public:
  typedef boost::shared_ptr<T> ObjPtr;
  typedef boost::shared_ptr<KDTreeNode> NodePtr;
  typedef boost::weak_ptr<KDTreeNode> NodeWeakPtr;
  typedef std::vector<ObjPtr> ObjContainer;

  enum NodeType {INTERNAL, LEAF};

  DoubleVector splitValue; // value associated to this node (if LEAF)
  // or value such that all leaves <= splitValue are in the left
  // subtree, and all those > splitValue are in the right.
  ObjPtr object; // object associated to this node (if LEAF)
  NodePtr leftChild; // left subtree (if INTERNAL)
  NodePtr rightChild; // right subtree (if INTERNAL)
  NodeWeakPtr parent;  // parent node (if any)
  NodeType typeOfNode; // INTERNAL or LEAF
  int dimensionDepth; // either 1, 2 or 3
  Box regionValue; // region including all points in this node's leaves.
  
  /*! \brief Find all points (leaves) in this subtree
   *  that are contained in searchRange.
   * 
   * \pre this is a KD-tree subtree.
   * 
   * \param searchRange Rectangular limits of point search.
   * \param[out] result Container with all objects associated
   *  to the leaves (points) of this subtree contained in searchRange.
   */
  void search(const Box &searchRange,
        ObjContainer &result);

  class NodeBuildInfo {
  public:    
    NodePtr node;
    ObjContainer sortDimX; // sorted list, by dimension 1 (x axis)
    ObjContainer sortDimY; // sorted list, by dimension 2 (y axis)
    ObjContainer sortDimZ; // sorted list, by dimension 3 (z axis)
    typename ObjContainer::size_type numElems;
  };

  static NodePtr buildTree(ObjContainer &sortDimX, 
      ObjContainer &sortDimY, ObjContainer &sortDimZ);
            
private:
  /*! \brief Obtain a container of tree leaves' objects sorted in BFS visit
   * order from the current node.
   * 
   * \param[in, out] result Container of the leaves' objects already visited
   *  (input parameter), and with this tree's leaves' objects added in return,
   *  in a BFS visit order.
   */
  void reportTree(ObjContainer &result);
  /*! \brief Find all points (leaves) in this subtree
   *  that are contained in searchRange.
   * 
   * \pre this is a KD-tree subtree. 
   * \pre this->findSplitNode(searchRange) == this 
   * 
   * \param searchRange Rectangular limits of point search.
   * \param[out] result Container with all objects associated
   *  to the leaves (points) of this subtree contained in searchRange.
   */
  void searchKDTree(const Box &searchRange,
        ObjContainer &result);
  /*! \brief Obtain the node of this subtree where 
   * leafs in searchRange can be on its left and right subtrees.
   * 
   * \pre this is a KD-tree subtree
   * 
   * \param searchRange Rectangular limits of point search.
   * 
   * \return Node of this subtree with leaves in searchRange 
   * in both of its leaves, and containing all of the leaves in
   * searchRange of this subtree. If no leaf of this subtree is in
   * searchRange, it returns a leaf of this tree.
   */
  NodePtr findSplitNode(const Box &searchRange);        
};

/*! \brief Break a set of points into two parts, and keep
 *  order by dimension.
 * 
 * \pre sortedMainDim is sorted by the Main dimension.
 *  sortedDimA is sorted by the secondary dimension A.
 *  sortedDimB is sorted by the secondary dimension B.
 *  All sorted containers have the same elements.
 * 
 * \post Given the index of the partition point in the main dimension
 * (medianIndex), sortedDimA is broken into lower and upper parts,
 * both sorted by the secondary dimension A, with the lower part 
 * containing those points <= the partition point (in the main dimension).
 * Similarly, for sortedDimB.
 * 
 * \param[in] medianIndex index on sortedMainDim marking the breaking point.
 *  It defines the median point.
 * \param[in] sortedMainDim container sorted by the main dimension.
 * \param[out] oLowerMainDim sorted subset of sortedMainDim,
 *  containing all points <=(in main dimension) median point.
 * \param[out] oHigherMainDim sorted subset of sortedMainDim
 *  containing all points >(in main dimension) median point.
 * \param[in] sortedDimA container sorted by a secondary dimension.
 * \param[out] oLowerDimA sorted subset of sortedDimA
 *  containing same points as in oLowerMainDim, but sorted by
 *  dimension A.
 * \param[out] oHigherDimA sorted subset of sortedDimA
 *  containing same points as in oHigherMainDim, but sorted by 
 *  dimension A.
 * \param[in] sortedDimB container sorted by another secondary dimension.
 * \param[out] oLowerDimB sorted subset of sortedDimB
 *  containing same points as in oLowerMainDim, but sorted by
 *  dimension B.
 * \param[out] oHigherDimB sorted subset of sortedDimB
 *  containing same points as in oHigherMainDim, but sorted by 
 *  dimension B.
 *    
 */
template <class T>
void
breakSorted(int dimension, 
    typename KDTreeNode<T>::ObjContainer::size_type medianIndex,
    typename KDTreeNode<T>::ObjContainer& sortedMainDim,
    typename KDTreeNode<T>::ObjContainer& oLowerMainDim,
    typename KDTreeNode<T>::ObjContainer& oHigherMainDim,
    typename KDTreeNode<T>::ObjContainer& sortedDimA,
    typename KDTreeNode<T>::ObjContainer& oLowerDimA,
    typename KDTreeNode<T>::ObjContainer& oHigherDimA,
    typename KDTreeNode<T>::ObjContainer& sortedDimB,
    typename KDTreeNode<T>::ObjContainer& oLowerDimB,
    typename KDTreeNode<T>::ObjContainer& oHigherDimB) {


  
  DoubleVector median = sortedMainDim[medianIndex]->getCoordinates();

  oLowerMainDim.insert(oLowerMainDim.end(), sortedMainDim.begin(),
    sortedMainDim.begin() + medianIndex + 1);

  oHigherMainDim.insert(oHigherMainDim.end(),
    sortedMainDim.begin() + medianIndex + 1, sortedMainDim.end());
    
    // break dimA
    typename KDTreeNode<T>::ObjContainer::iterator it;
    for(it = sortedDimA.begin(); it != sortedDimA.end(); ++it) {
      if (lessThanOrEqualDim(dimension, (*it)->getCoordinates(), median)) {
        oLowerDimA.push_back(*it);
      } else {
        oHigherDimA.push_back(*it);
      }      
    }
    
    for(it = sortedDimB.begin(); it != sortedDimB.end(); ++it) {
      if (lessThanOrEqualDim(dimension, (*it)->getCoordinates(), median)) {
        oLowerDimB.push_back(*it);
      } else {
        oHigherDimB.push_back(*it);
      }      
    }



    assert(oLowerDimB.size() + oHigherDimB.size()
      == sortedDimB.size());
    
    assert(oLowerDimA.size() == oLowerDimB.size());
    assert(oLowerDimA.size() == oLowerMainDim.size());
    assert(oHigherDimA.size() == oHigherDimB.size());
    assert(oHigherDimA.size() == oHigherMainDim.size());
    
}


template <class T> 
void KDTreeNode<T>::reportTree(ObjContainer &result) {
  if (typeOfNode == LEAF) {
    result.push_back(object);
  } else {
    if (leftChild) {
      leftChild->reportTree(result);
    }
    if (rightChild) {
      rightChild->reportTree(result);
    }
  }
}        


template <class T> 
typename KDTreeNode<T>::NodePtr 
KDTreeNode<T>::findSplitNode(const Box &searchRange) {
  if (typeOfNode == LEAF) {
    return this->shared_from_this();
  } else {
    // it must have two children (not a leaf 
    // => at least left child)

    if (lessThanOrEqualDim(dimensionDepth, 
              searchRange.maxPoint, splitValue)) {
      return leftChild->findSplitNode(searchRange);
    } else if (greaterThanDim(dimensionDepth, 
                  searchRange.minPoint, splitValue)) {
      return rightChild->findSplitNode(searchRange); 
    } else {
      return this->shared_from_this();
    }
    
  }
}

template <class T> 
void KDTreeNode<T>::search(const Box &searchRange, 
    ObjContainer &result) {
  NodePtr splitNode = findSplitNode(searchRange);
  splitNode->searchKDTree(searchRange, result);      
}

template <class T> 
void KDTreeNode<T>::searchKDTree(const Box &searchRange, 
    ObjContainer &result) {
  if (typeOfNode == LEAF) {
    if (contains(searchRange, splitValue)) {
      result.push_back(object);
    }
  } else {
    if (contains(searchRange, leftChild->regionValue)) {
      leftChild->reportTree(result);
    } else if (intersects(leftChild->regionValue, searchRange)) {
      leftChild->searchKDTree(searchRange, result); 
    }
  
    if (rightChild) {
      if (contains(searchRange, rightChild->regionValue)) {
        rightChild->reportTree(result);
      } else if (intersects(rightChild->regionValue, searchRange)) {
        rightChild->searchKDTree(searchRange, result);
      }
    }
  }
}

/*! \brief Build a KD-tree representing the points associated
 * to the objects in sortDimX, sortDimY, and sortDimZ.
 * 
 * \pre The objects contained in sortDimX, sortDimY and sortDimZ
 *  are the same.
 * \pre sortDimX is sorted by x coordinate.
 * \pre sortDimY is sorted by y coordinate.
 * \pre sortDimZ is sorted by z coordinate.
 * 
 * \param sortDimX container of objects sorted by x coordinate.
 * \param sortDimY the same container of objects, but sorted by
 *  y coordinate.
 * \param sortDimZ the same container of objects, but sorted by
 *  z coordinate.
 * 
 * \return The root node of a KD-tree representing the points in
 * sortDimX, sortDimY and sortDimZ.
 */
template <class T> 
typename KDTreeNode<T>::NodePtr 
KDTreeNode<T>::buildTree(ObjContainer &sortDimX, 
      ObjContainer &sortDimY, ObjContainer &sortDimZ) {

  typedef boost::shared_ptr<NodeBuildInfo> BuildInfoPtr;
  
  std::list<BuildInfoPtr> buildQueue;
  
  if (sortDimX.empty()) {
    return NodePtr();
  }


  // create first node
  NodePtr rootNode(new KDTreeNode());
  BuildInfoPtr buildInfo(new NodeBuildInfo());
  buildInfo->numElems = sortDimX.size(); 

  rootNode->dimensionDepth = 1;
  rootNode->regionValue.minPoint 
    = DoubleVector(sortDimX[0]->getCoordinates().x, 
        sortDimY[0]->getCoordinates().y, sortDimZ[0]->getCoordinates().z);
  rootNode->regionValue.maxPoint
    = DoubleVector(sortDimX[buildInfo->numElems-1]->getCoordinates().x,
        sortDimY[buildInfo->numElems-1]->getCoordinates().y,
        sortDimZ[buildInfo->numElems-1]->getCoordinates().z);
  
  buildInfo->node = rootNode;
  buildInfo->sortDimX = sortDimX;
  buildInfo->sortDimY = sortDimY;
  buildInfo->sortDimZ = sortDimZ;
  
  // insert first node into queue
  buildQueue.push_back(buildInfo);
  
  
  // invariant: each node in the buildQueue has defined:
  // 1 KDTreeNode object (parent, regionValue and dimensionDepth)
  // 2 sortDimX, sortDimY, sortDimZ (for that KDTreeNode object)
  // 3 numElems (= size(sortDimX) = size(sortDimY) = size(sortDimZ)

  // while queue not empty
  // 1 get next node
  // 2 break node into children nodes
  // 3 insert children nodes into queue
  while (not buildQueue.empty()) {
    buildInfo = buildQueue.front();
    buildQueue.pop_front();

    if (buildInfo->numElems == 1) {
      buildInfo->node->typeOfNode = LEAF;
      buildInfo->node->object = buildInfo->sortDimX[0];
      buildInfo->node->splitValue = buildInfo->node->object->getCoordinates();
    } else {
      buildInfo->node->typeOfNode = INTERNAL;
      typename ObjContainer::size_type medianIndex 
        = boost::numeric_cast<typename ObjContainer::size_type>(
            floor(buildInfo->numElems / 2.0) - 1);
      // create right child
      int newDimension = (buildInfo->node->dimensionDepth % 3) + 1;
      
      BuildInfoPtr left(new NodeBuildInfo());
      left->node.reset(new KDTreeNode());
      left->node->parent = buildInfo->node;
      buildInfo->node->leftChild = left->node;  
      left->node->dimensionDepth = newDimension;  
      left->numElems = medianIndex + 1;

      BuildInfoPtr right(new NodeBuildInfo());
      right->node.reset(new KDTreeNode());
      right->node->parent = buildInfo->node;
      buildInfo->node->rightChild = right->node;
      right->node->dimensionDepth = newDimension;
      right->numElems = (buildInfo->numElems - left->numElems);
      
      switch(buildInfo->node->dimensionDepth) {
        case 1: {
          buildInfo->node->splitValue 
            = buildInfo->sortDimX[medianIndex]->getCoordinates();
          breakSorted<T>(buildInfo->node->dimensionDepth, 
            medianIndex,
            buildInfo->sortDimX, left->sortDimX, right->sortDimX,
            buildInfo->sortDimY, left->sortDimY, right->sortDimY,
            buildInfo->sortDimZ, left->sortDimZ, right->sortDimZ);
          break;            
        }
        case 2: {
          buildInfo->node->splitValue 
            = buildInfo->sortDimY[medianIndex]->getCoordinates();
          breakSorted<T>(buildInfo->node->dimensionDepth,
            medianIndex,
            buildInfo->sortDimY, left->sortDimY, right->sortDimY,
            buildInfo->sortDimZ, left->sortDimZ, right->sortDimZ,
            buildInfo->sortDimX, left->sortDimX, right->sortDimX);
          break;            
        }
        case 3: {
          buildInfo->node->splitValue 
            = buildInfo->sortDimZ[medianIndex]->getCoordinates();
          breakSorted<T>(buildInfo->node->dimensionDepth,
            medianIndex,
            buildInfo->sortDimZ, left->sortDimZ, right->sortDimZ,
            buildInfo->sortDimX, left->sortDimX, right->sortDimX,
            buildInfo->sortDimY, left->sortDimY, right->sortDimY);
          break;            
        }
        default: assert(false); break;
      }

      left->node->regionValue.minPoint 
        = DoubleVector(left->sortDimX[0]->getCoordinates().x, 
          left->sortDimY[0]->getCoordinates().y, 
          left->sortDimZ[0]->getCoordinates().z);
      left->node->regionValue.maxPoint
        = DoubleVector(left->sortDimX[left->numElems-1]->getCoordinates().x,
          left->sortDimY[left->numElems-1]->getCoordinates().y,
          left->sortDimZ[left->numElems-1]->getCoordinates().z);

      right->node->regionValue.minPoint 
        = DoubleVector(right->sortDimX[0]->getCoordinates().x, 
          right->sortDimY[0]->getCoordinates().y, 
          right->sortDimZ[0]->getCoordinates().z);
      right->node->regionValue.maxPoint
        = DoubleVector(right->sortDimX[right->numElems-1]->getCoordinates().x,
          right->sortDimY[right->numElems-1]->getCoordinates().y,
          right->sortDimZ[right->numElems-1]->getCoordinates().z);
        
      buildQueue.push_back(left);
      buildQueue.push_back(right);

    }
        
  }
  
  
  // return first node
  return rootNode;

}

/*! \brief Check that the kd-tree root is a kd-tree.
 * 
 * Conditions checked are:
 * 1.- For each node, if it has an even number of leaves under it,
 *  its left and right subtrees both have the same number of leaves.
 *  For odd number of leaves, the right subtree has one leaf more
 *  than the left subtree.
 * 2.- For any node, the maximum point/leaf in the left subtree is <=(by that
 *  node dimension) than the split value of that node; and the
 *  minimum point in the right subtree is >(by that node dimension)
 *  than the split value of that node.
 * 
 * A single leaf is always a correct KD-tree.
 * 
 * \param[out] subTreeRegion Region encompassing all
 *  points/leaves under root.
 * \param[out] subTreeSize Number of points/leaves under root.
 * 
 * \return True if root is a KD-tree, false otherwise.
 */
template <class T>
bool checkTree(typename KDTreeNode<T>::NodePtr root, 
  Box &subTreeRegion, int &subTreeSize) {

  if (root->typeOfNode == KDTreeNode<T>::LEAF) {
    subTreeRegion = root->regionValue;
    subTreeSize = 1;
    return true;
  } else {
    Box leftSubTreeBox;
    Box rightSubTreeBox;
    int leftSubTreeSize;
    int rightSubTreeSize;
    
    checkTree<T>(root->leftChild, leftSubTreeBox, leftSubTreeSize);
    checkTree<T>(root->rightChild, rightSubTreeBox, rightSubTreeSize);

    bool result 
      = ( (leftSubTreeSize % 2) == 0) 
          ? (leftSubTreeSize == rightSubTreeSize)
          : (leftSubTreeSize == (rightSubTreeSize - 1));

    result = result
      and lessThanOrEqualDim(root->dimensionDepth, leftSubTreeBox.maxPoint, root->splitValue)
      and greaterThanDim(root->dimensionDepth, rightSubTreeBox.minPoint, root->splitValue);

    subTreeRegion.minPoint = leftSubTreeBox.minPoint;
    subTreeRegion.maxPoint = rightSubTreeBox.maxPoint;
    subTreeSize = leftSubTreeSize + rightSubTreeSize;

    return result;           
    
  }


}

/*! \brief Check that the kd-tree root is a kd-tree.
 * 
 * Conditions checked are:
 * 1.- For each node, if it has an even number of leaves under it,
 *  its left and right subtrees both have the same number of leaves.
 *  For odd number of leaves, the right subtree has one leaf more
 *  than the left subtree.
 * 2.- For any node, the maximum point/leaf in the left subtree is <=(by that
 *  node dimension) than the split value of that node; and the
 *  minimum point in the right subtree is >(by that node dimension)
 *  than the split value of that node.
 * 
 * A single leaf is always a correct KD-tree.
 * 
 * \return True if root is a KD-tree, false otherwise.
 */
template <class T>
bool checkTree(typename KDTreeNode<T>::NodePtr root) {
  
  if (root->typeOfNode == KDTreeNode<T>::LEAF) {
    return true;
  } else {
    Box leftSubTreeBox;
    Box rightSubTreeBox;
    int leftSubTreeSize;
    int rightSubTreeSize;
    
    checkTree<T>(root->leftChild, leftSubTreeBox, leftSubTreeSize);
    checkTree<T>(root->rightChild, rightSubTreeBox, rightSubTreeSize);

    bool result 
      = ( ( (leftSubTreeSize + rightSubTreeSize )% 2) == 0) 
          ? (leftSubTreeSize == rightSubTreeSize)
          : (leftSubTreeSize == (rightSubTreeSize - 1));

    result = result
      and lessThanOrEqual(leftSubTreeBox.maxPoint, root->splitValue)
      and greaterThan(rightSubTreeBox.minPoint, root->splitValue);

    return result;           
  }
} 

#endif /*KDTREE_HPP_*/
