/*! \file azucar/utils/monitoredQueue.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef MONITOREDQUEUE_HPP_INCLUDED_
#define MONITOREDQUEUE_HPP_INCLUDED_

#include <deque>
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>
#include <boost/utility.hpp>

/*! \brief Queue of elements, with concurrency access through
 *  a Monitor pattern.
 * 
 * Only one thread can access the queue at the same time. All other
 * threads are blocked until it is their turn.
 * 
 * In blocking methods (popFront), the calling thread is blocked
 * until the queue is not empty. The non-blocking alternative is
 * popFrontDoNotWait. 
 */
template<class Elem> class MonitoredQueue : boost::noncopyable {
public:
  /*! \brief Create an empty MonitoredQueue.
   */
  MonitoredQueue();
  /*! \brief Obtain the first element of the queue, and remove it
   *  from the queue. If the queue is empty, it blocks until an
   *  element becomes available.
   */
  const Elem popFront();
  /*! \brief Obtain the first element of the queue, and remove it
   *  from the queue. If the queue is empty, it does not block,
   *  and returns false.
   * 
   * \param e[out] Element extracted from the front of the queue.
   *  Only valid if return value is true.
   * 
   * \return True, if an element could be extracted from the queue,
   *  false otherwise.
   */
  bool popFrontDoNotWait(Elem &e); // e is an output parameter
  /*! \brief Add an element at the end of the queue.
   */
  void pushBack(const Elem &e);
  ~MonitoredQueue();
private:
  boost::condition nonEmptyCondition; //!< Condition variable
    //!< to signal a non empty queue.
  mutable boost::mutex m_mutex; //!< Mutex to control concurrent
    //!< access to the Monitor.
  std::deque<Elem> internalQueue; //!< Internal queue to store
    //!< the elements.
}; 


template<class Elem> MonitoredQueue<Elem>::MonitoredQueue() {};
template<class Elem> MonitoredQueue<Elem>::~MonitoredQueue() {};

template<class Elem>  const Elem MonitoredQueue<Elem>::popFront() {
  boost::mutex::scoped_lock scoped_lock(m_mutex);
  if (internalQueue.empty()) {
    nonEmptyCondition.wait(scoped_lock);
  }
  Elem frontElement = internalQueue.front();
  internalQueue.pop_front();
  return frontElement;
}

template<class Elem> bool MonitoredQueue<Elem>::popFrontDoNotWait(Elem &e) {
  boost::mutex::scoped_lock scoped_lock(m_mutex);
  if (internalQueue.empty()) {
    return false;
  } else {
    e = internalQueue.front();
    internalQueue.pop_front();
    return true;
  }
}
template<class Elem>  void MonitoredQueue<Elem>::pushBack(const Elem &e) {
  boost::mutex::scoped_lock scoped_lock(m_mutex);
  internalQueue.push_back(e);
  nonEmptyCondition.notify_one();
  return;
}

#endif /*MONITOREDQUEUE_HPP_INCLUDED_*/
