/*! \file azucar/utils/network.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef NETWORK_HPP_
#define NETWORK_HPP_

#include <string>
#include <fstream>
#include <iomanip>


#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/array.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/range/iterator_range.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include <boost/asio.hpp>

#include <log4cpp/Category.hh>


#include <loki/ScopeGuard.h>

#include "azucar/utils/util.hpp"
#include "azucar/utils/exceptions.hpp"


namespace network {
const long length_field_size = 10; //!< Size for the header
  //!< field of a message, which contains the length, in bytes,
  //!< of the data part of the message.
const std::size_t buffer_size = 512; //!< Size for the read/write
  //!< buffer used for reading/writing through a socket.


const std::string category_name = "network_cat";

log4cpp::Category &logger();


/*! \brief Exception specific for network problems.
 * 
 * \sa azucar::utils::BaseException
 */
class NetworkException : public azucar::utils::BaseException {
public:
  /*! \sa azucar::utils::BaseException::BaseException()
   */
  NetworkException(const std::string &msg);
  
  /*! \sa azucar::utils::BaseException::~BaseException()
   */
  ~NetworkException() throw();       
};


/*! \brief Controller of an asynchronous network service.
 * 
 * It offers operations to start the service (blockingRun,
 * where the current thread is the one offering the service),
 * to notify the service that it must finish (asyncTerminate),
 * and to check whether the service has been notified to
 * end (through isTerminateSet).
 */
class NetAsyncService {
private:
  boost::asio::io_service &service_; //!< Network service managed
    //!< through this object.
  bool terminateSet_; //!< If true, the network service must
    //!< finish as soon as possible.
  boost::mutex m_mutex; //!< Mutex to control access to
    //!< terminateSet_.
public:
  /*! \brief Create a controller for the asynchronous service.
   * 
   * \param service Network asynchronous service to control.
   */
  NetAsyncService(boost::asio::io_service& service);
  
  /*! \brief Start servicing with this thread.
   * 
   * Current thread will block until notified to finish through
   * this same object with asyncTerminate. Meanwhile, it will
   * service all network requests.
   * 
   * \internal It must be checked whether there is any 
   * problem in between call of run() and reset(), in case
   * a client posts a network request. 
   */
  void blockingRun();

  /*! \brief Notify a possible thread running this service to end.
   */
  void asyncTerminate();
  
  /*! \brief Inform whether asyncTerminate has been called on
   *  this object.
   */
  bool isTerminateSet();
};

/*! \brief Represents a network operation through its lifetime.
 * 
 * A network operation consists either on reading or writing
 * a message through a network socket. An additional network operation
 * is available, which discards all further data received
 * through the associated socket. A token can only serve a single
 * operation during its lifetime.
 * 
 * A message consists of a header and data, and represents
 * a field of information. The header contains metadata for
 * interpretation of the data.
 * 
 * For special data (such as a file), see FileNetworkOpToken. 
 */
template<class Socket>
class NetworkOpToken : public boost::noncopyable {
public:
//  boost::asio::ip::tcp::socket &socket_; //!< Socket assigned for this network operation.
  Socket &socket_;
protected:
  std::string data_; //!< The data to send/receive.
  boost::condition operationDone; //!< Condition variable
    //!< to signal the end of the subyacent asynchronous operation.
    //!< Used with m_mutex.
  boost::mutex m_mutex; //!< Mutex to ensure exclusive access
    //!< to the socket, and for waiting for the subyacent 
    //!< asynchronous operation to be finished (\sa operationDone). 
public:
  /*! \brief Error Status Type
   */
  enum Error {
      NO_ERROR, //!< No error happened. 
      COMMUNICATION_ERROR //!< A communication error happened.
  };
  static std::string end_of_record; //!< Single character string
    //!< to signal end of record. It is useful for
    //!< allowing sending packets of null strings, which require
    //!< this delimiter.
  Error error; //!< Error status.
  boost::array<char, length_field_size> headerBuffer_; //!< Buffer for storing the
    //!< message header.
  size_t data_pending_; //!< buffer length still pending to be sent/received.
  boost::array<char, buffer_size> readBuffer_; //!< Buffer for message data.

  /*! \brief Synchronous read 
   * of the next data received through the associated
   * socket. 
   * 
   * Check error status before retrieving the data by using 
   * getData().
   */
  void readField();
  
  /*! \brief Synchronous write (send) 
   * the given data through the associated
   *  socket.
   * 
   * \param str The plain data to send (it must be codified
   *  in a format understandable by the receiving peer).
   */
  void writeField(const std::string &str);

  /*! \brief Synchronously read all data sent by the sending
   * peer, until end of transmission (EOF).
   */
  void readAndDiscard();
  
  /*! \brief Internal method to be called by the asynchronous
   *  processing handlers, in order to convert the network
   *  message read into the format expected (i.e.: write 
   *  to a file; \sa FileNetworkOpToken::processNewBuffer()).
   * 
   * \exception azucar::utils::BaseException If something goes wrong with
   *  the processing.
   */
  virtual void processNewBuffer(size_t bytes_transferred);
  
  /*! \brief Obtain the data received through a readField() 
   *  operation.
   * 
   * Check error status before obtaining the data.
   * 
   * The data is codified as plain text, in a format
   * that the caller must agree with the sender. 
   */
  const std::string &getData();

  /*! \brief To be used by the internal asynchronous service
   *  to notify a network error during the token operation,
   *  and to notify a waiting thread for the end of the 
   *  asynchronous request.
   */
  virtual void notifyOpEndWithError(const boost::system::error_code &error);
  
  /*! \brief To be used by the internal asynchronous service
   *  to notify a non network error during the token operation,
   *  and to notify a waiting thread for the end of the 
   *  asynchronous request.
   */
  virtual void notifyOpEndWithError();
  
  /*! \brief To be used by the internal asynchronous service
   *  to notify a successful network token operation completed
   *  and to notify a waiting thread for the end of the
   *  asynchronous request.
   */
  void notifyOpEnd();

  /*! \brief Create a token for an operation not yet indicated,
   *  on the associated socket.
   * 
   * \param socket The socket to use for the network operation.
   */
  NetworkOpToken(Socket &socket);

  virtual ~NetworkOpToken();

};





/*! \brief Represents a network operation for reading
 *  data and writing it to a file.
 */
template<class Socket>
class FileNetworkOpToken : public NetworkOpToken<Socket> {
private:
  std::ofstream file_; //!< Output file to write the data.
public:
  const boost::filesystem::path path_; //!< Path of the
  //!< output file to write the data.

  /*! \sa NetworkOpToken::processNewBuffer()
   */
  void processNewBuffer(size_t bytes_transferred);

  /*! \brief Create a token for reading data and writing it
   * in a file.
   * 
   * \param socket Asynchronous socket to read data from.
   * \param path Path of the output file to write the data to.
   *  The file is created if it does not exist (but its parent directory
   *  must already exist), or truncated if it exists.
   */
  FileNetworkOpToken(Socket &socket, 
      const boost::filesystem::path &path);
      
  /*! \sa NetworkOpToken::notifyOpEndWithError(const boost::system::error_code&)
   */
  void notifyOpEndWithError(const boost::system::error_code &error);


  /*! \sa NetworkOpToken::notifyOpEndWithError()
   */
  void notifyOpEndWithError();

  /*! \sa NetworkOpToken::notifyOpEnd()
   */
  void notifyOpEnd();

  ~FileNetworkOpToken();
       
};



/*! \brief Read synchronously from asynchronous socket.
 * 
 * \tparam T Type of the data to read (must be one of numeric types
 *  or a string).
 * \param socket asynchronous socket to read from.
 * 
 * \return Data read.
 * 
 * \exception NetworkException There was a problem with the network.
 * \exception azucar::utils::BaseException There was a problem converting the
 *  data read to T.
 */
template<class T, class Socket> const T read(Socket &socket) {
  NetworkOpToken<Socket> *token = new NetworkOpToken<Socket>(socket);
  ON_BLOCK_EXIT(azucar::utils::deleteFunction<NetworkOpToken<Socket> >, token);
  
  token->readField();
  
  if (token->error != NetworkOpToken<Socket>::NO_ERROR) {
    throw NetworkException("Network error");
  } else {
    try {
      return boost::lexical_cast<T>(boost::trim_copy(token->getData()));
    } catch (boost::bad_lexical_cast &ex) {
      throw azucar::utils::BaseException("Problem converting data read to right format.");
    }
  }
}



/*! \brief Write synchronously to an asynchronous socket.
 * 
 * \tparam T Type of the data to write (must be one of numeric types
 *  or a string).
 * \param socket asynchronous socket to write to.
 * \param t Data of type T to write through the socket.
 * 
 * \exception NetworkException There was a problem with the network.
 * \exception azucar::utils::BaseException There was a problem converting the
 *  data to a suitable format for writing.
 */
template<class T, class Socket> void write(Socket &socket,
                            const T& t) {

  NetworkOpToken<Socket> *token = new NetworkOpToken<Socket>(socket);
  ON_BLOCK_EXIT(azucar::utils::deleteFunction<NetworkOpToken<Socket> >, token);
  
  try {
    token->writeField(boost::lexical_cast<std::string>(t));
  } catch (boost::bad_lexical_cast &ex) {
    throw azucar::utils::BaseException("Problem converting data to write in the right format.");
  }
  
  if (token->error != NetworkOpToken<Socket>::NO_ERROR) {
    throw NetworkException("Network error");
  }
  
  return;
}

/*! \brief Handle for async network reading. Data read
 *  must be discarded.
 * 
 * If end of connection is detected, the thread waiting for
 * this operation in the token monitor is signaled about
 * the end of this operation.
 * 
 * If not, further reading from the network is performed,
 * until all data is discarded and the connection shut down.
 * 
 * If there is an error, the thread waiting for this
 * operation in the token monitor is signaled and notified
 * about the error.
 */
template<class Socket>
void discard_buffer(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred);
    
/*! \brief Handle for async network reading. Header
 *  data is read, and further processing is set up.
 * 
 * If there is an error, the thread waiting for this
 * operation in the token monitor is signaled and notified
 * about the error.
 */
template<class Socket>
void handle_header_read(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred);

/*! \brief Handle for async network reading. Message
 *  data is read, and further processing is set up.
 * 
 * Further data reading may be needed and is setup
 * in this function.
 * 
 * If operation is completed, the thread waiting for
 * this operation in the token monitor is signaled.
 * 
 * If there is an error, the thread waiting for this
 * operation in the token monitor is signaled and notified
 * about the error.
 */
template<class Socket>
void handle_data_read(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred);
    
///*! \brief Handle for async network reading. Message
// *  end of record is read.
// * 
// * If operation is completed, the thread waiting for
// * this operation in the token monitor is signaled.
// * 
// * If there is an error, the thread waiting for this
// * operation in the token monitor is signaled and notified
// * about the error.
// */
//void handle_end_of_record_read(NetworkOpToken *token,
//    const boost::system::error_code& error, size_t bytes_transferred);    


/*! \brief Handle for async network reading. Message
 *  header is written, and further processing is set up.
 * 
 * Further writing may be needed, and is set up
 * inside this function.
 * 
 * If there is an error, the thread waiting for this
 * operation in the token monitor is signaled and notified
 * about the error.
 */
template<class Socket>
void handle_header_write(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred);


/*! \brief Handle for async network reading. Message
 *  data is written, and further processing is set up.
 * 
 * Further writing may be needed, and is set up
 * inside this function.
 * 
 * If operation is completed, the thread waiting for
 * this operation in the token monitor is signaled.
 * 
 * If there is an error, the thread waiting for this
 * operation in the token monitor is signaled and notified
 * about the error.
 */
template<class Socket>
void handle_data_write(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred);

///*! \brief Handle for async network writing. Message
// *  end of record is written.
// * 
// * If operation is completed, the thread waiting for
// * this operation in the token monitor is signaled.
// * 
// * If there is an error, the thread waiting for this
// * operation in the token monitor is signaled and notified
// * about the error.
// */
//void handle_end_of_record_write(NetworkOpToken *token,
//    const boost::system::error_code& error, size_t bytes_transferred);
    

/*! \brief Asynchronous read of the header for the next message.
 * 
 * \param token Information about the connection to use,
 * and the way to store the data and header metadata.
 * 
 * \sa readField() for asynchronous behaviour. 
 */
template<class Socket>
void readHeader(NetworkOpToken<Socket> *token);

/*! \brief Asynchronous read 
 * of the next data received through the associated
 * socket in token.
 *  
 * \param token Information about the connection to use,
 * and the way to store the data and header metadata. 
 * 
 * If error, token error status is set.
 * 
 * If not error, data is stored by calling token
 * processNewBuffer method for each buffer of data read.
 * 
 * Control is returned as soon as the reading request is
 * sent to the asynchronous socket.
 */
template<class Socket>
void readField(NetworkOpToken<Socket> *token);

/*! \brief Asynchronous write (send) 
 * the given data through the associated
 *  socket.
 * 
 * \param token Information about the connection and
 *  the data to send.
 * 
 * Control is returned as soon as the writing request is
 * sent to the asynchronous socket.
 */
template<class Socket>
void writeField(NetworkOpToken<Socket> *token);

/*! \brief Asynchronous write of the header for the next message.
 * 
 * \param token Information about the connection to use,
 * and the way to obtain the data and header metadata.
 * 
 * \sa writeField() for asynchronous behaviour. 
 */
template<class Socket>
void writeHeader(NetworkOpToken<Socket> *token);


/*! \brief Discard, sinchronously, all data received through
 *  socket until EOF.
 * 
 * \param socket Network socket to receive data from.
 * 
 * \exception NetworkException There was a problem with the network.
 */
template<class Socket>
void discardAllData(Socket &socket) {
  NetworkOpToken<Socket> *token = new NetworkOpToken<Socket>(socket);
  ON_BLOCK_EXIT(azucar::utils::deleteFunction<NetworkOpToken<Socket> >, token);

  token->readAndDiscard();
  
  if (token->error != NetworkOpToken<Socket>::NO_ERROR) {
    throw NetworkException("Network error");
  }
    
  return;
}

template<class Socket>
void discard_buffer(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred) {

  if (error) {
    if (error == boost::asio::error::eof) {
      token->notifyOpEnd();
    } else {
      token->notifyOpEndWithError(error);
    }
  } else {
    boost::asio::async_read(token->socket_, 
      boost::asio::buffer(token->headerBuffer_, length_field_size),
      boost::bind(discard_buffer<Socket>, token,
        boost::asio::placeholders::error,
        boost::asio::placeholders::bytes_transferred));
  }  
}




/*! \brief Read synchronously from asynchronous socket and write
 *  data to file.
 * 
 * \param socket asynchronous socket to read from.
 * \param path Path to the file where data must be written. The
 *  file is created if it does not exist (but its parent directory
 *  must already exist), or truncated if it exists.
 * 
 * \exception NetworkException There was a problem with the network.
 */
template<class Socket>
void readToFile(Socket &socket, 
                const boost::filesystem::path &path) {
  FileNetworkOpToken<Socket> *token;
  token = new FileNetworkOpToken<Socket>(socket, path);
  
  ON_BLOCK_EXIT(azucar::utils::deleteFunction<FileNetworkOpToken<Socket> >, token);
  
  token->readField();

  if (token->error != NetworkOpToken<Socket>::NO_ERROR) {
    throw azucar::utils::BaseException("Network error");
  }
  return;

}

template<class Socket>
void readField(NetworkOpToken<Socket> *token) {
  readHeader(token);
}

template<class Socket>
void writeField(NetworkOpToken<Socket> *token) {
  writeHeader(token);
}

template<class Socket>
void writeHeader(NetworkOpToken<Socket> *token) {
    boost::array<char, length_field_size>::iterator end_point
      = token->headerBuffer_.begin();
    std::advance(end_point, length_field_size);
    boost::iostreams::filtering_ostream out(boost::make_iterator_range(token->headerBuffer_.begin(),
      end_point));
    out << std::setw(length_field_size) << token->getData().size();
    out.flush();


    if (out.fail()) {
//      token->notifyOpEndWithError(NetworkOpToken::COMMUNICATION_ERROR);
      token->notifyOpEndWithError();
    } else {
      boost::asio::async_write(token->socket_,
      boost::asio::buffer(token->headerBuffer_, length_field_size),
                boost::bind(handle_header_write<Socket>, token,
                boost::asio::placeholders::error,
                boost::asio::placeholders::bytes_transferred));
    }
}

template<class Socket>
void readHeader(NetworkOpToken<Socket> *token) {
    network::logger().debug("About to read header.");
    boost::asio::async_read(token->socket_, 
      boost::asio::buffer(token->headerBuffer_, length_field_size),
      boost::bind(handle_header_read<Socket>, token,
        boost::asio::placeholders::error,
        boost::asio::placeholders::bytes_transferred));
}

template<class Socket>
void handle_header_read(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred) {

  // if error -> notify error
  if (error) {
    network::logger().debug("Header read with error.");
    token->notifyOpEndWithError(error);
  } else {
    network::logger().debug("Header read ok.");
    boost::iostreams::filtering_istream in(boost::make_iterator_range(token->headerBuffer_));
    in >> token->data_pending_;
    if (in.fail()) {
      network::logger().debug("Error reading data_pending.");
//      token->notifyOpEndWithError(NetworkOpToken::COMMUNICATION_ERROR);
      token->notifyOpEndWithError();
    } else {
      if (token->data_pending_ == 0U) {
        // received empty message
        network::logger().debug("Received empty message.");
        token->notifyOpEnd();
      } else {
        size_t bytes_to_read = 
          (token->data_pending_ < buffer_size) ? token->data_pending_ : buffer_size;
        network::logger().debug("About to read data.");
        boost::asio::async_read(token->socket_, 
          boost::asio::buffer(token->readBuffer_, bytes_to_read),
          boost::bind(handle_data_read<Socket>, token,
            boost::asio::placeholders::error,
            boost::asio::placeholders::bytes_transferred));
      }
    }
  }
}

template<class Socket>
void handle_header_write(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred) {
  // if error -> notify error
  if (error) {
    network::logger().debug("Header written with error.");
    token->notifyOpEndWithError(error);
  } else {
    network::logger().debug("Header write ok.");
      boost::asio::async_write(token->socket_, 
        boost::asio::buffer(token->getData(), token->getData().size()),
        boost::bind(handle_data_write<Socket>, token,
          boost::asio::placeholders::error,
          boost::asio::placeholders::bytes_transferred));
  }
}

template<class Socket>
void handle_data_read(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred) {

  // if error -> notify error
  if (error) {
    network::logger().debug("Data read error.");
    token->notifyOpEndWithError(error);
  } else {
    token->data_pending_ -= bytes_transferred;
    try {
      token->processNewBuffer(bytes_transferred);
    } catch (azucar::utils::BaseException &ex) {
      network::logger().debug("Processing new buffer exception");
      token->notifyOpEndWithError();
    }
    if (token->data_pending_ == 0) {
//      boost::asio::async_read(token->socket_, 
//        boost::asio::buffer(token->readBuffer_, NetworkOpToken::end_of_record.size()),
//        boost::bind(handle_end_of_record_read, token,
//          boost::asio::placeholders::error,
//          boost::asio::placeholders::bytes_transferred));
      token->notifyOpEnd();
    } else {
      size_t bytes_to_read = 
        (token->data_pending_ < buffer_size) ? token->data_pending_ : buffer_size;
      network::logger().debug("About to read data again.");
      network::logger().debug(boost::lexical_cast<std::string>(token->data_pending_));
      boost::asio::async_read(token->socket_, 
        boost::asio::buffer(token->readBuffer_, bytes_to_read),
        boost::bind(handle_data_read<Socket>, token,
          boost::asio::placeholders::error,
          boost::asio::placeholders::bytes_transferred));
    }
  }
}

//void handle_end_of_record_read(NetworkOpToken *token,
//    const boost::system::error_code& error, size_t bytes_transferred) {
//
//  // if error -> notify error
//  if (error) {
//    network::logger().debug("Data read error.");
//    token->notifyOpEndWithError(error);
//  } else {
//    if (bytes_transferred != NetworkOpToken::end_of_record.size()) {
//      network::logger().debug("End of record wrongly read.");
//      token->notifyOpEndWithError(error);
//    } else {
//      token->notifyOpEnd();
//    }
//  }
//}


template<class Socket>
void handle_data_write(NetworkOpToken<Socket> *token,
    const boost::system::error_code& error, size_t bytes_transferred) {
  // if error -> notify error
  if (error) {
    network::logger().debug("Data write error.");
    token->notifyOpEndWithError(error);
  } else {
      token->notifyOpEnd();

//      boost::asio::async_write(token->socket_, 
//        boost::asio::buffer(NetworkOpToken::end_of_record, 
//          NetworkOpToken::end_of_record.size()),
//        boost::bind(handle_end_of_record_write, token,
//          boost::asio::placeholders::error,
//          boost::asio::placeholders::bytes_transferred));
  }
}

//void handle_end_of_record_write(NetworkOpToken *token,
//    const boost::system::error_code& error, size_t bytes_transferred) {
//  // if error -> notify error
//  if (error) {
//    network::logger().debug("Data write error.");
//    token->notifyOpEndWithError(error);
//  } else {
//      token->notifyOpEnd();
//  }
//}

template<class Socket>
void NetworkOpToken<Socket>::readField() {
    boost::mutex::scoped_lock scoped_lock(m_mutex);
    network::readField(const_cast<NetworkOpToken<Socket>*>(this));
    network::logger().debug("About to wait.");
    operationDone.wait(scoped_lock);
    network::logger().debug("After wait.");
  }
  
template<class Socket>
void NetworkOpToken<Socket>::writeField(const std::string &str) {
  boost::mutex::scoped_lock scoped_lock(m_mutex);
  data_ = str;
  network::writeField(const_cast<NetworkOpToken<Socket>*>(this));
  operationDone.wait(scoped_lock);
}
  
template<class Socket>
void NetworkOpToken<Socket>::readAndDiscard() {
    boost::mutex::scoped_lock scoped_lock(m_mutex);
    network::logger().debug("About to read and discard.");
    boost::asio::async_read(socket_, 
      boost::asio::buffer(headerBuffer_, length_field_size),
      boost::bind(discard_buffer<Socket>, this,
        boost::asio::placeholders::error,
        boost::asio::placeholders::bytes_transferred));
    network::logger().debug("About to wait and discard.");
    operationDone.wait(scoped_lock);
  }

template<class Socket>
void NetworkOpToken<Socket>::processNewBuffer(size_t bytes_transferred) {
    network::logger().debug("Processing buffer.");
    boost::array<char, length_field_size>::iterator end_point
      = this->readBuffer_.begin();
    std::advance(end_point, bytes_transferred);
    data_.append(this->readBuffer_.begin(), end_point);
  }
  
template<class Socket>
const std::string &NetworkOpToken<Socket>::getData() {
    network::logger().debugStream() << "Data is: '"
      << data_ << "'" << log4cpp::CategoryStream::EOL;
    return data_;
}

template<class Socket>
void NetworkOpToken<Socket>::notifyOpEndWithError() {
    this->error = COMMUNICATION_ERROR;
    boost::mutex::scoped_lock scoped_lock(m_mutex);
    network::logger().debug("About to notify op end with error.");
    operationDone.notify_one();
  }

template<class Socket>
void NetworkOpToken<Socket>::notifyOpEndWithError(const boost::system::error_code &error) {
  notifyOpEndWithError();
  }
  
template<class Socket>
void NetworkOpToken<Socket>::notifyOpEnd() {
    boost::mutex::scoped_lock scoped_lock(m_mutex);
    network::logger().debug("About to notify op end.");
    operationDone.notify_one();
  } 
  
template<class Socket>
NetworkOpToken<Socket>::NetworkOpToken(Socket &socket) 
    : socket_(socket), error(NO_ERROR), data_pending_(0) {};
    
template<class Socket>
NetworkOpToken<Socket>::~NetworkOpToken() {};

template<class Socket>
std::string NetworkOpToken<Socket>::end_of_record = "*";


template<class Socket>
void FileNetworkOpToken<Socket>::processNewBuffer(size_t bytes_transferred) {
    boost::array<char, length_field_size>::iterator end_point
      = this->readBuffer_.begin();
    std::advance(end_point, bytes_transferred);
    boost::iostreams::filtering_istream in(
        boost::make_iterator_range(this->readBuffer_.begin(), end_point));
    // should not fail with the filtering_istream, so we do not check
    // failures there
    try {        
      file_ << in.rdbuf();
    } catch (std::ios_base::failure &failure) {
      network::logger().debug("Failure writing network data to file.");
      throw azucar::utils::BaseException("Failure writing network data to file.");
    }
  }
  
template<class Socket>
FileNetworkOpToken<Socket>::FileNetworkOpToken(Socket &socket, 
      const boost::filesystem::path &path) 
      : NetworkOpToken<Socket>(socket), path_(path) {
    file_.open(path_.string().c_str());
    file_.exceptions(std::ios_base::badbit | std::ios_base::failbit);
  };

template<class Socket>
void FileNetworkOpToken<Socket>::notifyOpEndWithError(const boost::system::error_code &error) {
    file_.close();
    NetworkOpToken<Socket>::notifyOpEndWithError(error);
  }
  
template<class Socket>
void FileNetworkOpToken<Socket>::notifyOpEndWithError() {
    file_.close();
    NetworkOpToken<Socket>::notifyOpEndWithError();
  }

  
template<class Socket>
void FileNetworkOpToken<Socket>::notifyOpEnd() {
  file_.close();
  NetworkOpToken<Socket>::notifyOpEnd();
}  
  
template<class Socket>
FileNetworkOpToken<Socket>::~FileNetworkOpToken() {
    if (file_.is_open()) {
      file_.close();
    }
  }


}



#endif /*NETWORK_HPP_*/
