/*! \file azucar/utils/param.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef PARAM_HPP_
#define PARAM_HPP_

/*! \brief Optional or default parameter for execution of a function,
 * service or program.
 * 
 * When isDefined() is false, the function, service or program
 * may decide, according to its own semantics, whether the
 * parameter is not used (that is, it is optional), or
 * whether it is given a default value.
 */
template<class T> class NewParam {
private:
  bool defined_; //!< Whether the user has defined the parameter or
  //!< not.
  T value_; //!< Value of the parameter, as defined by the user.
  //!< Only valid when defined_ is true.
public:
  /*! \brief Create an undefined parameter.
   */
  NewParam() : defined_(false) {};
  /*! \brief Create a defined parameter, with the given value.
   */
  NewParam(const T &value) : defined_(true), value_(value) {};
  /*! \brief Create a new parameter, as a copy of other.
   */
  NewParam(const NewParam<T> &other) : defined_(other.defined_),
    value_(other.value_) {};
    
  /*! \brief Is this parameter defined?
   */
  bool isDefined() const;

  /*! \brief Obtain the value of this parameter.
   * 
   * \pre isDefined() is true.
   */
  operator T() const;

  /*! \brief Convert this parameter in a defined parameter of value
   * otherValue.
   */
  NewParam<T> &operator=(const T& otherValue) {
    defined_ = true;
    value_ = otherValue;
    return *this;
  }
  
  /*! \brief Convert this parameter in a copy of other.
   */
  NewParam<T> &operator=(const NewParam<T>& other) {
    if (this != &other) {
      defined_ = other.defined_;
      value_ = other.value_;
    }
    
    return *this;
  }
  
  /*! \brief Obtain the value of this parameter.
   * 
   * \pre isDefined() is true.
   */
  const T getValue(const T otherValue) const;
};

template<class T> bool NewParam<T>::isDefined() const {
  return defined_;
}

template<class T> NewParam<T>::operator T() const {
  assert(defined_);
  return value_;
}

template<class T> const T NewParam<T>::getValue(const T otherValue) const {
  if (defined_) {
    return value_;
  } else {
    return otherValue;
  }
}

//template<class T> class Param {
//private:
//  bool required_;
//  bool defined_;
//  T value_;
//public:
//  Param() : required_(false), defined_(false) {};
//  Param(bool required, const T value) : required_(required),
//    value_(value), defined_(true) {};
//  bool isRequired() const;
//  void setRequired(bool required);
//  const T getValue() const;
//  operator T() const;
//  void setValue(const T value);
//  bool isDefined() const;
//};
//
//template<class T> bool Param<T>::isRequired() const {
//  return required_;
//}
//
//template<class T> void Param<T>::setRequired(bool required) {
//  required_ = required;
//}
//
//template<class T> const T Param<T>::getValue() const {
//  assert(defined_);
//  return value_;
//}
//
//template<class T> Param<T>::operator T() const {
//  return this->getValue();
//}
//
//template<class T> void Param<T>::setValue(const T value) {
//  value_ = value;
//  defined_ = true;
//}
//
//template<class T> bool Param<T>::isDefined() const{
//  return defined_;
//}

#endif /*PARAM_HPP_*/
