/*! \file azucar/utils/process.hpp
 *
 * \brief Utilities for running processes. The interface is system independent.
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef PROCESS_HPP_INCLUDED_
#define PROCESS_HPP_INCLUDED_

#include <string>
#include <map>
#include <vector>

#include <boost/function.hpp>

#include "azucar/utils/exceptions.hpp"

/*! \brief Parameters needed to launch a new process.
 */
class ProcessRequest {
public:
  std::string processPath_; //!< path to the executable file to run.
  std::vector<std::string> argv_; //!< arguments to pass to the new process (first argument should
  //!< be the program name).
  std::map<std::string,std::string> env_; //!< environment in which to execute the program, where env keys are the
  //!< environment variables, and the env value associated to an env key is the
  //!< environment value associated to the environment variable.
  std::string inputFilePath_; //!< The program will receive input (stdin) from the file 
  //!< identified by inputFilePath (which could be /dev/null),
  std::string outputAndErrorFilePath_; //!< it will write its 
  //!< standard output and standard error to file outputAndErrorFilePath.
  std::string workingDirectory_; //!< Path where the process will be launched.
  long pid_; //!< PID of the launched process,
  //!< only valid after process has been launched.
  //!< Initially, its value is 0.

  /*! Create a ProcessRequest with the needed parameters. See
   * object attribute definitions for information about
   * the parameters.
   */
  ProcessRequest(const std::string &processPath,
       const std::vector<std::string> &argv,
       const std::map<std::string,std::string> &env,
       const std::string &inputFilePath,
       const std::string &outputAndErrorFilePath,
       const std::string &workingDirectory) 
      : processPath_(processPath), argv_(argv), env_(env),
        inputFilePath_(inputFilePath), 
        outputAndErrorFilePath_(outputAndErrorFilePath),
        workingDirectory_(workingDirectory),
        pid_(0) {};  
};

/*! \brief Exception specific for process problems.
 * 
 * \sa azucar::utils::BaseException
 */
class ProcessException : public azucar::utils::BaseException {
public:
  /*! \sa azucar::utils::BaseException::BaseException()
   */
  ProcessException(const std::string &msg);
  
  /*! \sa azucar::utils::BaseException::~BaseException()
   */
  ~ProcessException() throw();       
};

//! Convert an OS environ set of variables to the corresponding
//! map structure.
/*!
 * \param environ a pointer to pointer of char 
 * (that is, a pointer to an array of char*,
 * where the array is const and the char* are const too).
 * Each char* string (null terminated) in environ must be of the following
 * form: KEY=VALUE
 * \return the corresponding map indexed by string equivalent of KEY,
 * and whose value is the string equivalent of VALUE.
 * 
 * \note This access should be thread-safe as long as no other
 * thread is writing the environ variable. 
 * 
 * \exception azucar::utils::BaseException If there is formatting error in environ.
 */
std::map<std::string, std::string> convertEnvironToMap(const char * const *environ);

/*! Obtain process ID of current process.
 * 
 * \return PID of current process.
 */
long getProcessID();

//! It executes, synchronously, the program whose path is processPath.
/*!
 * \param processPath path to the executable file to run.
 * \param argv arguments to pass to the new process (first argument should
 * be the program name).
 * \param env environment in which to execute the program, where env keys are the
 * environment variables, and the env value associated to an env key is the
 * environment value associated to the environment variable.
 *  * \param inputFilePath The program will receive input (stdin) from the file 
 * identified by inputFilePath (which could be /dev/null), 
 * \param outputAndErrorFilePath it will write its 
 * standard output and standard error to file outputAndErrorFilePath.
 *
 * \note VERY IMPORTANT If executing in a multithreading environment,
 * this must be a BOOST multithreading environment.
 * This code currently only works on unix systems, in a 
 * single threaded environment, or in a multithreading
 * BOOST environment using PTHREADS.
 * 
 * \note OBSOLETE In a multithreading environment, this function
 * calls sigaction(). Therefore, access to this function
 * should be done without other thread calling a related
 * signal function (like sigaction() or sigwait()), as
 * the Open Unix standard states for the sigaction() function:
 * <i>The result of the use of sigaction() and a sigwait() 
 * function concurrently within a process on the same signal is 
 * unspecified.</i> OBSOLETE
 * 
 * \note In case the process executed returns -1 as the exit code,
 * there is no way to distinguish it from an error in calling the
 * process to execute.
 * 
 * \return This function will return the exit code of the execution of the program
 * identified by processPath, or -1 or 127 in case of error.
 */
int processExecute(const std::string &processPath,
		   const std::vector<std::string> &argv,
		   const std::map<std::string,std::string> &env,
		   const std::string &inputFilePath,
		   const std::string &outputAndErrorFilePath,
       const std::string &workingDirectory = "");

/*! \brief It executes, asynchronously, the program whose path is processPath.
 * \param processPath path to the executable file to run.
 * \param argv arguments to pass to the new process (first argument should
 * be the program name).
 * \param env environment in which to execute the program, where env keys are the
 * environment variables, and the env value associated to an env key is the
 * environment value associated to the environment variable.
 *  * \param inputFilePath The program will receive input (stdin) from the file 
 * identified by inputFilePath (which could be /dev/null), 
 * \param outputAndErrorFilePath it will write its 
 * standard output and standard error to file outputAndErrorFilePath.
 *
 * \return This function will return -1 in case of error, child pid otherwise.
 * Notice that it will not return the exit code of the process, as in 
 * processExecute, since it executes asynchronously.
 */
long processExecuteAsynchronous(const std::string &processPath,
       const std::vector<std::string> &argv,
       const std::map<std::string,std::string> &env,
       const std::string &inputFilePath,
       const std::string &outputAndErrorFilePath,
       const std::string &workingDirectory = "",
       bool doClearZombies = true);

/*! \brief Wait for all finished child processes that were launched
 * asnynchronously.
 * 
 * \post Any children launched asynchronously (through process
 * ExecuteAsynchronous, timedProcessExecute
 * or timedMultiProcessExecuteAsynchrnonous)
 * that has already finished is cleared.
 */
void clearZombies();

/*! \brief Execute several processes and wait, to a maximum
 * time, for completion.
 * 
 * maxSecondsToWait and nanoPollTime are hints, and no real-time
 * notion of these values is considered.
 * 
 * \post As a secondary effect, all zombie processes launched
 * asynchronously with processExecuteAsynchronous,
 * timedProcessExecute or timedMultiProcessExecute
 * are waited if maxSecondsToWait time is exceeded. Therefore,
 * NO CONCURRENT CALLS TO processExecuteAsynchronous,
 * timedProcessExecute nor timedMultiProcessExecute should be
 * executed until the functions have returned or, 
 * in case of processExecuteAsynchronous, the child has
 * already been waited and cleaned. 
 *   
 * \param requests Information about the processes requested to
 * be launched.
 * \param maxSecondsToWait Maximum number of seconds to wait
 * for completion of requested processes. If this time goes by
 * and there are processes not yet finished, they are killed.
 * \param nanoPollTime Period of time for polling for completion
 * of processes launched.
 *
 * \return This function will return true if the processes
 * launched finish before maxSeondsToWait elapses, or false if at
 * least one process had to be killed for exceeding maxSecondsToWait
 * time of execution.
 * 
 * \sa processExecuteAsynchronous
 * \sa timedProcessExecute
 */
bool timedMultiProcessExecute(std::vector<ProcessRequest> &requests,
      unsigned int maxSecondsToWait,
      long nanoPollTime);

/*! \brief As timedMultiProcessExecute, but separate launching of all processes
 * by nanoDelayTime.
 *
 * \param maxSecondsToWait Maximum number of seconds to wait
 * for completion of requested processes, after all of them have been launched.
 * If this time goes by and there are processes not yet finished, they are killed.
 * \param nanoDelayTime Time (in nanoseconds) to wait before lanching each
 *  process after the previous one.
 *
 */
bool timedMultiProcessExecuteWithDelay(std::vector<ProcessRequest> &requests,
      unsigned int maxSecondsToWait,
      long nanoPollTime, long nanoDelayTime);

/*! \brief Execute one process and wait, to a maximum
 * time, for completion.
 * 
 * maxSecondsToWait and nanoPollTime are hints, and no real-time
 * notion of these values is considered.
 * 
 * \post As a secondary effect, all zombie processes launched
 * asynchronously with processExecuteAsynchronous,
 * timedProcessExecute or timedMultiProcessExecute
 * are waited if maxSecondsToWait time is exceeded. Therefore,
 * NO CONCURRENT CALLS TO processExecuteAsynchronous,
 * timedProcessExecute nor timedMultiProcessExecute should be
 * executed until the functions have returned or, 
 * in case of processExecuteAsynchronous, the child has
 * already been waited and cleaned. 
 *   
 * \param processPath path to the executable file to run.
 * \param maxSecondsToWait Maximum number of seconds to wait
 * for completion of the launched process. If this time goes by
 * and the process has not yet finished, it is killed.
 * \param nanoPollTime Period of time for polling for completion
 * of the launched process.
 * \param argv arguments to pass to the new process (first argument should
 * be the program name).
 * \param env environment in which to execute the program, where env keys are the
 * environment variables, and the env value associated to an env key is the
 * environment value associated to the environment variable.
 * \param inputFilePath The program will receive input (stdin) from the file 
 * identified by inputFilePath (which could be /dev/null), 
 * \param outputAndErrorFilePath it will write its 
 * standard output and standard error to file outputAndErrorFilePath.
 * \param workingDirectory Path where the process will be launched.
 *
 * \return This function will return true if the process
 * launched finishes before maxSeondsToWait elapses, or false if it
 * had to be killed for exceeding maxSecondsToWait
 * time of execution.
 * 
 * \sa processExecuteAsynchronous
 * \sa timedMultiProcessExecute
 */
bool timedProcessExecute(const std::string &processPath,
       unsigned int maxSecondsToWait,
       long nanoPollTime,
       const std::vector<std::string> &argv,
       const std::map<std::string,std::string> &env,
       const std::string &inputFilePath,
       const std::string &outputAndErrorFilePath,
       const std::string &workingDirectory = "");


void processExecuteIPC(const std::string &processPath,
       void *args, size_t argsSize,
       void *results, size_t resultsSize,
       unsigned int maxSecondsToWait,
       const std::vector<std::string> &argv,
       const std::map<std::string,std::string> &env,
       const std::string &inputFilePath,
       const std::string &outputAndErrorFilePath,
       const std::string &workingDirectory = "");

void clientExecuteIPC(int shmid, int semid, void *args, 
            size_t argsSize, void *results,
            size_t resultsSize, unsigned int secondsToWaitPerCall,
            boost::function<void (void)> callbackFunction);            

/*! Kill (SIGKILL) the process with the given pid.
 * 
 * \post As a secondary effect, all zombie processes launched
 * asynchronously with processExecuteAsynchronous,
 * timedProcessExecute or timedMultiProcessExecute
 * are waited. Therefore,
 * NO CONCURRENT CALLS TO processExecuteAsynchronous,
 * timedProcessExecute nor timedMultiProcessExecute should be
 * executed until the functions have returned or, 
 * in case of processExecuteAsynchronous, the child has
 * already been waited and cleaned.
 * 
 * \param pid Process ID of the process to kill. 
 */
void callKill(long pid);

     

#endif // PROCESS_HPP_INCLUDED_
