/*! \file azucar/utils/quantity.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef QUANTITY_HPP_
#define QUANTITY_HPP_

#include <string>
#include <iostream>
#include <cassert>
#include <functional>
#include <cmath>

#include "azucar/utils/mathConstants.hpp"

namespace quantity {

class Unit {
private:
	Unit &operator=(const Unit &other); // unimplemented. Not allowed.
public:
  std::string symbol_;
  
  Unit(const std::string &symbol) : symbol_(symbol) {};
  Unit() : symbol_("Dimensionless") {};
  bool operator==(const Unit &right) const {
    return this->symbol_ == right.symbol_;
  }

  Unit(const Unit& other) : symbol_(other.symbol_) {};

  Unit &transform(const Unit &newUnit);

};

class Prefix {
public:
  const std::string symbol_;
  
  Prefix(const std::string &symbol) : symbol_(symbol) {};
};

class Quantity {
public:
  double value_;
  Unit unit_;
  
  Quantity(double value, const Unit &unit) : value_(value),
    unit_(unit) {};
  
  Quantity() : value_(0.0), unit_() {};
  
  Quantity(const Quantity &other) : value_(other.value_),
    unit_(other.unit_) {};
  
  Quantity &operator=(const Quantity &other);
  
  bool operator==(const Quantity &right) const {
    return (this->value_ == right.value_)
      and (this->unit_ == right.unit_);
  }
  
  bool operator<(const Quantity &right) const {
    assert (this->unit_ == right.unit_);
    
    return (this->value_ < right.value_);
  }
  
  Quantity &transform(const Quantity &newValue);

};

// For temperature scale and units, see 
// http://physics.nist.gov/cuu/Units/kelvin.html

const Unit DIMENSIONLESS = Unit();
const Unit CELSIUS("Celsius"); 
const Unit KELVIN("Kelvin");
const Unit JOULE("Joule");
const Unit KJ_PER_MOLE("kiloJoulePerMole");
const Unit KJ_PER_MOLE_PER_E2("kiloJoulePerMolePere^2");
const Unit KJ_PER_MOLE_SQ("(kiloJoulePerMole)^2");
const Unit KCAL_PER_MOLE("kiloCalPerMole");
const Unit KCAL_PER_MOLE_SQ("(kiloCalPerMole)^2");
const Unit MOLAR("Molar");
const Unit K_T("kT");
const Unit K_T_PER_E("kTpere");
const Unit K_T_PER_E2("kTpere^2");
const Unit E_CHARGE("e");
const Unit COULOMB("C");
const Unit ANGSTROM("Angstrom");
const Unit METER("meter");
const Unit KCAL_PER_MOLE_KELVIN("kiloCalPerMoleKelvin");
const Unit JOULE_PER_MOLE_KELVIN("JoulePerMoleKelvin");
const Unit JOULE_PER_KELVIN("JoulePerKelvin");
const Unit KJ_PER_MOLE_KELVIN("kiloJoulePerMoleKelvin");
const Unit FARAD_PER_METER("FaradPerMeter");
const Unit METER_PER_FARAD("meterPerFarad");
const Unit MOLE_TO_MINUS_ONE("mol^-1");
// meadPotentialUnit is an energy unit in the following
// units:
// e^2 * k_e /(Angstrom) where k_e is the Coulomb constant.
const Unit MEAD_POTENTIAL_UNIT("e^2*k_eperAngstrom");


const Unit &unit(const std::string &symbol);


namespace nist {
// see http://physics.nist.gov/cuu/Constants/Table/allascii.txt  
  const Quantity MOLAR_GAS_CONSTANT(8.314472, JOULE_PER_MOLE_KELVIN);
  const Quantity ELECTRIC_CONSTANT(8.854187817E-12, FARAD_PER_METER);
  const Quantity AVOGADRO_CONSTANT(6.02214179E+23, MOLE_TO_MINUS_ONE);
  const Quantity BOLTZMANN_CONSTANT(1.3806504E-23,JOULE_PER_KELVIN);
}

namespace conversion {
  // see http://physics.nist.gov/Pubs/SP811/appenB9.html
  // calorie(th)
  const Quantity ONE_CALORIE(4.184, JOULE);
  const Quantity ONE_E_CHARGE(1.602176487E-19,COULOMB);
  const Quantity ONE_ANGSTROM(1.0E-10,METER);

  const Quantity ABSOLUTE_ZERO_IN_CELSIUS(-273.15, CELSIUS);
  const Quantity to_Kelvin_scale(const Quantity &qInCelsius); 
  const Quantity to_Celsius_scale(const Quantity &qInKelvin);
}

const Quantity R_IN_KCAL(nist::MOLAR_GAS_CONSTANT.value_ / (conversion::ONE_CALORIE.value_ * 1000),KCAL_PER_MOLE_KELVIN);
const Quantity R_IN_KJ(nist::MOLAR_GAS_CONSTANT.value_ / 1000,KJ_PER_MOLE_KELVIN);
const Quantity COULOMB_CONSTANT(1.0 / (4.0 * PI * nist::ELECTRIC_CONSTANT.value_),
        METER_PER_FARAD);

namespace conversion {
  const Quantity ONE_MEAD_POTENTIAL_UNIT(COULOMB_CONSTANT.value_ * pow(conversion::ONE_E_CHARGE.value_,2.0)
		  / conversion::ONE_ANGSTROM.value_,JOULE);

}

} // end namespace quantity

namespace std {
  ostream& operator<< (ostream &out, const quantity::Unit &u);

  ostream& operator<< (ostream &out, const quantity::Quantity &q);

}

#endif /*QUANTITY_HPP_*/
