/*! \file azucar/utils/testUtils.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef TESTUTILS_HPP_
#define TESTUTILS_HPP_

#include <string>
#include <vector>

#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/format.hpp>
#include <boost/test/unit_test.hpp>

namespace test {

template <typename T>
bool sharedPtrIsEmpty(const boost::shared_ptr<T> &t) {
  return t.use_count() == 0;
}

template <typename T>
bool sharedPtrIsNull(const boost::shared_ptr<T> &t) {
  return (t.get() == 0) and (t.use_count() == 1);
}

template <typename T>
bool sharedPtrIsEmptyOrNull(const boost::shared_ptr<T> &t) {
  return sharedPtrIsEmpty(t) or sharedPtrIsNull(t);
}

/*! \brief Visitor class that prints qualified test names for all the tests it visits.
 *
 * Names are written to std output.
 *
 * Qualified names include the suite path and the test name. For instance:
 * "suite1::suite2::test1"
 */
class TestSuiteVisitorPrinter : public boost::unit_test::test_tree_visitor
{
public:
    TestSuiteVisitorPrinter();

    virtual void visit( boost::unit_test::test_case const& tu );

    virtual bool    test_suite_start( boost::unit_test::test_suite const& ts);
    virtual void    test_suite_finish( boost::unit_test::test_suite const& );

private:
    // Vector containing the ordered suite path for the current test case (leafs) being visited.
    // Empty when at root of the test tree.
    std::vector<std::string> suitePath_;
};


bool compareTextFiles(const std::string &file1, const std::string &file2,
    const std::vector<std::string> &excluded_line_starts);

bool compareTextFiles(const std::string &file1, const std::string &file2,
    const std::string &excluded_line_start = "");

bool compareTextFilesRegEx(const std::string &file1, const std::string &file2);

/*! \brief Check if 2 files are binary equal.
 *
 * \param file1 Path of one of the files.
 * \param file2 Path of the other file.
 *
 * \return true if file1 and file2 are binary equal; false if they are not.
 *
 * May throw an exception if any of the two files does not exist or cannot be read.
 */
bool compareFiles(const std::string &file1, const std::string &file2);

}
#endif /*TESTUTILS_HPP_*/
