/*! \file azucar/utils/threadMonitor.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef THREADMONITOR_HPP_
#define THREADMONITOR_HPP_

#include <boost/utility.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>

/*! \brief Control life cycle of threads.
 */
class ThreadMonitor : boost::noncopyable {
private:
  boost::condition notifyCondition; //!< Condition variable
    //!< to signal end of wait.
  mutable boost::mutex m_mutex; //!< Mutex to control concurrent
    //!< access to the Monitor.
  int normalThreadsRunning_; //!< normal threads to wait
    //!< until notification is sent.
  bool immediateExit_; //!< signals an immediate exit is needed.    
public:
  /*! \brief Create an empty ThreadMonitor.
   */
  ThreadMonitor() 
    : normalThreadsRunning_(0), 
      immediateExit_(false) {}; 

  /*! \brief Wait for all normal threads to finish as expected (no normal threads
   * in the monitoring list), or for 
   * a key thread to finish or for a normal thread to finish unexpectedly
   * (with an error). Only one thread can call waitNotification.
   * Otherwise, a deadlock can happen.
   */
  void waitNotification();
  
  /*! \brief Remove a thread of type normal from the monitoring
   * list. 
   */
  void notifyNormalThreadEnd();
  
  /*! \brief Add a new thread of type normal to the list of
   * threads controlled by this monitor.
   * 
   * Threads of type key do not need to be added unless they
   * want to be treated as normal threads too (besides
   * behaving as a key thread). 
   */
  void addNormalThread();
  
  /*! \brief Notify the thread in waitNotification() (or calling it
   * later) that a normal thread has finished with errors.
   */
  void notifyErrorThreadEnd();
  
  /*! \brief Notify the thread in waitNotification() (or calling it
   * later) that a key thread has finished.
   */
  void notifyKeyThreadEnd();
  
  ~ThreadMonitor();
}; 

extern ThreadMonitor globalThreadMonitor; //!< Global thread monitor
  //!<to be used by all threads.

#endif /*THREADMONITOR_HPP_*/
