/*! \file azucar/utils/util.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef UTIL_HPP_
#define UTIL_HPP_
#include <istream>
#include <functional>
#include <set>
#include <fstream>
#include <string>
// for random numbers
#include <cstdlib>
#include <sstream>

#include <boost/shared_ptr.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/tr1/random.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/program_options.hpp>
#include <boost/function.hpp>
#include <boost/numeric/conversion/converter.hpp>
#include <boost/multi_array.hpp>
#include <boost/shared_array.hpp>

#include <loki/ScopeGuard.h>
#include <loki/LokiTypeInfo.h>

#include "azucar/utils/exceptions.hpp"

namespace po = boost::program_options;


namespace azucar {
namespace utils {


/*! \brief Check if an input file stream has reached EOF.
 * 
 * Notice that istream::eof does not actually notify about
 * reaching EOF, but about having actually read it.
 * 
 * \return true if EOF has been reached. false otherwise.
 */
bool eof(std::istream &input);

template <class T>
const std::string toString(const T &value) {
  std::ostringstream outString;
  outString << value;
  outString.flush();
  return outString.str();
}

template<class Target>
Target trimConvert(const std::string &source) {
  std::string trimmedVersion = boost::algorithm::trim_copy(source);
  return boost::lexical_cast<Target>(trimmedVersion);
}

template<class Target>
Target trimConvertSubstr(const std::string &source, std::string::size_type off = 0,
            std::string::size_type count = std::string::npos) {
  std::string trimmedVersion = boost::algorithm::trim_copy(source.substr(off, count));
  return boost::lexical_cast<Target>(trimmedVersion);
} 

const std::string 
trimSubstr(const std::string &source, std::string::size_type off = 0,
          std::string::size_type count = std::string::npos);

class InputFile {
public:
  std::ifstream stream;  
  InputFile(const boost::filesystem::path &filePath);
  InputFile(const std::string &fileName);
  InputFile(const char *filename);
  void skipLine();
  void skipLines(unsigned int n);
  bool nextLineIsEmpty(std::string &nextLine);
  void close();
  ~InputFile();
};

class OutputFile {
public:
  std::ofstream stream;
  OutputFile(const boost::filesystem::path &filePath,
    bool append=false);
  OutputFile(const std::string &fileName, bool append=false);
  OutputFile(const char *filename, bool append=false);
  void close();
  ~OutputFile();
};


/*! \brief Handy function to be used when binding a function
 *  is needed, and the function uses the operator delete.
 * 
 * Since delete is an operator and not a function, it must be
 * wrapped inside this function to be used as a function argument.
 * 
 * 
 * \param T Type of the object to delete.
 * \param t Pointer to an object of type T on which execute
 *  delete t.
 */
template<class T> void deleteFunction(T *t) {
  delete t;
}

/*! \brief Put current thread to sleep.
 * 
 * \param seconds Approximate time to sleep, in seconds.
 */
void threadSleep(unsigned int seconds);

template<class R, class T> class mem_fun_smart_t 
    : public std::unary_function<boost::shared_ptr<T>, R> {
  R(T::*pmf) ();
public:
  explicit mem_fun_smart_t(R(T::*p)()) : pmf(p) {}
  R operator() (boost::shared_ptr<T> &p) { return ((p.get())->*pmf)();}

};

template<class R, class T> mem_fun_smart_t<R,T> mem_fun_smart(R(T::*f)()) {
  return mem_fun_smart_t<R,T>(f);
}

template<class R, class T, class A> class mem_fun1_smart_t 
    : public std::binary_function<boost::shared_ptr<T>, A, R> {
  R(T::*pmf) (A);
public:
  explicit mem_fun1_smart_t(R(T::*p)(A)) : pmf(p) {}
  R operator() (boost::shared_ptr<T> &p, A &param1) { 
      return ((p.get())->*pmf)(param1);}

};

template<class R, class T, class A> 
mem_fun1_smart_t<R,T,A> 
mem_fun_smart(R(T::*f)(A)) {
  return mem_fun1_smart_t<R,T,A>(f);
}

template<class In, class Op>Op for_each_key(In firstIt, In lastIt, Op f) {
  while (firstIt != lastIt) {
    f(firstIt->first);
    ++firstIt;
  }
  return f;
}


template<class In, class Op>Op for_each_value(In firstIt, In lastIt, Op f) {
  while (firstIt != lastIt) {
    f(firstIt->second);
    ++firstIt;
  }
  return f;
}

template<class In, class Op>In find_if_value(In firstIt, In lastIt, Op f) {
  while (firstIt != lastIt) {
    if (f(firstIt->second)) {
      return firstIt;
    }
    ++firstIt;
  }
  return lastIt;
}


bool isInASCIIPrintableRange(const std::string &s);
bool isInASCIIPrintableRange(char c);

const std::string htmlBrk(const std::string &str);
const std::string htmlEscape(const std::string &input);

bool hasEmailAddressFormat(const std::string &address);

///////////////// NUMERIC CONVERSIONS ////////////////////
typedef boost::numeric::converter<double,int> Int2Double;

///////////////// ARRAYS /////////////////////////////////
typedef boost::multi_array<double, 1> Array1D;
typedef boost::multi_array<double, 2> Array2D;
typedef boost::multi_array<double, 3> Array3D;


//////////////// RANDOM NUMBERS //////////////////////////
/*!
 * \deprecated Use one of ::createUniformIntRandomGenerator(),
 *   ::createUniformLongRandomGenerator(), or ::createUniformRealRandomGenerator()
 *   instead.
 *
 * \brief Class for generating random numbers.
 *
 *
 */
class Randuni {
private:
  double max; // limit for integer or real range where the 
  // random number is generated.
public:
  /**
   * Create a uniform random number generator (both integer and real).
   * Range: [0,maxValue) for real numbers, [0,maxValue] for integers.
   */
  Randuni(double maxValue = 1.0):max(maxValue) {}
  static void seed(long s) {srand(s);}

  /** 
   * Generate random numbers uniformely between [0,max)
   */
  double draw() { return ((double) (rand()) /RAND_MAX) * max;}
  
  /**
   * Generate integer random number uniformely between [0,max]
   */
  int drawInt() { return (int) (( (double) rand() / RAND_MAX) * (max+1));}
};

typedef boost::mt19937 GlobalRandomGeneratorType; /*!< \brief Random Engine Type
	used for the ::globalRandomGenerator */

extern GlobalRandomGeneratorType globalRandomGenerator; /*!< \brief Random Engine object
	used globally by the %azucar.utils random functions.
	It is initialized, always with the same default value (depending on the actual
	type of ::GlobalRandomGeneratorType), so if you intend to change this behaviour,
	make sure to change its seed. */


typedef std::tr1::variate_generator<GlobalRandomGeneratorType&,
              std::tr1::uniform_int<int> > UniformIntRandomGenerator; /*!< \brief Random number
              generation type for integer numbers in a given closed range.
              Engine used is shared with external usage, because it is passed
              as a reference. */

typedef std::tr1::variate_generator<GlobalRandomGeneratorType&,
              std::tr1::uniform_int<long> > UniformLongRandomGenerator; /*!< \brief Random number
              generation type for long numbers in a given closed range.
              Engine used is shared with external usage, because it is passed as a reference. */

typedef std::tr1::variate_generator<GlobalRandomGeneratorType&,
              std::tr1::uniform_real<double> > UniformRealRandomGenerator; /*!< \brief Random number
              generation type for double numbers in a given range [min,max).
              Engine used is shared with external usage, because it is passed as a reference. */
              

/*!
 * \brief Create a uniform int random number generator using globalRandomGenerator as engine.
 *
 * \param min Minimum integer number of the available range to generate. Included in the range.
 * \param max Maximum integer number of the available range to generate. Included in the range.
 */
UniformIntRandomGenerator createUniformIntRandomGenerator(int min, int max);

/*!
 * \brief Create a uniform long random number generator using globalRandomGenerator as engine.
 *
 * \param min Minimum long number of the available range to generate. Included in the range.
 * \param max Maximum long number of the available range to generate. Included in the range.
 */
UniformLongRandomGenerator createUniformLongRandomGenerator(long min, long max);

/*!
 * \brief Create a uniform double random number generator using globalRandomGenerator as engine.
 *
 * \param min Minimum number of the available range to generate. Included in the range.
 * \param max Maximum number of the available range to generate. Not included in the range.
 */
UniformRealRandomGenerator createUniformRealRandomGenerator(double min, double max);              


const std::string createTmpDirectory(const std::string &nameTemplate);

class TempFile {
private:
  boost::filesystem::path path_;
public:
  TempFile(const std::string &nameTemplate);
  const boost::filesystem::path &getPath() const;
  void getInputStream(std::ifstream &in);
  void getOutputStream(std::ofstream &out);
  ~TempFile();
};

/** \brief Obtain the relative path of p with regard to base.
 *
 * https://svn.boost.org/trac/boost/ticket/1976#comment:2
 * http://stackoverflow.com/questions/5772992/get-relative-path-from-two-absolute-paths
 *
 * Example: relativePath(/foo/new, /foo/bar) => ../new
 *
 * ALGORITHM:
 *  iterate path and base
 * compare all elements so far of path and base
 * whilst they are the same, no write to output
 * when they change, or one runs out:
 *   write to output, ../ times the number of remaining elements in base
 *   write to output, the remaining elements in path
 */
const boost::filesystem::path
relativePath(const boost::filesystem::path p, const boost::filesystem::path base);

/** \brief Obtain a normalized path for p (without ./ nor ../ elements).
 *
 * \param p A non-empty path to be normalized.
 *
 * \return A normalized version of p (without ./ nor ../ elements).
 */
const boost::filesystem::path
normalizePath(const boost::filesystem::path p);


///////////////// OBTAIN PROGRAM PATH ////////////////////
const boost::filesystem::path getSelfPath(); 

class ManagedArgv {
private:
  int size_;
  boost::shared_array<char *> args_;
  int currentIndex_;
public:
  ManagedArgv(int size);  

  void append(const std::string &newArg);  

  char **getData();
  
  ~ManagedArgv();
};



//////////////// OPTION PARSING OPERATIONS ///////////////

/*! Given an assigned map of program options variable, this type of
 * functions checks it is a valid set. If so, it returns true. If not,
 * it returns false and it may print some error or warning message to
 * stderr. */
typedef boost::function<bool (po::variables_map&)> CheckOptionsFunc;

/*! This type of functions print a usage message for a program
 * to stderr.
 */
typedef boost::function<void (void)> UsageFunc;
 

enum ProgramOptionsResult {NORMAL_EXECUTION, INFO_EXECUTION, ERROR_EXECUTION};

/*! \brief Read command line parameters and store them in vm.
 * 
 * \param[out] vm Map of processed command line parameters.
 * \param checkF Function to check validity of vm once assigned.
 * \param usageF Function to print a usage message if needed.
 * \param argc Number of command line parameters.
 * \param argv Command line parameters.
 * \param specificProgramOptions Description of the options permitted
 *   in command line. There are no positional parameters. Parameters
 *   --help and --version are automatically added by this function.
 * 
 * \return True if the command line contained a valid set of parameters
 *   (stored at vm). False otherwise (it prints error messages in stderr). 
 */
ProgramOptionsResult
processProgramOptions(po::variables_map &vm,
  const CheckOptionsFunc &checkF,
  const UsageFunc &usageF,
  const std::string &version_string,
  int argc, char **argv,
  const po::options_description &specificProgramOptions);

/*! \brief Read command line parameters and store them in vm.
 * 
 * \param[out] vm Map of processed command line parameters.
 * \param checkF Function to check validity of vm once assigned.
 * \param usageF Function to print a usage message if needed.
 * \param argc Number of command line parameters.
 * \param argv Command line parameters.
 * \param specificProgramOptions Description of the options permitted
 *   in command line. Those that are positional parameters are
 *   included i npositionalOptions. Parameters
 *   --help and --version are automatically added by this function.
 * \param positionalOptions Set of the positional parameters from
 *   specificProgramOptions.
 * 
 * \return True if the command line contained a valid set of parameters
 *   (stored at vm). False otherwise (it prints error messages in stderr). 
 */
ProgramOptionsResult 
processProgramOptions(po::variables_map &vm,
  const CheckOptionsFunc &checkF,
  const UsageFunc &usageF,
  const std::string &version_string,
  int argc, char **argv,
  const po::options_description &namedProgramOptions,
  const po::positional_options_description &positionalOptions);

//////////////// SET OPERATIONS //////////////////////////

//! A = A U B

template<class Key> void joinSets(std::set<Key> &a, const std::set<Key> b) {
  using namespace std;
  typename set<Key>::const_iterator it;
  for(it = b.begin(); it != b.end(); ++it) {
    a.insert(*it);
  }
}

//! A = A - B
template<class Key> void removeSet(std::set<Key> &a, const std::set<Key> b) {
  using namespace std;
  typename set<Key>::const_iterator it;
  for(it = b.begin(); it != b.end(); ++it) {
    typename set<Key>::iterator foundPos;
    foundPos = a.find(*it);
    if (foundPos != a.end()) {
      a.erase(foundPos);
    }
  }
}

#define ON_BLOCK_EXIT      ::Loki::ScopeGuard __attribute__ ((unused)) LOKI_ANONYMOUS_VARIABLE(scopeGuard) = ::Loki::MakeGuard
#define ON_BLOCK_EXIT_OBJ  ::Loki::ScopeGuard __attribute__ ((unused)) LOKI_ANONYMOUS_VARIABLE(scopeGuard) = ::Loki::MakeObjGuard

using ::Loki::TypeInfo;


}
}

#endif /*UTIL_HPP_*/
