/*! \file azucar/utils/xerces.hpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#ifndef XERCES_HPP_
#define XERCES_HPP_

#include <string>
#include <istream>

#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/BinInputStream.hpp>
#include <xercesc/sax/InputSource.hpp>

namespace azucar {
namespace utils {

/*! \brief Control initialization and termination of Xerces-C++ library.
 *
 * Expected usage, per process, is:
 * 1. Create an object of this class, with automatic destruction.
 * boost::scoped_ptr<XercesLibrary> xerces(new XercesLibrary);
 * 2. Init the library, before doing any operation with Xerces:
 * xerces->init();
 * 3. After working with Xerces, and before exiting the program, terminate the
 * library:
 * xerces->terminate();
 *
 * In any case, once the program exits, for whatever reason, the object will be
 * destroyed and, if it had not been terminated, a call to terminate will be done.
 */
class XercesLibrary {
public:
	bool initialized;
	/*! \brief Create an object to control the Xerces-C++ library.
	 *
	 */
	XercesLibrary();

	/*! \brief Initialize the Xerces-C++ library.
	 *
	 * \pre No previous initialization of Xerces-C++ library has been done.
	 * \post Xerces-C++ library is ready for use.
	 *
	 * \exception XMLException (xerces)
	 */
	void init();

	/*! \brief Terminate the Xerces-C++ library.
	 *
	 * \pre init() has been previousely called, and no previous call to terminate()
	 * has been done.
	 * \post Xerces-C++ library has been orderly terminated, and no further use
	 * of the library is allowed.
	 */
	void terminate();
	/*! \brief Destroy this object, and terminate the Xerces-C++ library if it had
	 * been initialized and not yet terminated.
	 *
	 */
	~XercesLibrary();
private:
	// forbid copy constructor and assignment
	XercesLibrary(const XercesLibrary&);
	XercesLibrary& operator=(const XercesLibrary&);
};

const std::string XMLChToString(const XMLCh *const sourceStr);

// Implementation from std_input_stream class from Code Synthesis XSD 3.3.0
// http://www.codesynthesis.com/products/xsd/
class StdIstreamInputStream : public XERCES_CPP_NAMESPACE_QUALIFIER BinInputStream {
public:
	StdIstreamInputStream(std::istream &is);
	virtual unsigned int curPos() const;
	virtual unsigned int readBytes(XMLByte* const toFill, const unsigned int maxToRead);
#if _XERCES_VERSION >= 30000
	virtual const XMLCh* getContentType () const;
#endif
	virtual ~StdIstreamInputStream();
private:
  std::istream& is_;
};

// After std_input_source class from Code Synthesis XSD 3.3.0
// http://www.codesynthesis.com/products/xsd/
class StdIstreamInputSource : public XERCES_CPP_NAMESPACE_QUALIFIER InputSource {
public:
	StdIstreamInputSource(std::istream &is);
	virtual ~StdIstreamInputSource();
	virtual XERCES_CPP_NAMESPACE_QUALIFIER BinInputStream* makeStream() const;
private:
  mutable std::istream* is_;
};

}
}


#endif /* XERCES_HPP_ */
