/*! \file apps/CGILint.cpp
 *
 * \author Jorge Estrada
 * 
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 *
 * \ingroup Applications
 * 
 * \brief CGILint is designed to test CGI applications.
 * 
 * It is based on Perl CGI_Lint application 
 * (http://www.oreilly.com/openbook/cgi/ch12_05.html) described on the O'Reilly
 * book "CGI Programming on the World Wide Web" 1st edition, 1996.
 * 
 * Usage:
 *  
 * <tt>CGILint.exe CGILint.cfg CGILint.dat</tt>\n
 * For executing, it requires two arguments: an environment variable file 
 * (CGILint.cfg) and a data file (CGILint.dat).\n 
 * The environment variable file contains all the environment 
 * variables that will be passed to the CGI application. It contains one variable
 * per line in the  format VARIABLE=data\n
 * The data file will contain all the POST data passed to the CGI application.
 * The format is dataVariable=data\n
 * When passing files, the format is fileVariable=*filename
 * so that filename will be read by CGILint and passed to the CGI application.
 * 
 * CGILint.cfg must include at least the following variables:\n
 * REQUEST_METHOD=post\n
 * SERVER_ROOT=.\n
 * SCRIPT_NAME=my/path/to/script.exe\n
 * 
 * At this time, REQUEST_METHOD can only be 'post'. SERVER_ROOT and SCRIPT_NAME
 * can take any sensible value.
 * 
 * The variables CONTENT_LENGTH and CONTENT_TYPE are set by CGILint, so
 * values set in CGILint.cfg will be overwritten.
 * 
 * Sample files are as follows:
 * <ul> <li> CGILint.cfg
 * \verbatim
REQUEST_METHOD = post
SERVER_ROOT = .
SCRIPT_NAME = ../source/serviciowebpau.cgi
LD_LIBRARY_PATH = /usr/local/cgicc/lib 
\endverbatim
 * <li> CGILint.dat
 * \verbatim
solventRadius = 1.4
userfile = *1FTG.pdb
\endverbatim
 * </ul>
 */

#include <string>
#include <fstream>
#include <iostream>
#include <map>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/assign/std/set.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/foreach.hpp>
#include <boost/bind.hpp>

#include "azucar/utils/exceptions.hpp"
#include "azucar/utils/process.hpp"
#include "azucar/utils/util.hpp"
#include "azucar/utils/version.hpp"

using namespace std;

// the following is needed to convert the API_X arguments into a string
// see the VERSION_STRING define below.
// For documentation about the C preprocessor language, see
// http://gcc.gnu.org/onlinedocs/cpp/Stringification.html#Stringification
#define XSTRINGIFY(s) STRINGIFY(s)
#define STRINGIFY(s) #s

#define VERSION_STRING XSTRINGIFY(API_MAJOR) "." XSTRINGIFY(API_MINOR) "." XSTRINGIFY(API_PATCH)

const std::string version_string = VERSION_STRING;

//! Temporary file to write all the MIME data to pass to the CGI application
//! It is not deleted after execution.
const string mime_file_name = "tempMIMEinput.txt";
//! Temporary file to write the output (standard and error) of the CGI
//! application.
const string output_cgi_file = "tempCGIoutput.txt";
//! String code to separate mime sections. It is assumed to not be contained
//! on the data passed to the CGI application.
const string mime_separator = "AaB03x";

/*!
 * Class to contain all important data for the CGI application.
 * It includes the MIME file name, the MIME separator, and the
 * environment variables.
 */
class CGIData {
public:
  string mimeFilename;
  string separator;
  map<string, string> env;
};

//data is input output
//! Read an input file with environment variables' definitions, and
//! insert them into data.
/*!
 * The input file can contain blank lines and blanks around the equal sign
 * or before and after the variable name or value are ignored.
 * \param inputEnvFilename The name of the file containing the 
 * environment variables' definitions.
 * \param[in,out] data Object whose env member will be updated, adding
 * new environment variables found on inputEnvFilename, or overwriting
 * existing environment variables, also found on inputEnvFilename.
 * \exception azucar::utils::BaseException In case of wrong format in input data file.
 */
void processEnvironment(const string &inputEnvFilename, CGIData &data) {
  ifstream inputEnvFile(inputEnvFilename.c_str());

  while(not inputEnvFile.eof()) {
    string line;
    getline(inputEnvFile, line);
    boost::algorithm::trim(line);

    if (line.empty()) {
      continue;
    }

    string::size_type breakPosition = line.find("=");

    if (breakPosition == string::npos) {
      throw azucar::utils::BaseException("Wrong format in input data file");
    }


    string name = line.substr(0, breakPosition) ;
    string value = line.substr(breakPosition+1, string::npos);

    boost::algorithm::trim(name);
    boost::algorithm::trim(value);

    data.env[name] = value;
  }

  inputEnvFile.close();
}

// data is input output
//! Read the CGI data file and convert to a valid standard input format for a CGI
//! application.
/*!
 * The input file can contain blank lines and blanks around the equal sign
 * or before and after the variable name or value are ignored.
 * \param inputDataFilename name of the CGI data file to read, containing
 * the CGI variables to pass to the CGI application.
 * \param[in,out] data Object containing information on the name of the
 * file to use as input for the CGI application. That file will be created with
 * the CGI variables to pass to the CGI application.
 * \exception azucar::utils::BaseException if the input data file has the wrong format.
 */
boost::intmax_t processData(const string &inputDataFilename, CGIData &data) {
  ifstream inputDataFile(inputDataFilename.c_str());
  ofstream outputMIMEFile(data.mimeFilename.c_str());

// The following commented text is not needed, since it is provided
// through the environment variables.
// Besides, the CGI 1.1 RFC 3875 states that :
// "As transfer-codings are not supported on the request-body, the server
//    MUST remove any such codings from the message-body, and recalculate
//    the CONTENT_LENGTH."

//   outputMIMEFile << "Content-Type: multipart/form-data; boundary="
// 		 << data.separator << "\r\n" << "\r\n";

  while(not inputDataFile.eof()) {
    string line;
    getline(inputDataFile, line);
    boost::algorithm::trim(line);
    if (line.empty()) {
      continue;
    }

    string::size_type breakPosition = line.find("=");

    if (breakPosition == string::npos) {
      throw azucar::utils::BaseException("Wrong format in input data file");
    }


    string name = line.substr(0, breakPosition) ;
    string value = line.substr(breakPosition+1, string::npos);

    boost::algorithm::trim(name);
    boost::algorithm::trim(value);

    if (not boost::algorithm::starts_with(value, "*")) {
      // it is a normal name-value pair
      outputMIMEFile << "--" << data.separator << "\r\n";
      outputMIMEFile << "Content-Disposition: form-data; name=\""
		     << name << "\"\r\n\r\n";
      outputMIMEFile << value;
      outputMIMEFile << "\r\n";
    } else {
      // it is a file
      outputMIMEFile << "--" << data.separator << "\r\n";
      outputMIMEFile << "Content-Disposition: form-data; name=\""
		     << name << "\"; filename=\""
		     << value.substr(1) << "\"\r\n\r\n";
      
      ifstream dataFile(value.substr(1).c_str());
      if (dataFile.fail()) {
        throw azucar::utils::BaseException("Could not open file in form file field.");
      }
      // The following comes from chapter 27 of GNU libstdc++ manual.
      // Section: "Copying a file"
      outputMIMEFile << dataFile.rdbuf();
      dataFile.close();
      outputMIMEFile << "\r\n";
    }
    
  }

  outputMIMEFile << "--" << data.separator << "--" << "\r\n";
  inputDataFile.close();
  outputMIMEFile.close();
  return 
    boost::filesystem::file_size(boost::filesystem::path(data.mimeFilename));
}

enum ProgramOption {INPUT_ENV_FILE,
  INPUT_DATA_FILE,};

std::map<ProgramOption, const char *> program_option_names
  = boost::assign::map_list_of(INPUT_ENV_FILE, "input-environment-file")
      (INPUT_DATA_FILE, "input-data-file");

std::vector<ProgramOption> requiredOptions
    = boost::assign::list_of(INPUT_ENV_FILE)
      (INPUT_DATA_FILE);


bool checkOptions(po::variables_map &vm) {

  BOOST_FOREACH(ProgramOption option, requiredOptions) {
    if (vm[program_option_names[option]].empty()) {
      cerr << "Missing '" << program_option_names[option]
        << "' option." << endl;
      return false;
    }
  }

  return true;
}

void usage(const boost::filesystem::path &programPath) {
  cerr << "Usage: " << programPath.filename() << " [options] "
    << program_option_names[INPUT_ENV_FILE]
    << " " << program_option_names[INPUT_DATA_FILE]
    << endl;

  return;
}


int main(int argc, char **argv) {
  po::options_description desc("Specific options");
  desc.add_options()
    (program_option_names[INPUT_ENV_FILE], po::value<std::string>(),
      "Input file containing the environment variables")
    (program_option_names[INPUT_DATA_FILE], po::value<std::string>(),
      "Input file containing the data")
    ;

  po::positional_options_description posDesc;
  posDesc.add(program_option_names[INPUT_ENV_FILE], 1);
  posDesc.add(program_option_names[INPUT_DATA_FILE], 1);

  po::variables_map vm;


  azucar::utils::ProgramOptionsResult optionsResult
    = azucar::utils::processProgramOptions(vm, checkOptions,
                                boost::bind(usage, argv[0]),
                                version_string,
                                argc, argv, desc, posDesc);

  if (optionsResult == azucar::utils::ERROR_EXECUTION) {
    return 1;
  } else if (optionsResult == azucar::utils::INFO_EXECUTION) {
    return 0;
  }

  try {
    string inputEnvFilename = vm[program_option_names[INPUT_ENV_FILE]].as<string>();
    string inputDataFilename = vm[program_option_names[INPUT_DATA_FILE]].as<string>();
    string scriptName;
    CGIData data;
    data.mimeFilename = mime_file_name;
    data.separator = mime_separator;
    processEnvironment(inputEnvFilename, data);
    boost::intmax_t dataSize = processData(inputDataFilename, data);
    data.env[string("CONTENT_LENGTH")] = boost::lexical_cast<string>(dataSize);
    data.env[string("CONTENT_TYPE")] = string("multipart/form-data; boundary=")
      + data.separator;

    // read SCRIPT_NAME
    // make sure REQUEST_METHOD exists and it is POST
    if (data.env.find(string("REQUEST_METHOD")) != data.env.end()) {
      if (data.env[string("REQUEST_METHOD")] != string("post")) {
	throw azucar::utils::BaseException("Only POST method is allowed at the moment");
      } 
    } else {
	throw azucar::utils::BaseException("No REQUEST_METHOD set in config file");
    }

    if (data.env.find(string("SCRIPT_NAME")) != data.env.end()) {
      scriptName = data.env[string("SCRIPT_NAME")];
    } else {
      throw azucar::utils::BaseException("No SCRIPT_NAME set in config file");
    }
     
    // use exec
    vector<string> argVector; // empty, except for the scriptName
    argVector.push_back(scriptName); // make sure to pass only the ending
    map<string,string> &envVector = data.env;
    string inputFilePath = data.mimeFilename;
    string outputAndErrorFilePath = output_cgi_file;
    boost::filesystem::path scriptPath = 
      boost::filesystem::system_complete(boost::filesystem::path(scriptName));


    int returnStatus = processExecute(scriptPath.string(),
				      argVector, envVector, inputFilePath,
				      outputAndErrorFilePath);

    if (returnStatus == -1) {
      throw azucar::utils::BaseException("Error executing CGI script");
    }
    cout << "CGI execution finished with status: " << returnStatus << endl;
    cout << "Intermediate MIME input file: " << inputFilePath << endl;
    cout << "CGI output and error: " << outputAndErrorFilePath << endl;
  } catch (azucar::utils::BaseException &ex) {
    cerr << "Exception caught: " << ex.getMessage() << endl;
    exit(1);
  } catch (...) {
    cerr << "Exception caught." << endl;
    exit(1);
  }
  return 0;
}
