/*! \file azucar/utils/configurationReader.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <iostream>
#include <cassert>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/filesystem/exception.hpp>


#include <log4cpp/Category.hh>

#include <loki/ScopeGuard.h>


#include "azucar/utils/exceptions.hpp"
#include "azucar/utils/configurationReader.hpp"

XERCES_CPP_NAMESPACE_USE

using namespace std;

namespace {
  const std::string category_name = "configurationReader";
  
  log4cpp::Category &logger() {
    static log4cpp::Category &theLogger =   
      log4cpp::Category::getInstance(category_name);
    return theLogger;
  }
}

namespace configuration {
  class MyErrorHandler : public DOMErrorHandler {
  public:
      bool handleError(const DOMError&);
  };
  
  bool MyErrorHandler::handleError(const DOMError& exception)
  {
      char* message = XMLString::transcode(exception.getMessage());
      logger().errorStream() << "Fatal Error: " << message
           << " at line: " << exception.getLocation()->getLineNumber()
           << log4cpp::CategoryStream::EOL;
      XMLString::release(&message);
  
      return true;
  }
  
  void ConfigurationReader::init(
    const boost::filesystem::path &configFile) {
    assert(not this->isActive());
    XMLCh *tempStr = XMLString::transcode("LS");
    DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(tempStr);
    XMLString::release(&tempStr);
    parser_ = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
    started_ = true;
          
    parser_->setFeature(XMLUni::fgDOMValidation, false);
    parser_->setFeature(XMLUni::fgDOMNamespaces, true);
    parser_->setFeature(XMLUni::fgDOMDatatypeNormalization, true);
    parser_->setFeature(XMLUni::fgDOMEntities, false);
    parser_->setFeature(XMLUni::fgDOMComments, false);
          
    errorHandler_ = new MyErrorHandler();
    parser_->setErrorHandler(errorHandler_);
  
    Loki::ScopeGuard parserGuard 
      = Loki::MakeGuard(&ConfigurationReader::release, this);
  
  
    try {
      doc_ = parser_->parseURI(configFile.string().c_str());
    }
    catch (const XMLException& toCatch) {
      char* message = XMLString::transcode(toCatch.getMessage());
      logger().infoStream() << "Exception message is: \n"
           << message << log4cpp::CategoryStream::EOL;;
      XMLString::release(&message);
      throw azucar::utils::BaseException("Could not correctly parse configuration file.");
    }
    catch (const DOMException& toCatch) {
      char* message = XMLString::transcode(toCatch.msg);
      logger().infoStream() << "Exception message is: \n"
           << message << log4cpp::CategoryStream::EOL;;
      XMLString::release(&message);
      throw azucar::utils::BaseException("Could not correctly parse configuration file.");
    }
    catch (...) {
      logger().info("Unexpected Exception");
      throw azucar::utils::BaseException("Could not correctly parse configuration file.");
    }
    
    parserGuard.Dismiss();
    return;
  }
  
  bool ConfigurationReader::isActive() const {
    return started_;
  }
  
  void ConfigurationReader::release() {
    if (isActive()) {
      parser_->release();
      delete errorHandler_;
    }
    started_ = false;
  }
  
  bool ConfigurationReader::existsTextOption(const std::string &name) const {
    try {
      XMLCh *xmlTag = XMLString::transcode(name.c_str());
      DOMNodeList *elements = doc_->getElementsByTagName(xmlTag);
      XMLString::release(&xmlTag);
      if (elements->getLength() != 1) {
        return false;
      }
      DOMNode *targetNode = elements->item(0);
      if (targetNode->getChildNodes()->getLength() != 1) {
        return false;
      }
    
      DOMNode *textNode = targetNode->getFirstChild();
      if (textNode->getNodeType() != DOMNode::TEXT_NODE) {
        return false;
      }
      
      return true;
    } catch (const XMLException& toCatch) {
      char* message = XMLString::transcode(toCatch.getMessage());
      cout << "Exception message is: \n"
           << message << "\n";
      XMLString::release(&message);
    }
    catch (const DOMException& toCatch) {
      char* message = XMLString::transcode(toCatch.msg);
      cout << "Exception message is: \n"
           << message << "\n";
      XMLString::release(&message);
    }
    catch (...) {
      cout << "Unexpected Exception \n" ;
    }
    
    throw azucar::utils::BaseException("ConfigurationReader::getTextOption exception");
  }
  
  const std::string ConfigurationReader::getTextOption(const std::string &name) const {
    try {
      XMLCh *xmlTag = XMLString::transcode(name.c_str());
      DOMNodeList *elements = doc_->getElementsByTagName(xmlTag);
      XMLString::release(&xmlTag);
      DOMNode *targetNode = elements->item(0);
      DOMNode *textNode = targetNode->getFirstChild();
      const XMLCh *textValue = textNode->getNodeValue();
      char *charTextValue = XMLString::transcode(textValue);
      string result(charTextValue);
      boost::algorithm::trim(result);
      XMLString::release(&charTextValue);
      return result;
    } catch (const XMLException& toCatch) {
      char* message = XMLString::transcode(toCatch.getMessage());
      cout << "Exception message is: \n"
           << message << "\n";
      XMLString::release(&message);
    }
    catch (const DOMException& toCatch) {
      char* message = XMLString::transcode(toCatch.msg);
      cout << "Exception message is: \n"
           << message << "\n";
      XMLString::release(&message);
    }
    catch (...) {
      cout << "Unexpected Exception \n" ;
    }
    
    throw azucar::utils::BaseException("ConfigurationReader::getTextOption exception");
  }
  
  ConfigurationReader::~ConfigurationReader() {
    this->release();
  }  
  
  bool checkAndLogPathIsDir(const boost::filesystem::path &p,
      const string &errStr) {
    if (not boost::filesystem::is_directory(p)) {
      logger().error(errStr);
      return false;
    } else {
      return true;
    }
  }
  
  bool checkAndLogPathIsRegular(const boost::filesystem::path &p,
      const string &errStr) {
//    if (not boost::filesystem::is_regular(p)) {
    if (not boost::filesystem::exists(p)
        or boost::filesystem::is_directory(p)) {
      logger().error(errStr);
      return false;
    } else {
      return true;
    }       
  }
  
  bool checkAndLogPathExists(const boost::filesystem::path &p,
      const string &errStr) {
    if (not boost::filesystem::exists(p)) {
      logger().error(errStr);
      return false;
    } else {
      return true;
    }        
  }
  
  bool checkAndLogPathIsExe(const boost::filesystem::path &p,
      const string &errStr) {
    return checkAndLogPathIsRegular(p, errStr);
  }
  
  
  bool checkAndLogLeaf(const string &str, const string &errStr) {
    try {
      boost::filesystem::path theLeaf(str);
      return true;
    } catch(boost::filesystem::filesystem_error &ex) {
      logger().error(errStr);
      return false;
    }
  }

  bool checkAndLogPrefix(const string &str, const string &errStr) {
    return checkAndLogLeaf(str, errStr);
  }
  
  bool checkAndLogSuffix(const string &str, const string &errStr) {
    return checkAndLogLeaf(str, errStr);
  }
  
  bool checkAndLogIsExt(const string &ext, const string &errStr) {
    try {
      boost::filesystem::path theExt(string("X") + ext);
      if (boost::filesystem::extension(theExt) == ext) {
        return true;
      } else {
        logger().error(errStr);
        return false;
      }
    } catch (boost::filesystem::filesystem_error &) {
      logger().error(errStr);
      return false;
    }
  }
  
  bool checkAndLogStrIsRegular(const string &filename,
      const string &errStr) {
    try {
      boost::filesystem::path thePath(filename);
      return checkAndLogPathIsRegular(thePath, errStr);
    } catch (boost::filesystem::filesystem_error &) {
      logger().error(errStr);
      return false;          
    }
  }

} // end namespace
