/*! \file azucar/utils/database/framework.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <boost/thread/mutex.hpp>

#include <log4cpp/Category.hh>

#include <mysql++/mysql++.h>


#include "azucar/utils/exceptions.hpp"
#include "azucar/utils/database/framework.hpp"
#include "azucar/utils/database/mysqlppHelper.hpp"

using namespace std;

namespace {
  const std::string category_name = "dbframework_cat";
  
  log4cpp::Category &logger() {
    static log4cpp::Category &theLogger =   
      log4cpp::Category::getInstance(category_name);
    return theLogger;
  }


}

namespace azucar {
namespace utils {
namespace database {
  /*! \brief Escape string for SQL command. 
   * 
   * Quotes must be added later to the string in the SQL command.
   */
  const std::string escapeString(const std::string &str) {
    string escapedStr = str;
    mysqlpp_escape_string_correction(escapedStr);
    return escapedStr;
  }
  
  DBException::DBException(const std::string &msg) 
        : azucar::utils::BaseException(msg) {};
    
  DBException::~DBException()  throw() {};

  // private namespace
  namespace {
    class DatabaseImpl : public Database {
    private:
      std::string dbName_; //!< Name of current database.
      std::string host_; //!< Host where current database is.
      std::string user_; //!< User for current database.
      std::string passwd_; //!< Password for user_ of current database.
      int port_; //!< TCP port where database manager of current database listens.
    public:
      boost::mutex db_mutex; //!< Mutex used for avoiding concurrent wrong
      //!< usage of current database connection.
      
      mysqlpp::Connection con_; //!< Connection for operating on current database.
      
    private:
      bool disconnected_;

    public:
      /*! \brief Check status of database connection and, if it is down,
       * try to reconnect.
       * 
       * \exception If reconnection fails.
       */
      void pingAndReconnect();
      /*! \brief Do a query to the database that can return row results.
       * 
       * \param query SQL query to execute.
       * 
       * \return A vector with the rows in the order returned. 
       *  Each row is returned as a map indexed by field name
       *  and with the value of the field as a string.
       * 
       * \exception DBException If query fails.
       */
//      friend const std::vector<std::map<std::string, std::string> > 
//        doRowQuery(boost::shared_ptr<Database> database, const std::string &query);
//
      /*! \brief Do a query to the database that cannot return row results.
       * 
       * \param query SQL query to execute.
       * 
       * \exception DBException If query fails.
       */
//      friend void doQuery(boost::shared_ptr<Database> database, const std::string &query);

      /*! \brief Do a query to the database that cannot return row results,
       *  and obtain the last auto increment value set by the query.
       * 
       * \param query SQL query to execute.
       * 
       * \exception DBException If query fails.
       */
//      friend int doQueryAndGetId(boost::shared_ptr<Database> database, const std::string &query);
    
    public:
      /*! \brief Connect to database dbName at host:port,
       * as user with passwd.
       * 
       * \note It is private, so that it serves as an interface
       *  and creation is only allowed through static methods
       * 
       * \param dbName Name of the database to connect to.
       * \param host Network host name of database manager to use
       *  for connecting to dbName.
       * \param user User name in the database manager used to connect
       *  to dbName.
       * \param passwd Password associated to user when connecting
       *  to dbName.
       * \param port Network port number where database manager listens.
       * 
       * \exception DBException If the connection cannot be established.
       */
      DatabaseImpl(const std::string &dbName, const std::string &host,
        const std::string &user, const std::string &passwd,
        int port);
        
      bool createDatabase(const std::string &newDb);

      bool changeDatabase(const std::string &newDb);
      
      void close();
      
      virtual ~DatabaseImpl();
    private:
      void disconnect();
    
    };
    
    
    DatabaseImpl::DatabaseImpl(const std::string &dbName, const std::string &host,
        const std::string &user, const std::string &passwd,
        int port) : Database(), dbName_(dbName), host_(host), user_(user),
          passwd_(passwd), port_(port), disconnected_(false) {

      try {
        con_.connect(dbName_.c_str(), host_.c_str(), user_.c_str(), 
              passwd_.c_str(), port_);
      } catch (mysqlpp::Exception &ex) {
        logger().debugStream() << "Could not connect to database."
          << " Error message: " << ex.what() 
          << log4cpp::CategoryStream::EOL;
        throw DBException("Could not connect to database.");
      }
    }

    void DatabaseImpl::disconnect() {
#if (MYSQLPP_HEADER_VERSION >= MYSQLPP_VERSION(3, 0, 0))
    	  con_.disconnect();
#else
    	  con_.close();
#endif
    }
    
    void DatabaseImpl::pingAndReconnect() {
      if (disconnected_) {
        throw DBException("Disconnected. No new operations are permited.");
      }
      
#if (MYSQLPP_HEADER_VERSION >= MYSQLPP_VERSION(3, 0, 0))
      if (con_.ping() == false) {
#else
      if (con_.ping() != 0) {
#endif

    	this->disconnect();
        
        try {
          con_.connect(dbName_.c_str(), host_.c_str(),
                  user_.c_str(), passwd_.c_str(), port_);
        } catch (mysqlpp::Exception &ex) {
          disconnected_ = true;
          logger().debugStream() << "Could not reconnect to"
            << " database. Error message: " << ex.what()
            << log4cpp::CategoryStream::EOL;
          throw DBException("Connection to database lost.");
        }
      }
    }
    
    void DatabaseImpl::close() {
      if (disconnected_) {
        throw DBException("Disconnected. No new operations are permited.");
      }
      
      this->disconnect();
      disconnected_ = true;
    }


    bool DatabaseImpl::createDatabase(const std::string &newDb) {
      boost::mutex::scoped_lock scoped_lock(db_mutex);
      if (disconnected_) {
        throw DBException("Disconnected. No new operations are permited.");
      }
      pingAndReconnect();
      return con_.create_db(newDb);
    }

    bool DatabaseImpl::changeDatabase(const std::string &newDb) {
      boost::mutex::scoped_lock scoped_lock(db_mutex);
      if (disconnected_) {
        throw DBException("Disconnected. No new operations are permited.");
      }
      pingAndReconnect();

      try {
        if (con_.select_db(newDb)) {
          dbName_ = newDb;
          return true;
        } else {
          return false;
        }
      } catch (mysqlpp::DBSelectionFailed &ex) {
        return false;
      }
    }

    DatabaseImpl::~DatabaseImpl() {
      // does not throw
      this->disconnect();
    }
    
  } // end anonymous namespace
    
  const vector<map<string, string> > 
  doRowQuery(boost::shared_ptr<Database> database, const std::string &query) {
    boost::shared_ptr<DatabaseImpl> db 
      = boost::dynamic_pointer_cast<DatabaseImpl>(database);
    boost::mutex::scoped_lock scoped_lock(db->db_mutex);
    db->pingAndReconnect();
    mysqlpp::Query theQuery = db->con_.query();
    try {
      vector<map<string, string> > result;
      MYSQLPP_RESULT queryResult = theQuery.store(query.c_str());
      mysqlpp::Row row;
      mysqlpp::Row::size_type i;
      for(i = 0; i != queryResult.size(); ++i) {
        row = queryResult.at(i);
        mysqlpp::Row::size_type j;
        map<string, string> mapRow;
        for(j = 0; j != row.size(); ++j) {
          mapRow[queryResult.field_name(j)] 
            = rowFieldAsString(row, j);
        }
        result.push_back(mapRow);
      }
      return result;
    } catch (mysqlpp::Exception &ex) {
      logger().debugStream() << "Problems with query to database."
        << " Error message: " << log4cpp::CategoryStream::EOL;
      throw DBException("Problems with query to database.");
    }
  } 
  
  void doQuery(boost::shared_ptr<Database> database, const std::string &query) {
    boost::shared_ptr<DatabaseImpl> db 
      = boost::dynamic_pointer_cast<DatabaseImpl>(database);
    boost::mutex::scoped_lock scoped_lock(db->db_mutex);
    db->pingAndReconnect();
    mysqlpp::Query theQuery = db->con_.query();
    try {
      theQuery.execute(query.c_str());
    } catch (mysqlpp::Exception &ex) {
      logger().debugStream() << "Problems with query to database."
        << " Error message: " << log4cpp::CategoryStream::EOL;
      throw DBException("Problems with query to database.");
    }
  }
  
  int doQueryAndGetId(boost::shared_ptr<Database> database, const std::string &query) {
    boost::shared_ptr<DatabaseImpl> db 
      = boost::dynamic_pointer_cast<DatabaseImpl>(database);
    boost::mutex::scoped_lock scoped_lock(db->db_mutex);
    db->pingAndReconnect();
    mysqlpp::Query theQuery = db->con_.query();
    try {
      theQuery.execute(query.c_str());
    } catch (mysqlpp::Exception &ex) {
      logger().debugStream() << "Problems with query to database."
        << " Error message: " << log4cpp::CategoryStream::EOL;
      throw DBException("Problems with query to database.");
    }
#if (MYSQLPP_HEADER_VERSION >= MYSQLPP_VERSION(3, 0, 0))
    return theQuery.insert_id();
#else
    return db->con_.insert_id();
#endif
  }
    
 
  boost::shared_ptr<Database> Database::openDatabase(const std::string &dbName, const std::string &host,
        const std::string &user, const std::string &passwd,
        int port) {
          
    return boost::shared_ptr<Database>(new DatabaseImpl(dbName, host, user, passwd, port));          
  }
  
} // end namespace
} // end namespace
} // end namespace
