/*! \file azucar/utils/database/mysqlInit.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <mysql.h>

#include "azucar/utils/database/mysqlInit.hpp"
#include "azucar/utils/exceptions.hpp"

using namespace std;

namespace {
  /*! \brief Init MySQL Library.
   * 
   * Library is only initialized if not previously done
   * (that is, it is in status Initializer::NON_INITIALIZED).
   * 
   * If error, status changes to Initializer::INIT_ERROR,
   * and an exception is thrown. If initialized, status
   * changes to Initializer::INITIALIZED.
   * 
   * \exception BaseException If it started as Initializer::NON_INITIALIZED
   *  and initialization failed.
   */
  void init(azucar::utils::Initializer& initObj) {
    if ((initObj.status == azucar::utils::Initializer::NON_INITIALIZED)
    		or (initObj.status == azucar::utils::Initializer::DESTROYED)){
      if (mysql_library_init(-1, 0, 0) != 0) {
        initObj.status = azucar::utils::Initializer::INIT_ERROR;
        throw azucar::utils::BaseException("Could not initialize MySQL library.");
      }
      initObj.status = azucar::utils::Initializer::INITIALIZED;
    }
  }
  
  /*! \brief Finalize MySQL library.
   * 
   * Library is only finalized if in Initializer::INITIALIZED
   * state.
   * 
   * In that case, it is finalized and status changes to 
   * Initializer::DESTROYED.
   */
  void destroy(azucar::utils::Initializer& initObj) {
    if (initObj.status == azucar::utils::Initializer::INITIALIZED) {
      mysql_library_end();
      initObj.status = azucar::utils::Initializer::DESTROYED;
    }
  }
};

namespace azucar {
namespace utils {
namespace database {


Initializer mysqlInitializer(&init, &destroy);

} // end namespace
} // end namespace
} // end namespace

