/*! \file azucar/utils/geometry.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include "azucar/utils/geometry.hpp"

#include <cmath>
#include <cassert>

#include <iostream>


#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

using namespace std;

double radiansToDegrees(double rad) {
  return (rad / M_PI) * 180.0;
}

double degreesToRadians(double deg) {
  return (deg / 180) * M_PI;
}

DoubleVector setPoint(const DoubleVector &a, const DoubleVector &b,
    const DoubleVector &c, double distance, double angle, double torsion) {
  // obtain d2 coordinates
  assert(angle >= 0.0 and angle <= M_PI);
  assert(distance >= 0.0);
  DoubleVector d2;
  double costheta = cos(M_PI - angle);
  double sintheta = sin(M_PI - angle);
  double cosphi = cos(torsion);
  double sinphi = sin(torsion);
  d2.x = distance * costheta;
  d2.y = distance * cosphi * sintheta;
  d2.z = distance * sinphi * sintheta;
  DoubleVector unitbc = unitVector(c - b);
  // actually, the normal would be defined for axis X as unitbc
  // and axis Y positive the projection of (a-b) on axis Y. Since 
  // angle ABC is always less than PI, then projection of (a-b) on axis Y
  // is always positive axis Y. Therefore, normal = crossProduct(unitbc, ba),
  // which is the same as crossProduct(ab, unitbc).
  DoubleVector ab = b - a;
  DoubleVector normal = unitVector(crossProduct(ab,unitbc));
  DoubleVector ncrossbc = unitVector(crossProduct(normal, unitbc));
  
  DoubleVector newCoordinates
    = d2.x * unitbc + d2.y * ncrossbc + d2.z * normal;
  
  return newCoordinates + c;  
}

DoubleVector setZMatrixPoint(const DoubleVector &point1, double distance,
    const DoubleVector &point2, double angle, 
    const DoubleVector &point3, double torsion) {
  return setPoint(point3, point2, point1, distance, angle, torsion);
}  


void quaternionToRotationMatrix(const gsl_vector *quaternion,
    gsl_matrix *rotation) {

  assert(quaternion->size == 4);
  assert(rotation->size1 == 3);
  assert(rotation->size2 == 3);

  // now, construct rotation matrix from eigenvector
  double q0 = gsl_vector_get(quaternion, 0);
  double q1 = gsl_vector_get(quaternion, 1);
  double q2 = gsl_vector_get(quaternion, 2);
  double q3 = gsl_vector_get(quaternion, 3);
  
  double b0 = 2.0 * q0;
  double b1 = 2.0 * q1;
  double b2 = 2.0 * q2;
  double b3 = 2.0 * q3;
  
  double q00 = b0*q0 -1.0;
  double q01 = b0*q1;
  double q02 = b0*q2;
  double q03 = b0*q3;
  double q11 = b1*q1;
  double q12 = b1*q2;
  double q13 = b1*q3;
  double q22 = b2*q2;
  double q23 = b2*q3;
  double q33 = b3*q3;
  
  *gsl_matrix_ptr(rotation, 0, 0) = q00 + q11;
  *gsl_matrix_ptr(rotation, 0, 1) = q12 - q03;
  *gsl_matrix_ptr(rotation, 0, 2) = q13 + q02;
  
  *gsl_matrix_ptr(rotation, 1, 0) = q12 + q03;
  *gsl_matrix_ptr(rotation, 1, 1) = q00 + q22;
  *gsl_matrix_ptr(rotation, 1, 2) = q23 - q01;

  *gsl_matrix_ptr(rotation, 2, 0) = q13 - q02;
  *gsl_matrix_ptr(rotation, 2, 1) = q23 + q01;
  *gsl_matrix_ptr(rotation, 2, 2) = q00 + q33;

}

double RMSD(const gsl_matrix *x, const gsl_matrix *y) {
  // do z=x-y
  int n = x->size2;
  gsl_matrix *z = gsl_matrix_alloc(3, n);
  gsl_matrix_memcpy(z, x);
  gsl_matrix_sub(z, y);

  // sum all elements in the result
  double rmsd = 0.0;
  for(int i = 0; i < 3; ++i) {
    for(int j=0; j < n; ++j) {
      rmsd += (gsl_matrix_get(z, i, j) * gsl_matrix_get(z, i, j));
    }
  }
  rmsd = sqrt(rmsd/n);
  // return as RMSD
  gsl_matrix_free(z);
  
  return rmsd;
}

double doRMSD(const gsl_matrix *x, const gsl_matrix *y,
  gsl_matrix *transform, gsl_vector *rotQuaternion) {

  assert(x->size1 == 3);
  assert(y->size1 == 3);
  assert(x->size2 == y->size2);
  assert(transform->size1 == 4);
  assert(transform->size2 == 4);
  assert(rotQuaternion->size == 4);
  
  int n = x->size2;

  gsl_vector *meanx = gsl_vector_alloc(3);
  gsl_vector *meany = gsl_vector_alloc(3);
  
  gsl_vector_set_zero(meanx);
  gsl_vector_set_zero(meany);
  
  for(int j=0; j < n; ++j) {
    gsl_vector_const_view xcol = gsl_matrix_const_column(x, j);
    gsl_vector_add(meanx, &xcol.vector);
    
    gsl_vector_const_view ycol = gsl_matrix_const_column(y, j);
    gsl_vector_add(meany, &ycol.vector);
  }
  
  gsl_vector_scale(meanx, 1.0/n);
  gsl_vector_scale(meany, 1.0/n);
  
  // we now have the centroids
  
  gsl_matrix *centeredx = gsl_matrix_alloc(3, n);
  gsl_matrix *centeredy = gsl_matrix_alloc(3, n);
  gsl_matrix_memcpy(centeredx, x);
  gsl_matrix_memcpy(centeredy, y);
  
  for(int j=0; j < n; ++j) {
    gsl_vector_view xcol = gsl_matrix_column(centeredx, j);
    gsl_vector_sub(&xcol.vector, meanx);
    gsl_vector_view ycol = gsl_matrix_column(centeredy, j);
    gsl_vector_sub(&ycol.vector, meany); 
  }

  double xnorm = 0.0;
  double ynorm = 0.0;
  
  for(int i=0; i < 3; ++i) {
    double tempSum;
    gsl_vector_view xrow = gsl_matrix_row(centeredx, i);
    gsl_blas_ddot(&xrow.vector, &xrow.vector, &tempSum);
    xnorm += tempSum;
    
    gsl_vector_view yrow = gsl_matrix_row(centeredy, i);
    gsl_blas_ddot(&yrow.vector, &yrow.vector, &tempSum);
    ynorm += tempSum;
  }
  
  
  gsl_matrix *r = gsl_matrix_alloc(3, 3);
  
  gsl_matrix_set_zero(r);
  
  gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, centeredx, centeredy, 0.0, r);
  
  // now, r = x * trans(y)
  
  gsl_matrix *nm = gsl_matrix_alloc(4, 4);
  
  *gsl_matrix_ptr(nm, 0, 0) = gsl_matrix_get(r, 0, 0) + gsl_matrix_get(r, 1, 1)
    + gsl_matrix_get(r, 2, 2);
  *gsl_matrix_ptr(nm, 1, 0) = gsl_matrix_get(r, 1, 2) - gsl_matrix_get(r, 2, 1);
  *gsl_matrix_ptr(nm, 2, 0) = gsl_matrix_get(r, 2, 0) - gsl_matrix_get(r, 0, 2);
  *gsl_matrix_ptr(nm, 3, 0) = gsl_matrix_get(r, 0, 1) - gsl_matrix_get(r, 1, 0);
  
  *gsl_matrix_ptr(nm, 0, 1) = gsl_matrix_get(nm, 1, 0);
  *gsl_matrix_ptr(nm, 1, 1) = gsl_matrix_get(r, 0, 0) - gsl_matrix_get(r, 1, 1) 
    - gsl_matrix_get(r, 2, 2);
  *gsl_matrix_ptr(nm, 2, 1) = gsl_matrix_get(r, 0, 1) + gsl_matrix_get(r, 1, 0);
  *gsl_matrix_ptr(nm, 3, 1) = gsl_matrix_get(r, 0, 2) + gsl_matrix_get(r, 2, 0);
  
  *gsl_matrix_ptr(nm, 0, 2) = gsl_matrix_get(nm, 2, 0);
  *gsl_matrix_ptr(nm, 1, 2) = gsl_matrix_get(nm, 2, 1);
  *gsl_matrix_ptr(nm, 2, 2) = -gsl_matrix_get(r, 0, 0) + gsl_matrix_get(r, 1, 1)
    - gsl_matrix_get(r, 2, 2);
  *gsl_matrix_ptr(nm, 3, 2) = gsl_matrix_get(r, 1, 2) + gsl_matrix_get(r, 2, 1);
  
  *gsl_matrix_ptr(nm, 0, 3) = gsl_matrix_get(nm, 3, 0);
  *gsl_matrix_ptr(nm, 1, 3) = gsl_matrix_get(nm, 3, 1);
  *gsl_matrix_ptr(nm, 2, 3) = gsl_matrix_get(nm, 3, 2);
  *gsl_matrix_ptr(nm, 3, 3) = -gsl_matrix_get(r, 0, 0) - gsl_matrix_get(r, 1, 1)
    + gsl_matrix_get(r, 2, 2);      

  // now, find maximum eigenvalue and eigenvector of n
//  gsl_eigen_symm_workspace *eigenws = gsl_eigen_symm_alloc(4);  
  gsl_eigen_symmv_workspace *eigenvws = gsl_eigen_symmv_alloc(4);
  gsl_matrix *evec = gsl_matrix_alloc(4, 4);
  gsl_vector *eval = gsl_vector_alloc(4);
  gsl_eigen_symmv(nm, eval, evec, eigenvws);

  gsl_eigen_symmv_sort(eval, evec, GSL_EIGEN_SORT_VAL_DESC);

  gsl_vector_view quaternion = gsl_matrix_column(evec, 0);
  gsl_vector_memcpy(rotQuaternion, &quaternion.vector);


  // set last row of transform to [0 1]
  gsl_vector_view lastRow = gsl_matrix_row(transform, 3);
  gsl_vector_set_zero(&lastRow.vector);
  gsl_vector_set(&lastRow.vector, 3, 1.0); 
  
  gsl_matrix_view rot = gsl_matrix_submatrix(transform, 0, 0, 3, 3);

  quaternionToRotationMatrix(&quaternion.vector, &rot.matrix);
  
  // translation is a 3-element vector
  gsl_vector_view translationExtended = gsl_matrix_column(transform, 3);
  gsl_vector_view translation 
    = gsl_vector_subvector(&translationExtended.vector, 0, 3);
  
  gsl_vector_set_zero(&translation.vector);
  
  gsl_blas_dgemv(CblasNoTrans, -1.0, &rot.matrix, meanx, 0.0, &translation.vector);
  gsl_vector_add(&translation.vector, meany);
  
  // now, calculate RMSD
  double rmsd = sqrt(std::max(0.0, ( (xnorm + ynorm) -2.0*(gsl_vector_get(eval, 0))))/n);


  gsl_vector_free(eval);
  gsl_matrix_free(evec);
  gsl_eigen_symmv_free(eigenvws);
  gsl_matrix_free(nm);
  gsl_matrix_free(r);
  gsl_matrix_free(centeredy);
  gsl_matrix_free(centeredx);
  gsl_vector_free(meanx);
  gsl_vector_free(meany);
  
  return rmsd;
}



bool lessThanOrEqual(const DoubleVector &v, const DoubleVector &w) {
  if (v.x < w.x) {
    return true;
  } else if (v.x == w.x) {
    if (v.y < w.y) {
      return true;
    } else if (v.y == w.y) {
      return v.z <= w.z;
    } else {
      return false;
    }
  } else {
    return false;
  }
}

bool lessThan(const DoubleVector &v, const DoubleVector &w) {
  if (v.x < w.x) {
    return true;
  } else if (v.x == w.x) {
    if (v.y < w.y) {
      return true;
    } else if (v.y == w.y) {
      return v.z < w.z;
    } else {
      return false;
    }
  } else {
    return false;
  }
}

bool greaterThan(const DoubleVector &v, const DoubleVector &w) {
  return not lessThanOrEqual(v, w);
}

bool lessThanOrEqualDim(int dimension, 
        const DoubleVector &v, const DoubleVector &w) {
          
  switch(dimension) {
    case 1: return lessThanOrEqual(v, w); break;
    case 2: {
      DoubleVector auxV(v.y, v.z, v.x);
      DoubleVector auxW(w.y, w.z, w.x);
      return lessThanOrEqual(auxV, auxW);
      break;
    }
    case 3: {
      DoubleVector auxV(v.z, v.x, v.y);
      DoubleVector auxW(w.z, w.x, w.y);
      return lessThanOrEqual(auxV, auxW);
      break;
    }
    default:
      assert(false);
  }
}

bool lessThanDim(int dimension,
        const DoubleVector &v, const DoubleVector &w) {
  switch(dimension) {
    case 1: return lessThan(v, w); break;
    case 2: {
      DoubleVector auxV(v.y, v.z, v.x);
      DoubleVector auxW(w.y, w.z, w.x);
      return lessThan(auxV, auxW);
      break;
    }
    case 3: {
      DoubleVector auxV(v.z, v.x, v.y);
      DoubleVector auxW(w.z, w.x, w.y);
      return lessThan(auxV, auxW);
      break;
    }
    default:
      assert(false);
  }
}        

double squareDistance(const DoubleVector &v1,
      const DoubleVector &v2) {
  return squareModulusVector(v2 - v1);        
}

bool greaterThanDim(int dimension,
        const DoubleVector &v, const DoubleVector &w) {
  return not lessThanOrEqualDim(dimension, v, w);          
}                  

bool contains(const Box &container, const Box &region) {
  return contains(container, region.minPoint)
    and contains(container, region.maxPoint);
}

bool contains(const Box &container, const DoubleVector &point) {
  return (container.minPoint.x <= point.x
    and container.minPoint.y <= point.y 
    and container.minPoint.z <= point.z
    and container.maxPoint.x >= point.x
    and container.maxPoint.y >= point.y
    and container.maxPoint.z >= point.z);
}



bool intersects(const Box &region1, const Box &region2) {
  double minx = fmax(region1.minPoint.x, region2.minPoint.x);
  double miny = fmax(region1.minPoint.y, region2.minPoint.y);
  double minz = fmax(region1.minPoint.z, region2.minPoint.z);
  double maxx = fmin(region1.maxPoint.x, region2.maxPoint.x);
  double maxy = fmin(region1.maxPoint.y, region2.maxPoint.y);
  double maxz = fmin(region1.maxPoint.z, region2.maxPoint.z);
  
  return ( (minx <= maxx) and (miny <= maxy) and (minz <= maxz));
}

bool contains(const MPFBox &container, const MPFBox &region) {
  return contains(container, region.minPoint)
    and contains(container, region.maxPoint);
}

bool contains(const MPFBox &container, const MPFVector &point) {
  return (cmp(container.minPoint.x, point.x) <= 0
    and cmp(container.minPoint.y, point.y) <= 0 
    and cmp(container.minPoint.z, point.z) <= 0
    and cmp(container.maxPoint.x, point.x) >= 0
    and cmp(container.maxPoint.y, point.y) >= 0
    and cmp(container.maxPoint.z, point.z) >= 0);
}

bool intersects(const MPFBox &region1, const MPFBox &region2) {
  mpf_class minx = (cmp(region1.minPoint.x, region2.minPoint.x) <= 0) ? region1.minPoint.x : region2.minPoint.x;
  mpf_class miny = (cmp(region1.minPoint.y, region2.minPoint.y) <= 0) ? region1.minPoint.y : region2.minPoint.y;
  mpf_class minz = (cmp(region1.minPoint.z, region2.minPoint.z) <= 0) ? region1.minPoint.z : region2.minPoint.z;
  mpf_class maxx = (cmp(region1.maxPoint.x, region2.maxPoint.x) <= 0) ? region1.minPoint.x : region2.minPoint.x;
  mpf_class maxy = (cmp(region1.maxPoint.y, region2.maxPoint.y) <= 0) ? region1.maxPoint.y : region2.maxPoint.y;
  mpf_class maxz = (cmp(region1.maxPoint.z, region2.maxPoint.z) <= 0) ? region1.maxPoint.z : region2.maxPoint.z;
  
  return ( (cmp(minx, maxx) <=0) 
    and (cmp(miny, maxy) <= 0) 
    and (cmp(minz, maxz) <= 0));
}


// /**
//  * Multiplication of a vector by a scalar.
//  */
// const Vector operator*(const Vector &v1, double scalar) {
//   Vector result;
//   result.x = v1.x * scalar;
//   result.y = v1.y * scalar;
//   result.z = v1.z * scalar;

//   return result;
// }

// /**
//  * Division of a vector by a scalar.
//  */
// const Vector operator/(const Vector &v1, double scalar) {
//   return v1 * (1.0 / scalar);
// }

// const Vector operator+(const Vector &v1, const Vector &v2) {
//   Vector result;
//   result.x = v1.x + v2.x;
//   result.y = v1.y + v2.y;
//   result.z = v1.z + v2.z;

//   return result;
// }


// /**
//  * Substraction of vector v2 to vector v1
//  */
// const Vector operator-(const Vector &v1, const Vector &v2) {
//   Vector result;
//   result.x = v1.x - v2.x;
//   result.y = v1.y - v2.y;
//   result.z = v1.z - v2.z;

//   return result;
// }

// /**
//  * Scalar product of vectors v1 and v2
//  */
// double operator*(const Vector &v1, const Vector &v2) {
//   return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
// }


// /**
//  * Modulus of vector v1
//  */
// double modulusVector(const Vector &v1) {
//   return sqrt(v1.x * v1.x + v1.y * v1.y + v1.z * v1.z);
// }

// /**
//  * Cross product v1 * v2
//  */
// Vector crossProduct(const Vector &v1, const Vector &v2) {
//   Vector result;
//   result.x = v1.y * v2.z - v1.z * v2.y;
//   result.y = -(v1.x * v2.z - v2.x * v1.z);
//   result.z = v1.x * v2.y - v1.y * v2.x;

//   return result;
// }


// /**
//  * Torsion angle in radians considering a-b-c-d
//  */
// double torsionAngle(const Vector &a, const Vector &b,
// 		    const Vector &c, const Vector &d) {

//   Vector ab;
//   ab.x = a.x - b.x;
//   ab.y = a.y - b.y;
//   ab.z = a.z - b.z;

//   Vector cb;
//   cb.x = c.x - b.x;
//   cb.y = c.y - b.y;
//   cb.z = c.z - b.z;

//   Vector cd;
//   cd.x = c.x - d.x;
//   cd.y = c.y - d.y;
//   cd.z = c.z - d.z;

//   Vector nda = crossProduct(ab, cb); // normals to the planes
//   Vector nea = crossProduct(cb, cd);

//   double bl = modulusVector(nda); // modulus
//   double el = modulusVector(nea); // modulus

//   double bel = nda * nea; // point product

//   assert( (bl * el) != 0);

//   bel /= (bl * el);

//   if (bel > 1.0)
//     {
//       bel = 1;
//     }
//   else if (bel < -1.0)
//     {
//       bel = -1;
//     }
  
//   double acosbel = acos(bel);

//   double signTest = cb * crossProduct(nda, nea);

//   if (signTest < 0) {
//     acosbel = -acosbel;
//   }

// //   acosbel = (acosbel > 0.0) ? M_PI - acosbel : M_PI + acosbel;

//   return acosbel;
// }
