/*! \file azucar/utils/process.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include "azucar/utils/process.hpp"
#include "azucar/utils/util.hpp"

#if defined(unix) || defined(__unix__) || defined(__unix)
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#endif

#include <set>
//#include <iostream>

#include <boost/config.hpp>
#include <boost/static_assert.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/date_time/posix_time/posix_time_types.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/utility.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/PropertyConfigurator.hh>
#include <log4cpp/NDC.hh>


// Code for signal managing (and system()-like implementation)
// is taken from Program 10.20, page 314-125 of 
// Advanced Programming in the UNIX Environment, W. Richard Stevens,
// 1993, Addison-Wesley (from now on, APUE).

// Code for maximum number of open file descriptors,
// taken from program 2.3, page 43, same book (APUE).



#if defined(__WIN32)
#endif

using namespace std;


namespace {
const std::string category_name = "process_cat";

log4cpp::Category &logger() {
  static log4cpp::Category &theLogger =   
    log4cpp::Category::getInstance(category_name);
  return theLogger;
}
  
} // end namespace


ProcessException::ProcessException(const std::string &msg) : azucar::utils::BaseException(msg) {};

ProcessException::~ProcessException() throw () {};


map<string, string> convertEnvironToMap(const char * const *environ) {
  map<string, string> resultEnv;

  for(int i=0; environ[i] != 0; ++i) {
    string keyAndValue = string(environ[i]);
    string::size_type equalPos = keyAndValue.find("=");
    if (equalPos == string::npos) {
      throw azucar::utils::StdStreamException("Char string is not formatted as environ");
    }
    string key = keyAndValue.substr(0, equalPos);
    string value = keyAndValue.substr(equalPos+1, string::npos);
    resultEnv[key] = value;
  }

  return resultEnv;
}

/*
 * This will allocate memory space needed for the C-type string (char *)
 * equivalent of inputString, and will copy the inputString into that
 * memory, returning a pointer to that memory.
 */
char *stringToChar(const string &inputString) {
  char *tmpPtr = new char[inputString.size() + 1]; // +1 for ending \0
  const char *cstr = inputString.c_str(); 
  copy(cstr, cstr + (inputString.size() + 1), tmpPtr); // +1 for ending \0
  return tmpPtr;
}

#if defined(unix) || defined(__unix__) || defined(__unix)

namespace {
  const int open_max_guess = 256;
  
  int open_max() {
    int result;
    errno = 0;
    if ( (result = sysconf(_SC_OPEN_MAX)) < 0) {
      if (errno == 0) {
        result = open_max_guess;
      } else {
        throw azucar::utils::StdStreamException("SC_OPEN_MAX error");
      }
    }
    
    return result;
  }
}

long getProcessID() {
  return getpid();
}

void executeInChild(const string &processPath,
		   const vector<string> &argv,
		   const map<string,string> &env,
		   const string &inputFilePath,
		   const string &outputAndErrorFilePath, 
       const string &workingDirectory) {
  // convert argv to array. It will allocate memory space for argvPtr
  // and argvArray
  vector<char *> argvPtr;
  vector<string>::const_iterator argvIt;
  for(argvIt = argv.begin(); argvIt != argv.end(); ++argvIt) {
    char *tmpPtr = stringToChar(*argvIt);
    argvPtr.push_back(tmpPtr);
  }

  char **argvArray = new char*[argvPtr.size()+1];
  argvArray[argvPtr.size()] = 0; // Null pointer
  copy(argvPtr.begin(), argvPtr.end(), argvArray);

  // convert env to vector<string> and later to array. It will allocate 
  // memory space for envVector and envArray
  vector<char *> envVector;
  map<string,string>::const_iterator envIt;
  for (envIt = env.begin(); envIt != env.end(); ++envIt) {
    string envString = (envIt->first) + string("=") + (envIt->second);
    char *tmpPtr = stringToChar(envString);
    envVector.push_back(tmpPtr);
  }
  char **envArray = new char*[envVector.size()+1];
  envArray[envVector.size()] = 0; // Null pointer
  copy(envVector.begin(), envVector.end(), envArray);

  int inFD;
  int outFD;

  // Open the file for stdin, make sure it obtains STDIN_FILENO, and
  // close the original descriptor.
  inFD = open(inputFilePath.c_str(), O_RDONLY);
  if (inFD == -1) {
    _exit(1);
  }
  if (dup2(inFD, STDIN_FILENO) == -1) {
    _exit(1);
  };
  close(inFD);

  // Open the file for stdout and stderr, 
  // make sure it obtains STDOUT_FILENO and STDERR_FILENO, and
  // close the original descriptor.
  outFD = open(outputAndErrorFilePath.c_str(), O_WRONLY | O_CREAT | O_TRUNC,
	       S_IWUSR | S_IRUSR);
  if (outFD == -1) {
    _exit(1);
  }
  if (dup2(outFD, STDOUT_FILENO) == -1) {
    _exit(1);
  }
  close(outFD);

  if (dup2(STDOUT_FILENO, STDERR_FILENO) == -1) {
    _exit(1);
  }
  
  // close all other file descriptors
  int maxNumOfOpenFds;
  try {
    maxNumOfOpenFds = open_max();
  } catch (azucar::utils::StdStreamException &ex) {
    _exit(1);
  }

  for (int fdnum = 3; fdnum < maxNumOfOpenFds; ++fdnum) {
    if (close(fdnum) == -1) {
      // if failure is due to other thing than trying to
      // close a not open descriptor
      // (though it also could be an invalid descriptor).
      if (errno != EBADF) {
        _exit(1);
      }
    };
  }

  // change directory if needed
  if (not workingDirectory.empty()) {
    if (chdir(workingDirectory.c_str()) == -1) {
      _exit(1);
    }
  }
  execve(processPath.c_str(), argvArray, envArray);
  _exit(127);

  // there is no need to clear the allocated memory, since _exit(1) will do
  // for us

}

void executeInChildSignal(const string &processPath,
       const vector<string> &argv,
       const map<string,string> &env,
       const string &inputFilePath,
       const string &outputAndErrorFilePath, 
       const string &workingDirectory,
       struct sigaction *saveintrPtr,
       struct sigaction *savequitPtr,
       sigset_t *savemaskPtr) {

//  sigaction(SIGINT, saveintrPtr, NULL);
//  sigaction(SIGQUIT, savequitPtr, NULL);
  sigprocmask(SIG_SETMASK, savemaskPtr, NULL);
  
  executeInChild(processPath, argv, env, inputFilePath,
    outputAndErrorFilePath, workingDirectory);        
}

void blockSIGCHLD(sigset_t &savemask) {
  
  sigset_t chldmask;
   
  sigemptyset(&chldmask); // now block SIGCHLD
  sigaddset(&chldmask, SIGCHLD);
#if defined(BOOST_HAS_THREADS)
#if defined(BOOST_HAS_PTHREADS)
  if (pthread_sigmask(SIG_BLOCK, &chldmask, &savemask) < 0) {
    logger().debugStream() << "Pthread Sigmask block error."
//      << " In ProcessExecute: " << processPath 
      << log4cpp::CategoryStream::EOL;
    throw ProcessException("Pthread Sigmask block error.");
  }
#else
// compilation error
  BOOST_STATIC_ASSERT(false);
#endif
#else
  if (sigprocmask(SIG_BLOCK, &chldmask, &savemask) < 0) {
    logger().debugStream() << "Sigprocmask block error."
//      << " In ProcessExecute: " << processPath 
      << log4cpp::CategoryStream::EOL;
    throw ProcessException("Sigprocmask block error.");
  }
#endif
}

void unblockSIGCHLD(sigset_t &savemask) {

#if defined(BOOST_HAS_THREADS)
#if defined(BOOST_HAS_PTHREADS)
  if (pthread_sigmask(SIG_SETMASK, &savemask, NULL) < 0) {
    logger().debugStream() << "Pthread Sigmask unblock error."
//      << " In ProcessExecute: " << processPath 
      << log4cpp::CategoryStream::EOL;
    throw ProcessException("Pthread Sigmask unblock error.");
  }
#else
// compilation error
  BOOST_STATIC_ASSERT(false);
#endif
#else
  if (sigprocmask(SIG_SETMASK, &savemask, NULL) < 0) {
    logger().debugStream() << "Sigprocmask unblock error."
//      << " In ProcessExecute: " << processPath 
      << log4cpp::CategoryStream::EOL;
    throw ProcessException("Sigprocmask unblock error.");
  }
#endif

}



int processExecute(const string &processPath,
		   const vector<string> &argv,
		   const map<string,string> &env,
		   const string &inputFilePath,
		   const string &outputAndErrorFilePath,
       const string &workingDirectory) {

  logger().infoStream() << "Preparing execution of: "
    << processPath << " with args: ";
  for(vector<string>::const_iterator it = argv.begin();
        it != argv.end(); ++it) {
    logger().infoStream() << "'" << *it << "' ";
  }
  logger().infoStream() << log4cpp::CategoryStream::EOL;

//  cerr << "Ready to execute: " << endl;
//  cerr << processPath;
//  vector<string>::const_iterator vit;
//  for(vit = argv.begin(); vit != argv.end(); ++vit) {
//    cerr << " " << *vit;
//  }
//  
//  cerr << endl;
//  cerr << "Input file path: " << inputFilePath << endl;
//  cerr << "Output file path: " << outputAndErrorFilePath << endl;
//  cerr << "Working directory: " << workingDirectory << endl;

  // execute
  pid_t pid = fork();
  int status;
  struct sigaction /*ignore,*/ saveintr, savequit;
  sigset_t savemask;

//  ignore.sa_handler = SIG_IGN; // ignore SIGINT and SIGQUIT,
//  // as in system() 
//  sigemptyset(&ignore.sa_mask);
//  ignore.sa_flags = 0;
//  if (sigaction(SIGINT, &ignore, &saveintr) < 0) {
//    return -1;
//  }
//  if (sigaction(SIGQUIT, &ignore, &savequit) < 0) {
//    return -1;
//  }
  
  blockSIGCHLD(savemask);
  
  ON_BLOCK_EXIT(&unblockSIGCHLD, Loki::ByRef(savemask));


  if ( pid == -1) {
    logger().debugStream() << "Process creation error."
      << " In ProcessExecute: " << processPath 
      << log4cpp::CategoryStream::EOL;
    throw ProcessException("Process creation error.");
  } else if (pid == 0) {
    // child process
    executeInChildSignal(processPath, argv, env, 
		   inputFilePath, outputAndErrorFilePath, workingDirectory,
       &saveintr, &savequit, &savemask);
  } else {
     // parent process
    // processExecute must be synchronous (that is, blocking until the child
    // ends its labor). Since some signals might interrupt the waitpid,
    // we check for EINTR and wait again if needed.
    while (waitpid(pid, &status, 0) == -1) {
      if (errno != EINTR) {
        logger().debugStream() << "Process wait error."
          << " In ProcessExecute: " << processPath
          << " Errno: " << errno 
          << log4cpp::CategoryStream::EOL;
        throw ProcessException("Process wait error.");
      }
    }
  }
  
//  if (sigaction(SIGINT, &saveintr, NULL) < 0) {
//    return -1;
//  }
//  if (sigaction(SIGQUIT, &savequit, NULL) < 0) {
//    return -1;
//  }

//  unblockSIGCHLD(savemask); // executed now on block exit
    
  return status;

}

namespace {
  
  void internalKill(long pid) {
    vector<string> argVector;
    argVector.push_back(string("/usr/bin/kill"));
    argVector.push_back(string("-9"));
    argVector.push_back(boost::lexical_cast<string>(pid));
  
    std::map<string,string> envVector = convertEnvironToMap(environ);
  
    // if pid in async list, block SIGCHLD
    
  
  
    try {
    /*int returnStatus =*/ processExecute(string("/usr/bin/kill"),
              argVector, envVector, string("/dev/null"),
              string("/dev/null"));
    } catch (ProcessException &ex) {
      throw ProcessException(string("Kill operation failed: ")
        + ex.getMessage());
    }
    
    // if pid in async list, waitpid for child (no blocking)
    // up to a time. If that time does expire, unblock
    // and throw exception
  }


  
  class AsyncChildrenSet : boost::noncopyable {
  private:
    boost::mutex monitorMutex_;
    std::set<pid_t> childrenSet_;
  public:
    AsyncChildrenSet() : childrenSet_() {};
    void insert(pid_t pid) {
      boost::mutex::scoped_lock scoped_lock(monitorMutex_);
      childrenSet_.insert(pid);
    };
    
    
    bool kill(pid_t pid) {
      boost::mutex::scoped_lock scoped_lock(monitorMutex_);
      if (childrenSet_.find(pid) != childrenSet_.end()) {
        sigset_t savemask;
        blockSIGCHLD(savemask);
        ON_BLOCK_EXIT(&unblockSIGCHLD, Loki::ByRef(savemask));

        internalKill(pid);
        
        // 
        for(int i=0; i < 5; ++i) { // 5 tries
          // do not do this first time
          if (i > 0) {
            struct timespec sleepTime;
            sleepTime.tv_sec = 5; // 5 seconds between tries
            sleepTime.tv_nsec = 0; 
            
            nanosleep(&sleepTime, NULL);
          }
          
          pid_t result = waitpid(pid, NULL, WNOHANG);
          if (result == -1) {
            if (errno == ECHILD) {
              // remove from set
              childrenSet_.erase(pid);
              break;
            } else {
              logger().debugStream() << "Error while waiting for"
                << " possible zombie child " << pid
                << " Errno: " << errno << log4cpp::CategoryStream::EOL;
              throw ProcessException("Error in waiting process.");
            }
          } else if (result != 0) {
            // remove from set
            childrenSet_.erase(pid);
            break;
          }
        }
//        unblockSIGCHLD(savemask); // executed now on block exit        
        
        
        return true;
      } else {
        return false;
      }
    }
    
    bool remove(pid_t pid) {
      boost::mutex::scoped_lock scoped_lock(monitorMutex_);
      return (childrenSet_.erase(pid) != 0);
    };
    void clearZombies() {
      boost::mutex::scoped_lock scoped_lock(monitorMutex_);
      std::set<pid_t>::iterator it;
      std::set<pid_t> childrenRemoved;
      for(it = childrenSet_.begin(); 
          it != childrenSet_.end(); ++it) {
        pid_t result = waitpid(*it, NULL, WNOHANG);
        if (result == -1) {
          if (errno == ECHILD) {
            // remove from set
            childrenRemoved.insert(*it);
          } else {
            logger().debugStream() << "Error while waiting for"
              << " possible zombie child " << *it
              << " Errno: " << errno << log4cpp::CategoryStream::EOL;
          }
        } else if (result != 0) {
          // remove from set
          childrenRemoved.insert(*it);
        }
      }
      for(it = childrenRemoved.begin(); it != childrenRemoved.end();
            ++it) {
        childrenSet_.erase(*it);
      }
    }
    ~AsyncChildrenSet() {};
  };
  
AsyncChildrenSet &asyncChildrenSet() {
  static AsyncChildrenSet theSet;
  return theSet;
}
}

void clearZombies() {
  asyncChildrenSet().clearZombies();
}

bool clearWaitedChild(pid_t pid) {
  return asyncChildrenSet().remove(pid);
}

long processExecuteAsynchronous(const string &processPath,
       const vector<string> &argv,
       const map<string,string> &env,
       const string &inputFilePath,
       const string &outputAndErrorFilePath,
       const string &workingDirectory,
       bool doClearZombies) {

  logger().infoStream() << "Preparing async execution of: "
    << processPath << " with args: ";
  for(vector<string>::const_iterator it = argv.begin();
        it != argv.end(); ++it) {
    logger().infoStream() << "'" << *it << "' ";
  }
  logger().infoStream() << log4cpp::CategoryStream::EOL;


  // try cleaning zombie processes first
  if (doClearZombies) {
    clearZombies();
  }

  // execute
  pid_t pid = fork();


  switch(pid) {
  case 0: // child process
    executeInChild(processPath, argv, env, 
       inputFilePath, outputAndErrorFilePath, workingDirectory);
    _exit(1); // shouldn't get here
    break;
  case -1: // error
    logger().debugStream() << "Process creation error."
      << " In ProcessExecuteAsynchronous: " << processPath 
      << log4cpp::CategoryStream::EOL;
    throw ProcessException("Process creation error.");
  default: // parent process
    asyncChildrenSet().insert(pid);
    return pid; // child was forked, everything seems OK, return 0 (OK code).
    break;
  }

  // shouldn't get here
}

void callKill(long pid) {

  if (not asyncChildrenSet().kill(pid)) {
    internalKill(pid);
  }
  
//  internalKill(pid);
  
  clearZombies(); // try to clean up if possible
}

// IPC STRUCTURES
namespace {
#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
/* union semun is defined by including <sys/sem.h> */
#else
/* according to X/OPEN we have to define it ourselves */
union semun {
     int val;                  /* value for SETVAL */
     struct semid_ds *buf;     /* buffer for IPC_STAT, IPC_SET */
     unsigned short *array;    /* array for GETALL, SETALL */
             /* Linux specific part: */
     struct seminfo *__buf;    /* buffer for IPC_INFO */
};
#endif

unsigned int secondsLeft(const boost::posix_time::ptime &deadline) {
  boost::posix_time::time_duration timeLeft
    = deadline - boost::posix_time::second_clock::local_time();

  return ( (timeLeft.total_seconds() > 0) ? timeLeft.total_seconds() : 0);
}

class BidirectionalSemaphore {
public:
  int semid_;
  bool creator_;
  bool closed_;
  BidirectionalSemaphore() : creator_(false), closed_(false) {
    semid_ = semget(IPC_PRIVATE, 2, S_IRUSR | S_IWUSR);
    if (semid_ == -1) {
      logger().debug("Error creating semaphore set.");
      throw ProcessException("Error creating semaphore set.");
    }
    creator_ = true;



    semun data;
    data.val = 0;

    if (semctl(semid_, 0, SETVAL, data) != 0) {
      semctl(semid_, 0, IPC_RMID, data);
      logger().debug("Error setting semaphore values.");
      throw ProcessException("Error setting semaphore values.");
    }
    data.val = 0;
    if (semctl(semid_, 1, SETVAL, data) != 0) {
      semctl(semid_, 0, IPC_RMID, data);
      logger().debug("Error setting semaphore values.");
      throw ProcessException("Error setting semaphore values.");
    }

  }

  BidirectionalSemaphore(int semid) : semid_(semid), creator_(false),
              closed_(false) {};

  void doOp(int semNum, int op) {
      struct sembuf semops[1];

      semops[0].sem_num = semNum;
      semops[0].sem_op = op;
      semops[0].sem_flg = 0;

      if (semop(semid_, semops, 1) != 0) {
        this->close();
        logger().debug("Semaphore operation error.");
        throw ProcessException("Semaphore operation error.");
      }
  }
  
  bool doOp(int semNum, int op, unsigned int secondsToWait) {
      struct sembuf semops[1];

      semops[0].sem_num = semNum;
      semops[0].sem_op = op;
      semops[0].sem_flg = 0;

      struct timespec timeout;
    
      timeout.tv_sec = secondsToWait;
      timeout.tv_nsec = 0;
    
      if (semtimedop(semid_, semops, 1, &timeout) != 0) {
        if (errno == EAGAIN) {
          return false;
        }
        this->close();
        logger().debug("Semaphore timed operation error.");
        throw ProcessException("Semaphore timed operation error.");
      }
    
      return true;
  }

  void close() {
    if (creator_ and not closed_) {
      semun data;
      semctl(semid_, 0, IPC_RMID, data);
    }

    closed_ = true;
  }

  void silentClose() {
    closed_ = true;
  }

  ~BidirectionalSemaphore() {
    if (not closed_) {
      this->close();
    }
  }
};

const size_t shared_memory_size = sizeof(double)*12000; // in chars

class ShMemCall {
public:
  int shmid_;
  boost::shared_ptr<BidirectionalSemaphore> semaph;
  void *shmptr_;
  bool closed_;
  size_t bufSize_;
  ShMemCall(size_t size = shared_memory_size) : bufSize_(size) {
    semaph.reset(new BidirectionalSemaphore());

    // create shared memory
    if ( (shmid_ = shmget(IPC_PRIVATE, size, (SHM_R | SHM_W))) < 0) {
      logger().debug("ShMemCall: Could not get shared memory.");
      throw ProcessException("Could not get shared memory.");
    }

    if ( (shmptr_ = shmat(shmid_, 0, 0)) == (void *) -1) {
      logger().debug("ShMemCall: Could not attach shared memory.");
      throw ProcessException("Could not attach shared memory.");
    }

    closed_ = false;

    return;
  }

  bool call(void *args, size_t size, void *result, size_t resultSize, 
      unsigned int secondsToWait) {
    // write args
    // while not everything written:
    //      wait(0)
    size_t sizeLeft = size;
    char *destPtr = static_cast<char *>(shmptr_);
    char *srcPtr = static_cast<char *>(args);

    boost::posix_time::ptime maxTime(boost::posix_time::second_clock::local_time() + boost::posix_time::seconds(secondsToWait));

    while (sizeLeft > 0) {
      int currentSize = ((sizeLeft > bufSize_) ? bufSize_ : sizeLeft);
      if (not semaph->doOp(0, 0, secondsLeft(maxTime))) {
        return false;
      }
      memcpy(destPtr, srcPtr, currentSize);
    //      write()
    //      signal(2)

      if (not semaph->doOp(0, 2, secondsLeft(maxTime))) {
        return false;
      }
      sizeLeft = sizeLeft - currentSize;
      destPtr = destPtr + currentSize;
      srcPtr = srcPtr + currentSize;
    }


    srcPtr = static_cast<char *>(shmptr_);
    destPtr = static_cast<char *>(result);
    sizeLeft = resultSize;
    while (sizeLeft > 0) {
      size_t currentSize = ((sizeLeft > bufSize_) ? bufSize_ : sizeLeft);
    // read result
    // while not everything read:
    //      wait(1)
      if (not semaph->doOp(1, -1, secondsLeft(maxTime))) {
        return false;
      }
      memcpy(destPtr, srcPtr, currentSize);
    //      read()
    //      wait(1)
      if (not semaph->doOp(1, -1, secondsLeft(maxTime))) {
        return false;
      }
      sizeLeft = sizeLeft - currentSize;
      destPtr = destPtr + currentSize;
      srcPtr = srcPtr + currentSize;
    }
    
    return true;
  }

  void close() {
    if (not closed_) {
      semaph->close();
      if (shmdt(shmptr_) != 0) {
        logger().debug("ShMemCall.close: Could not dettach shared memory.");
        throw ProcessException("Could not dettach shared memory.");
      }

      if (shmctl(shmid_, IPC_RMID, 0) < 0) {
        logger().debug("ShMemCall.close: Could not remove shared memory.");
        throw ProcessException("Could not remove shared memory.");
      }
    
      closed_ = true;
    }
  }

  void silentClose() {
    if (not closed_) {
      semaph->silentClose();
      if (shmdt(shmptr_) != 0) {
        logger().debug("ShMemCall.silentClose: Could not dettach shared memory.");
        throw ProcessException("Could not dettach shared memory.");
      }
    }
    closed_ = true;
  }

  ~ShMemCall() {
    if (not closed_) {
      this->close();
    }
  }
};

class ShMemServe {
public:
  int shmid_;
  boost::shared_ptr<BidirectionalSemaphore> semaph;
  void *shmptr_;
  bool closed_;
  size_t bufSize_;
  ShMemServe(int shmid, int semid, size_t size = shared_memory_size) 
      : bufSize_(size) {
    semaph.reset(new BidirectionalSemaphore(semid));

    shmid_ = shmid;

    if ( (shmptr_ = shmat(shmid_, 0, 0)) == (void *) -1) {
      logger().debug("ShMemServe: Could not attach shared memory.");
      throw ProcessException("Could not attach shared memory.");
    }

    closed_ = false;

    return;
  }

  bool readArgs(size_t size, void *result, unsigned int secondsToWait) {
    // while not everything read:
    //      wait(1)
    size_t sizeLeft = size;
    char *destPtr = static_cast<char *>(result);
    char *srcPtr = static_cast<char *>(shmptr_);
    boost::posix_time::ptime maxTime(boost::posix_time::second_clock::local_time() + boost::posix_time::seconds(secondsToWait));
    while (sizeLeft > 0) {
      size_t currentSize = ((sizeLeft > bufSize_) ? bufSize_ : sizeLeft);
      if (not semaph->doOp(0, -1, secondsLeft(maxTime))) {
        return false;
      }
      memcpy(destPtr, srcPtr, currentSize);
    //      read()
    //      wait(1)
      if (not semaph->doOp(0, -1, secondsLeft(maxTime))) {
        return false;
      }
      sizeLeft = sizeLeft - currentSize;
      destPtr += currentSize;
      srcPtr += currentSize;
    }
    
    return true;
  }

  bool reply(void *result, size_t size, unsigned int secondsToWait) {
    // write reply
    // while not everything written:
    //      wait(0)
    size_t sizeLeft = size;
    char *destPtr = static_cast<char *>(shmptr_);
    char *srcPtr = static_cast<char *>(result);
    boost::posix_time::ptime maxTime(boost::posix_time::second_clock::local_time() + boost::posix_time::seconds(secondsToWait));
    while (sizeLeft > 0) {
      size_t currentSize = ((sizeLeft > bufSize_) ? bufSize_ : sizeLeft);
      if (not semaph->doOp(1, 0, secondsLeft(maxTime))) {
        return false;
      }
      memcpy(destPtr, srcPtr, currentSize);
    //      write()
    //      signal(2)
      if (not semaph->doOp(1, +2, secondsLeft(maxTime))) {
        return false;
      }
      sizeLeft = sizeLeft - currentSize;
      destPtr += currentSize;
      srcPtr += currentSize;
    }
    
    return true;
  }

  void close() {
    if (not closed_) {
      semaph->close();
      if (shmdt(shmptr_) != 0) {
        logger().debug("ShMemServe.close: Could not detach shared memory.");
        throw ProcessException("Could not detach shared memory.");
      }

      closed_ = true;
    }
  }

  void silentClose() {
    if (not closed_) {
      semaph->silentClose();
      if (shmdt(shmptr_) != 0) {
        logger().debug("ShMemServe.silentClose: Could not detach shared memory.");
        throw ProcessException("Could not detach shared memory.");
      }
    }
    closed_ = true;
  }

  ~ShMemServe() {
    if (not closed_) {
      this->close();
    }
  }
};

} // end anonymous namespace

bool timedMultiProcessExecuteWithDelay(std::vector<ProcessRequest> &requests,
      unsigned int maxSecondsToWait,
      long nanoPollTime, long nanoDelayTime=0L) {
	assert(nanoDelayTime >= 0L);
	assert(nanoPollTime > 0L);
  assert(nanoPollTime < 999999999L);
  assert(maxSecondsToWait > 0);

  boost::posix_time::ptime maxTime(boost::posix_time::second_clock::local_time() + boost::posix_time::seconds(maxSecondsToWait));

  struct timespec sleepTime;

  // launch processes
  int requestIndex = 0;
  std::vector<int> pendingRequests;
  std::vector<ProcessRequest>::iterator requestIt;
  for(requestIt = requests.begin(); requestIt != requests.end();
        ++requestIt) {

  	// if we must wait some delay time, and this is not the first request,
  	// wait that time
  	if ((nanoDelayTime != 0L) and (requestIt != requests.begin())) {
  	  struct timespec remainingTime;
      sleepTime.tv_sec = 0;
      sleepTime.tv_nsec = nanoDelayTime;

      bool sleepCompleted = false;

      while (not sleepCompleted) {
      	int result = nanosleep(&sleepTime, &remainingTime);

      	if (result == -1) {
          if (errno == EINTR) {
        		sleepTime.tv_sec = remainingTime.tv_sec;
        		sleepTime.tv_nsec = remainingTime.tv_nsec;
          } else {
            logger().debugStream() << "Error while delaying multi process execution: "
              << errno
              << log4cpp::CategoryStream::EOL;
            throw ProcessException("Process could not be executed with required delay.");
          }
      	} else {
      		sleepCompleted = true;
      	}
      }
  	}

    ProcessRequest &request = *requestIt;
    
    request.pid_ = processExecuteAsynchronous(request.processPath_, 
      request.argv_, request.env_,
      request.inputFilePath_, 
      request.outputAndErrorFilePath_, 
      request.workingDirectory_, false); // do not clear zombies

    pendingRequests.push_back(requestIndex);      

    requestIndex++;      
  }
  
  // wait for processes
  while (not pendingRequests.empty()) {
    sleepTime.tv_sec = 0;
    sleepTime.tv_nsec = nanoPollTime;
    
    nanosleep(&sleepTime, NULL);
    
    std::vector<int> stillPendingRequests;
    
    std::vector<int>::iterator pendingIt;
    for(pendingIt = pendingRequests.begin(); 
        pendingIt != pendingRequests.end(); ++pendingIt) {

      long currentPid = requests[*pendingIt].pid_;
      pid_t result = waitpid(currentPid, NULL, WNOHANG);

      if (result == -1) {
        if (errno == ECHILD) {
          logger().debugStream() << "Error while waiting for"
            << " timed child " << currentPid << " ECHILD."
            << log4cpp::CategoryStream::EOL;
          throw ProcessException("Process could not be waited.");
        } else {
          logger().debugStream() << "Error while waiting for"
            << " timed child " << currentPid
            << " Errno: " << errno << log4cpp::CategoryStream::EOL;
          throw ProcessException("Error in waiting process.");
        }
      } else if (result != 0) {
        // remove from set
        if (not clearWaitedChild(currentPid)) {
          logger().debugStream() << "Error when clearWaitedChild: "
            << currentPid << log4cpp::CategoryStream::EOL;
        }
        // do not insert pid in stillPendingRequests
      } else {
        stillPendingRequests.push_back(*pendingIt);
      }

    }
    
    
    pendingRequests = stillPendingRequests;
    
    // if current time exceeds deadline, kill process and throw
    // exception
    boost::posix_time::time_duration timeLeft
      = maxTime - boost::posix_time::second_clock::local_time();
    if (timeLeft.is_negative()) {
      
      for(pendingIt = pendingRequests.begin(); 
          pendingIt != pendingRequests.end(); ++pendingIt) {
        long currentPid = requests[*pendingIt].pid_;
        callKill(currentPid);
        clearZombies();
        logger().debugStream() << "Time exceeded for process "
          << currentPid << log4cpp::CategoryStream::EOL;
      }
      return false;
    }
  }

  // no pending requests, so everything is done 
  return true;

}


bool timedMultiProcessExecute(std::vector<ProcessRequest> &requests,
      unsigned int maxSecondsToWait,
      long nanoPollTime) {
  // execute without delay
	return timedMultiProcessExecuteWithDelay(requests, maxSecondsToWait, nanoPollTime, 0L);
}

bool timedProcessExecute(const std::string &processPath,
       unsigned int maxSecondsToWait,
       long nanoPollTime,
       const std::vector<std::string> &argv,
       const std::map<std::string,std::string> &env,
       const std::string &inputFilePath,
       const std::string &outputAndErrorFilePath,
       const std::string &workingDirectory) {

  assert(nanoPollTime > 0L);
  assert(nanoPollTime < 999999999L);
  assert(maxSecondsToWait > 0);

  boost::posix_time::ptime maxTime(boost::posix_time::second_clock::local_time() + boost::posix_time::seconds(maxSecondsToWait));

  struct timespec sleepTime;

  long pid = processExecuteAsynchronous(processPath, argv, env,
    inputFilePath, outputAndErrorFilePath, workingDirectory);

  // get deadline
  
  // go waiting msPollTime until deadline is reached
  
  for(;;) {
    sleepTime.tv_sec = 0;
    sleepTime.tv_nsec = nanoPollTime;
    
    nanosleep(&sleepTime, NULL);
    
    pid_t result = waitpid(pid, NULL, WNOHANG);
    if (result == -1) {
      if (errno == ECHILD) {
        logger().debugStream() << "Error while waiting for"
          << " timed child " << pid << " ECHILD."
          << log4cpp::CategoryStream::EOL;
        throw ProcessException("Process could not be waited.");
      } else {
        logger().debugStream() << "Error while waiting for"
          << " timed child " << pid
          << " Errno: " << errno << log4cpp::CategoryStream::EOL;
        throw ProcessException("Error in waiting process.");
      }
    } else if (result != 0) {
      // remove from set
      if (not clearWaitedChild(pid)) {
        logger().debugStream() << "Error when clearWaitedChild: "
          << pid << log4cpp::CategoryStream::EOL;
      }
      return true;
    }
    
    // if current time exceeds deadline, kill process and throw
    // exception
    boost::posix_time::time_duration timeLeft
      = maxTime - boost::posix_time::second_clock::local_time();
    if (timeLeft.is_negative()) {
      callKill(pid);
      clearZombies();
      logger().debugStream() << "Time exceeded for process "
        << pid << log4cpp::CategoryStream::EOL;
      return false;
    }
  }
  // should never get here
}
       
// caller owns args and results
void processExecuteIPC(const std::string &processPath,
       void *args, size_t argsSize,
       void *results, size_t resultsSize,
       unsigned int maxSecondsToWait,
       const std::vector<std::string> &argv,
       const std::map<std::string,std::string> &env,
       const std::string &inputFilePath,
       const std::string &outputAndErrorFilePath,
       const std::string &workingDirectory) {
        

  logger().infoStream() << "Preparing IPC execution of: "
    << processPath << " with args: ";
  for(vector<string>::const_iterator it = argv.begin();
        it != argv.end(); ++it) {
    logger().infoStream() << "'" << *it << "' ";
  }
  logger().infoStream() << log4cpp::CategoryStream::EOL;

  // prepare shmem
  try {
    ShMemCall shmem;
  
    std::vector<std::string> newArgs;
    if (not argv.empty()) {
      newArgs.push_back(argv[0]);
    } else {
      newArgs.push_back(processPath);
    }
    newArgs.push_back(boost::lexical_cast<string>(shmem.shmid_)); // shmid
    newArgs.push_back(boost::lexical_cast<string>(shmem.semaph->semid_)); // semid
    newArgs.push_back(boost::lexical_cast<string>(argsSize));
    newArgs.push_back(boost::lexical_cast<string>(resultsSize));
    if (not argv.empty()) {
      newArgs.insert(newArgs.end(), argv.begin()+1, argv.end());
    }

    // resources destroyed in child without calling destructors
    long pid = processExecuteAsynchronous(processPath,
         newArgs, env, inputFilePath, outputAndErrorFilePath,
         workingDirectory);
         
    if (pid == -1) {
      logger().debug("Auxiliary IPC process creation failed.");
      throw ProcessException("Auxiliary process creation failed.");
    }         

    if (not shmem.call(args, argsSize, results, resultsSize,
        maxSecondsToWait)) {
      callKill(pid);
      logger().debug("IPC Server call wait time expired.");
      throw ProcessException("Wait time expired.");
    }
    
    // always kill the child, just in case
    callKill(pid);
  } catch (ProcessException &ex) {
    throw ProcessException(string("Failure when clientExecuteIPC.")
      + ex.getMessage());
  }  
}

// caller owns args and results data
void clientExecuteIPC(int shmid, int semid, 
            void *args, size_t argsSize,
            void *results, size_t resultsSize, 
            unsigned int secondsToWaitPerCall,
            boost::function<void (void)> callbackFunction) {
//  boost::shared_array<char> args(new char[argsSize]);
//  boost::shared_array<char> results(new char[resultsSize]);
  
  try {
    ShMemServe cliShmem(shmid, semid);
    if (not cliShmem.readArgs(argsSize, args,  
      secondsToWaitPerCall)) {
      logger().debug("IPC: Client read wait time expired.");
      throw ProcessException("Wait time expired.");
    }
  
    callbackFunction();
  
    if (not cliShmem.reply(results, resultsSize, 
      secondsToWaitPerCall)) {
      logger().debug("IPC: Client reply wait time expired.");
      throw ProcessException("Wait time expired.");        
    }
  } catch (ProcessException &ex) {
    throw ProcessException(string("Failure when clientExecuteIPC.")
      + ex.getMessage());
  }
}

#endif
// end of UNIX specific version
