/*! \file azucar/utils/propertyDBXMLReader.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include "azucar/utils/propertyDBXMLReader.hpp"
#include "azucar/utils/propertyDBMap.hpp"

#include <utility>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/cast.hpp>
#include <boost/algorithm/string.hpp>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMErrorHandler.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/sax/HandlerBase.hpp>

#include "azucar/utils/exceptions.hpp"
#include "azucar/utils/baseXMLException.hpp"
#include "azucar/utils/xerces.hpp"
#include "azucar/utils/util.hpp"

XERCES_CPP_NAMESPACE_USE


namespace {
	class MyErrorHandler : public HandlerBase {
	public:
			void error(const SAXParseException&);
			virtual ~MyErrorHandler() {};
	};

	void MyErrorHandler::error(const SAXParseException& exception)
	{
		const std::string message = azucar::utils::XMLChToString(exception.getMessage());
//			logger().errorStream() << "Fatal Error: " << message
//					 << " at line: " << exception.getLineNumber()
//					 << " column: " << exception.getColumnNumber()
//					 << log4cpp::CategoryStream::EOL;
	}

	const boost::shared_ptr<XercesDOMParser> createDOMParser(ErrorHandler *const handler) {
		const boost::shared_ptr<XercesDOMParser> parser(new XercesDOMParser());
		parser->setValidationScheme(XercesDOMParser::Val_Never);
		parser->setDoNamespaces(true);
		parser->setCreateCommentNodes(false);
		parser->setCreateEntityReferenceNodes(false);

		parser->setErrorHandler(handler);

		return parser;
	}

	bool hasSingleChild(const DOMNode* const elem) {
		return elem->getChildNodes()->getLength() == 1;
	}

	bool isTextNode(const DOMNode* const node) {
		return node->getNodeType() == DOMNode::TEXT_NODE;
	}

	bool hasSingleTextChild(const DOMNode* const elem) {
		if (not hasSingleChild(elem)) {
			return false;
		}

		return isTextNode(elem->getFirstChild());
	}

	const std::string getText(const DOMElement* const elem) {
		assert(hasSingleChild(elem));
		DOMNode *child = elem->getFirstChild();
		assert(isTextNode(child));

		return azucar::utils::XMLChToString(child->getNodeValue());
	}

	const std::pair<std::string, std::string> getPropertyAndValue(const DOMElement *elem) {
		const std::string propertyName = azucar::utils::XMLChToString(elem->getTagName());
		std::string value = getText(elem);
		boost::algorithm::trim(value);

		return std::make_pair(propertyName, value);
	}

}

namespace azucar {
namespace utils {
boost::shared_ptr<PropertyDB> readPropertyDBXML(const boost::filesystem::path &filepath) {
	// automatically closed on return
	azucar::utils::InputFile input(filepath);

	return readPropertyDBXML(input.stream);
}

boost::shared_ptr<PropertyDB> readPropertyDBXML(const std::string &filename) {
	return readPropertyDBXML(boost::filesystem::path(filename));
}
boost::shared_ptr<PropertyDB> readPropertyDBXML(std::istream &input) {
	// create parser
	const boost::scoped_ptr<ErrorHandler> errorHandler(new MyErrorHandler());

	const boost::shared_ptr<XercesDOMParser> parser = createDOMParser(errorHandler.get());

	// parse document
	const StdIstreamInputSource inputSource(input);

	parser->parse(inputSource);

	DOMDocument *doc = parser->getDocument();
	DOMElement *root = doc->getDocumentElement();

	// check root
	const std::string rootTag = azucar::utils::XMLChToString(root->getTagName());

	assert(rootTag == "azucar_utils_propertyDB");

	// check well-formedness of children

	// create PropertyDB with elements
	// check validity of PropertyDB
	boost::shared_ptr<PropertyDB> propDB(new PropertyDBMap());

	DOMNode* child;
  for (child = root->getFirstChild(); child != 0; child=child->getNextSibling()) {
  	if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
  		DOMElement* elem = boost::polymorphic_downcast<DOMElement*>(child);

  		if (not hasSingleTextChild(elem)) {
  			throw BaseException("Properties can only have a text value. Make sure "
  					"you are not using nested properties.");
  		}
  		const std::pair<std::string, std::string> property = getPropertyAndValue(elem);

  		// check property is not repeated
  		if (propDB->has(property.first)) {
  			throw BaseException(std::string("Repetead property: ") + property.first + ".");
  		} else {
  			propDB->set(property.first, property.second);
  		}

  	} else {
  		if (isTextNode(child)) {
  			std::string text = azucar::utils::XMLChToString(child->getNodeValue());
  			boost::algorithm::trim(text);
  			if (text.empty()) {
  				continue;
  			} else {
  	  		throw BaseException(std::string("Expected property tag, found: ")
  	  			+ " non empty text.");
  			}
  		} else {
  			throw BaseException(std::string("Expected property tag, found: ")
  				+ azucar::utils::XMLChToString(child->getNodeName()) + ".");
  		}
  	}
  }


	// clean-up

	// return PropertyDB

	return propDB;
}

}
}
