/*! \file azucar/utils/quantity.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include "azucar/utils/quantity.hpp"

#include <map>
#include <cassert>
#include <boost/assign/list_of.hpp>

namespace quantity {

const Unit &unit(const std::string &symbol) {
  static std::map<std::string, Unit> table;
  static bool tableSet = false;
  
  if (not tableSet) {
    tableSet = true;
    table = boost::assign::map_list_of
      (DIMENSIONLESS.symbol_, DIMENSIONLESS)
      (CELSIUS.symbol_, CELSIUS)
      (KELVIN.symbol_, KELVIN)
      (JOULE.symbol_, JOULE)
      (KJ_PER_MOLE.symbol_, KJ_PER_MOLE)
      (KJ_PER_MOLE_PER_E2.symbol_, KJ_PER_MOLE_PER_E2)
      (KJ_PER_MOLE_SQ.symbol_, KJ_PER_MOLE_SQ)
      (KCAL_PER_MOLE.symbol_, KCAL_PER_MOLE)
      (KCAL_PER_MOLE_SQ.symbol_, KCAL_PER_MOLE_SQ)
      (MOLAR.symbol_, MOLAR)
      (K_T.symbol_, K_T)
      (K_T_PER_E.symbol_, K_T_PER_E)
      (K_T_PER_E2.symbol_, K_T_PER_E2)
      (E_CHARGE.symbol_, E_CHARGE)
      (COULOMB.symbol_, COULOMB)
      (ANGSTROM.symbol_, ANGSTROM)
      (KCAL_PER_MOLE_KELVIN.symbol_, KCAL_PER_MOLE_KELVIN)
      (JOULE_PER_MOLE_KELVIN.symbol_, JOULE_PER_MOLE_KELVIN)
      (JOULE_PER_KELVIN.symbol_, JOULE_PER_KELVIN)
      (KJ_PER_MOLE_KELVIN.symbol_, KJ_PER_MOLE_KELVIN)
      (FARAD_PER_METER.symbol_, FARAD_PER_METER)
      (METER_PER_FARAD.symbol_, METER_PER_FARAD)
      (MOLE_TO_MINUS_ONE.symbol_, MOLE_TO_MINUS_ONE)
      (MEAD_POTENTIAL_UNIT.symbol_, MEAD_POTENTIAL_UNIT);
  }
  
  return table[symbol];
}

Unit &Unit::transform(const Unit &newUnit) {
	this->symbol_ = newUnit.symbol_;

	return *this;
}


Quantity &Quantity::operator=(const Quantity &other) {
  if (this != &other) {
    this->value_ = other.value_;
    assert (this->unit_ == other.unit_);
  }
  
  return *this;
}

Quantity &Quantity::transform(const Quantity &newValue) {
	this->value_ = newValue.value_;
	this->unit_.transform(newValue.unit_);

	return *this;
}



namespace conversion {
const Quantity to_Kelvin_scale(const Quantity &qInCelsius) {
  assert(qInCelsius.unit_ == CELSIUS);
  
  return Quantity(qInCelsius.value_ 
                  - conversion::ABSOLUTE_ZERO_IN_CELSIUS.value_, 
                  KELVIN);
} 

const Quantity to_Celsius_scale(const Quantity &qInKelvin) {
  assert(qInKelvin.unit_ == KELVIN);
  
  return Quantity(qInKelvin.value_ 
                  + conversion::ABSOLUTE_ZERO_IN_CELSIUS.value_, 
                  CELSIUS);
} 

} // end namespace conversion

} // end namespace quantity


namespace std {
  ostream& operator<< (ostream &out, const quantity::Unit &u) {
    return (out << u.symbol_);
  }

  ostream& operator<< (ostream &out, const quantity::Quantity &q) {
    return (out << q.value_ << " (" << q.unit_ << ") ");
  }

}


