/*! \file azucar/utils/testUtils.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include "azucar/utils/testUtils.hpp"

#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include "azucar/utils/util.hpp"

using namespace std;

namespace {
const std::string category_name = "test_util_cat";

log4cpp::Category &logger() {
  static log4cpp::Category &theLogger =
    log4cpp::Category::getInstance(category_name);
  return theLogger;
}
}

namespace test {

TestSuiteVisitorPrinter::TestSuiteVisitorPrinter() : suitePath_() {}

void TestSuiteVisitorPrinter::visit( boost::unit_test::test_case const& tu ) {
	// join all the suite path components to create a text qualified suite name
	// separated by double colons.
	string suitePathString = boost::algorithm::join(suitePath_, "::");

	// Print the qualified test name (that is, the qualified suite name + test name)
	// to std out.
	cout << suitePathString << "::" << tu.p_name << endl;
}

bool    TestSuiteVisitorPrinter::test_suite_start( boost::unit_test::test_suite const& ts)   {
	// Add the suite name to the suite path (qualified suite name)
	suitePath_.push_back(ts.p_name);
	return true;
}

void    TestSuiteVisitorPrinter::test_suite_finish( boost::unit_test::test_suite const& )  {
	// Remove the suite name from the suite path (qualified suite name)
	suitePath_.pop_back();
}


bool compareTextFiles(const string &file1, const string &file2,
    const string &excluded_line_start) {
  vector<string> excluded_line_starts;
  excluded_line_starts.push_back(excluded_line_start);
  return compareTextFiles(file1, file2, excluded_line_starts);
}

bool compareTextFiles(const string &file1, const string &file2,
    const vector<string> &excluded_line_starts) {
  ifstream input1(file1.c_str());
  ifstream input2(file2.c_str());

  bool checkStartOfLine = not excluded_line_starts.empty();

  while (not azucar::utils::eof(input1) and not azucar::utils::eof(input2)) {
    string currentLine1;
    string currentLine2;
    getline(input1, currentLine1);
    getline(input2, currentLine2);
    if (checkStartOfLine) {
      vector<string>::const_iterator vecit;
      bool matchFound = false;
      for(vecit = excluded_line_starts.begin();
          vecit != excluded_line_starts.end(); ++vecit) {
        if (boost::algorithm::contains(currentLine1, *vecit)) {
          matchFound = true;
          break;
        }
      }
      if (matchFound) {
        continue;
      }
    }

    if (currentLine1 != currentLine2) {
      return false;
    }
  }

  bool bothFilesCompleted = azucar::utils::eof(input1) and azucar::utils::eof(input2);
  input1.close();
  input2.close();
  return bothFilesCompleted;
}


bool compareTextFilesRegEx(const string &file1, const string &file2) {
  ifstream input1(file1.c_str());
  ifstream input2(file2.c_str());
  boost::regex dateExpr("[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}");
  boost::regex dateMonthExpr("[0-9]{4}-[a-zA-Z]{3}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}");
  boost::regex completeDateExpr("[0-9]{4}-[a-zA-Z]{3}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2} [^[:digit:]]*");
  boost::regex timeExpr("[0-9]{2}:[0-9]{2}:[0-9]{2}");
  boost::regex yearExpr("[0-9]{4}");
  boost::regex numberExpr("[0-9]+");
  const string tag_delimiter("###");
  const string date_tag("###DATE###");
  const string date_month_tag("###DATEMONTH###");
  const string complete_date_tag("###COMPLETEDATE###");
  const string year_tag("###YEAR###");
  const string number_tag("###NUMBER###");
  const string time_tag("###TIME###");

  while (not azucar::utils::eof(input1) and not azucar::utils::eof(input2)) {
    string currentLine1;
    string currentLine2;
    getline(input1, currentLine1);
    getline(input2, currentLine2);

    // while find $$$:
    string::size_type position = currentLine1.find(tag_delimiter);

    while (position != string::npos) {
      // compare first part
      string beginLine1 = currentLine1.substr(0, position);
      string beginLine2 = currentLine2.substr(0, position);
      if (beginLine1 != beginLine2) {
        return false;
      }
      // get same position in file2
      currentLine1 = currentLine1.substr(position);
      currentLine2 = currentLine2.substr(position);
      boost::match_results<std::string::const_iterator> theMatch;
      std::string::const_iterator endLine2It = currentLine2.end();
      if (boost::algorithm::starts_with(currentLine1, date_tag)) {
        // match DATE
        if (boost::regex_search(currentLine2, theMatch,
                                dateExpr, boost::match_continuous)) {
          currentLine1 = currentLine1.substr(date_tag.size());
          currentLine2 = string(theMatch[0].second, endLine2It);
        } else {
          logger().errorStream() << "Date tag do not match"
            << currentLine1 << log4cpp::CategoryStream::EOL
            << currentLine2 << log4cpp::CategoryStream::EOL;
          return false;
        }
      } else if (boost::algorithm::starts_with(currentLine1, date_month_tag)) {
        // match DATE
        if (boost::regex_search(currentLine2, theMatch,
                                dateMonthExpr, boost::match_continuous)) {
          currentLine1 = currentLine1.substr(date_month_tag.size());
          currentLine2 = string(theMatch[0].second, endLine2It);
        } else {
          logger().errorStream() << "DateMonth tag do not match"
            << currentLine1 << log4cpp::CategoryStream::EOL
            << currentLine2 << log4cpp::CategoryStream::EOL;
          return false;
        }
      } else if (boost::algorithm::starts_with(currentLine1, complete_date_tag)) {
        // match COMPLETE DATE
        if (boost::regex_search(currentLine2, theMatch,
                                completeDateExpr, boost::match_continuous)) {
          currentLine1 = currentLine1.substr(complete_date_tag.size());
          currentLine2 = string(theMatch[0].second, endLine2It);
        } else {
          logger().errorStream() << "CompleteDate tag do not match"
            << currentLine1 << log4cpp::CategoryStream::EOL
            << currentLine2 << log4cpp::CategoryStream::EOL;
          return false;
        }
      } else if (boost::algorithm::starts_with(currentLine1, time_tag)) {
        if (boost::regex_search(currentLine2, theMatch,
                                timeExpr, boost::match_continuous)) {
          currentLine1 = currentLine1.substr(time_tag.size());
          currentLine2 = string(theMatch[0].second, endLine2It);
        } else {
          logger().errorStream() << "Time tag do not match"
            << currentLine1 << log4cpp::CategoryStream::EOL
            << currentLine2 << log4cpp::CategoryStream::EOL;
          return false;
        }

      } else if (boost::algorithm::starts_with(currentLine1, year_tag)) {
        // match year
        if (boost::regex_search(currentLine2, theMatch,
                                yearExpr, boost::match_continuous)) {

          currentLine1 = currentLine1.substr(year_tag.size());
          currentLine2 = string(theMatch[0].second, endLine2It);
        } else {
          logger().errorStream() << "Year tag do not match"
            << currentLine1 << log4cpp::CategoryStream::EOL
            << currentLine2 << log4cpp::CategoryStream::EOL;
          return false;
        }
      } else if (boost::algorithm::starts_with(currentLine1, number_tag)) {
        // match number
        if (boost::regex_search(currentLine2, theMatch,
                                numberExpr, boost::match_continuous)) {
          currentLine1 = currentLine1.substr(number_tag.size());
          currentLine2 = string(theMatch[0].second, endLine2It);
        } else {
          logger().errorStream() << "Number tag do not match"
            << currentLine1 << log4cpp::CategoryStream::EOL
            << currentLine2 << log4cpp::CategoryStream::EOL;
          return false;
        }
      } else {
        assert("Search template unknown");
      }

      position = currentLine1.find(tag_delimiter);
    }

    if (currentLine1 != currentLine2) {
    // return false if there is no match
      logger().errorStream() << "Rest of lines do not match: "
        << currentLine1 << log4cpp::CategoryStream::EOL
        << currentLine2 << log4cpp::CategoryStream::EOL;
      return false;
    }

  }

  bool bothFilesCompleted = azucar::utils::eof(input1) and azucar::utils::eof(input2);
  input1.close();
  input2.close();
  return bothFilesCompleted;
}


bool compareFiles(const string &file1, const string &file2) {
  char buffer1[512];
  char buffer2[512];

  ifstream input1(file1.c_str(), ios_base::in | ios_base::binary);
  ifstream input2(file2.c_str(), ios_base::in | ios_base::binary);

  while (not azucar::utils::eof(input1) and not azucar::utils::eof(input2)) {
    input1.read(buffer1, 512);
    int read1 = input1.gcount();
    input2.read(buffer2, 512);
    int read2 = input2.gcount();
    if (read1 != read2) {
      return false;
    }
    if (read1 > 0) {
      for(int i=0; i < read1; ++i) {
        if (buffer1[i] != buffer2[i]) {
          return false;
        }
      }
    }
  }

  bool bothFilesCompleted = azucar::utils::eof(input1) and azucar::utils::eof(input2);
  input1.close();
  input2.close();
  return bothFilesCompleted;
}

}
