/*! \file azucar/utils/util.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

// uses <cstdlib> from util.hpp for mkdtemp()


#include <sstream>
#include <unistd.h>
#include <list>

#include <boost/thread/thread.hpp>
#include <boost/thread/xtime.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/regex.hpp>
#include <boost/shared_array.hpp>
#include <boost/filesystem.hpp>

#include "azucar/utils/util.hpp"

using namespace std;

namespace azucar {
namespace utils {

const std::string 
trimSubstr(const std::string &source, std::string::size_type off,
          std::string::size_type count) {
  string substr = source.substr(off, count);
  boost::algorithm::trim(substr);
  return substr;
}


bool eof(istream &input) {
  // this first check is needed, in order to avoid 
  // internal temporal badbits in the input.peek() function.
  if (input.eof()) { 
      return true;
  }
  
  istream::int_type peekC = input.peek();
  return (peekC == istream::traits_type::eof());
}

InputFile::InputFile(const boost::filesystem::path &filePath) {
  this->stream.open(filePath.string().c_str());
  this->stream.exceptions(ios_base::badbit | ios_base::failbit);
}

InputFile::InputFile(const std::string &fileName) {
  this->stream.open(fileName.c_str());
  this->stream.exceptions(ios_base::badbit | ios_base::failbit);
}

InputFile::InputFile(const char *filename) {
  this->stream.open(filename);
  this->stream.exceptions(ios_base::badbit | ios_base::failbit);
}

void InputFile::skipLine() {
  string dummyLine;
  getline(this->stream, dummyLine);
  return;
}

void InputFile::skipLines(unsigned int n) {
  for(unsigned int i = 0; i < n; ++i) {
    skipLine();
  }
}

bool InputFile::nextLineIsEmpty(std::string &nextLine) {
  getline(this->stream, nextLine);
  return (boost::empty(
            boost::find_token(nextLine, 
                !(boost::algorithm::is_space()), 
                boost::algorithm::token_compress_off)));
}

void InputFile::close() {
  if (this->stream.is_open()) {
    this->stream.close();
  }
}

InputFile::~InputFile() {
  this->close();
}

OutputFile::OutputFile(const boost::filesystem::path &filePath,
      bool append) {
  ios_base::openmode mode = ios_base::out;
  if (append) {
    mode |= ios_base::app;
  }
  this->stream.open(filePath.string().c_str(), mode);
  this->stream.exceptions(ios_base::badbit | ios_base::failbit);
}

OutputFile::OutputFile(const std::string &fileName, bool append) {
  ios_base::openmode mode = ios_base::out;
  if (append) {
    mode |= ios_base::app;
  }
  this->stream.open(fileName.c_str(), mode);
  this->stream.exceptions(ios_base::badbit | ios_base::failbit);
}

OutputFile::OutputFile(const char *filename, bool append) {
  ios_base::openmode mode = ios_base::out;
  if (append) {
    mode |= ios_base::app;
  }
  this->stream.open(filename, mode);
  this->stream.exceptions(ios_base::badbit | ios_base::failbit);
}

void OutputFile::close() {
  if (this->stream.is_open()) {
    this->stream.close();
  }
}

OutputFile::~OutputFile() {
  this->close();
}


bool isInASCIIPrintableRange(const std::string &s) {
  std::string::const_iterator it;
  for(it = s.begin(); it != s.end(); ++it) {
    if ( ( (*it) < 32) or ( (*it) > 126)) {
      return false;
    }
  }
  return true;
}

bool isInASCIIPrintableRange(char c) {
  return ((c >= 32) or (c <= 126));
}

const std::string htmlBrk(const std::string &str) {
  ostringstream outputText;
  string::const_iterator it;
  for(it = str.begin(); it != str.end(); ++it) {
    outputText << *it;
    outputText << "<br>";
  }
  return outputText.str();
}



const string htmlEscape(const std::string &input) {
  string result;
  
  std::string::const_iterator it;
  for(it = input.begin(); it != input.end(); ++it) {
    switch(*it) {
      case '<': result.append("&lt;"); break;
      case '>': result.append("&gt;"); break;
      case '&': result.append("&amp;"); break;
      case '"': result.append("&quot;"); break;
      case '\n': result.append("<br>"); break;
      default: {
          if (not isInASCIIPrintableRange(*it)) {
            result.append("?");
          } else {
            result.append(1, *it);
          }
          break;
        }
    }
  }
  
  return result;
}

bool hasEmailAddressFormat(const std::string &address) {
  string::size_type atPos = address.find('@');
  if (atPos == string::npos) {
    return false;
  }
  
  string localPart = address.substr(0, atPos);
  string domainPart = address.substr(atPos + 1);
  
  boost::algorithm::trim(localPart);
  boost::algorithm::trim(domainPart);

  
  
  if ( (localPart.size() > 64) or (domainPart.size() > 255)
    or localPart.empty() or domainPart.empty() ) {
    return false;
  }


  string allowedChar = "[[:alnum:]!#\\$%&'*\\+\\-/=\\?\\^_`\\{\\|\\}~]";
  boost::regex expression(allowedChar + "+(\\." + allowedChar + "+)*");

  if (not boost::regex_match(localPart, expression)) {
    return false;
  } 
  
  if (not boost::regex_match(domainPart, expression)) {
    return false;
  }
  
  return true;
}

const std::string createTmpDirectory(const std::string &nameTemplate) {
  assert(boost::algorithm::ends_with(nameTemplate, "XXXXXX"));
  string::size_type length = nameTemplate.size();
  
  boost::shared_array<char> cstringPtr(new char[length+1]);
  cstringPtr[length] = '\0';
  if (nameTemplate.copy(cstringPtr.get(), length) != length) {
    throw azucar::utils::BaseException("Could not create directory.");
  }
  char *cresult = mkdtemp(cstringPtr.get());
  if (cresult == NULL) {
    // should never come here, since XXXXXX are at the end of the string
    assert(false and "mkdtemp returned NULL.");
  }
  string result(cstringPtr.get()); // create the result (copy from the C-string)
  return result;
}

  TempFile::TempFile(const std::string &nameTemplate) {
    assert(boost::algorithm::ends_with(nameTemplate, "XXXXXX"));
    string::size_type length = nameTemplate.size();
    
    boost::shared_array<char> cstringPtr(new char[length+1]);
    cstringPtr[length] = '\0';
    if (nameTemplate.copy(cstringPtr.get(), length) != length) {
      throw azucar::utils::BaseException("Could not create directory.");
    }
    int tempFd = mkstemp(cstringPtr.get());
    if (tempFd == -1) {
      throw azucar::utils::BaseException("mkstemp failed.");
    }

    if (close(tempFd) == -1) {
      throw azucar::utils::BaseException("creation of temporary file failed");
    }
  
    this->path_ = boost::filesystem::path(std::string(cstringPtr.get()));
    
    return;
  }
  
  
  const boost::filesystem::path &TempFile::getPath() const {
    return this->path_;
  }
  
  void TempFile::getInputStream(std::ifstream &in) {
    assert(not in.is_open());
    
    in.open(this->path_.string().c_str());
    
    return;
  }
  
  
  void TempFile::getOutputStream(std::ofstream &out) {
    assert(not out.is_open());
    
    out.open(this->path_.string().c_str());
    
    return;
  }
  
  
  TempFile::~TempFile() {
    try {
      boost::filesystem::remove(this->path_);
    } catch (boost::filesystem::basic_filesystem_error<boost::filesystem::path> &ex) {
      // do nothing; should log the error.
    }
  }

namespace {
//  using boost::filesystem::dot;
//  using boost::filesystem::slash;

  // Cache system-dependent dot, double-dot and slash strings

  // these definitions do not work in latest versions of Boost
//  const std::string _dot  = std::string(1, dot<path>::value);
//  const std::string _dots = std::string(2, dot<path>::value);
//  const std::string _sep = std::string(1, slash<path>::value);

  const std::string _dot  = std::string(".");
  const std::string _dots = std::string("..");
  const std::string _sep = std::string("/");
}

const boost::filesystem::path
normalizePath(const boost::filesystem::path p) {
	if (p.empty()) {
    throw azucar::utils::BaseException("path empty; couldn't normalize path");
	}

	// iterate through all path components.
	// skip _dot
	// remove previous component when finding _dots (unless there is no component to remove)
	// store all separators, since they may be important.
	std::list<std::string> pathList;

	for(boost::filesystem::path::iterator path_it = p.begin(); path_it != p.end(); ++path_it) {
		if (*path_it == _dot) {
			continue;
		} else if (*path_it == _dots) {
			if (not pathList.empty()) {
				pathList.pop_back();
			} else {
				pathList.push_back(*path_it);
			}
		} else {
			pathList.push_back(*path_it);
		}
	}

	boost::filesystem::path result;

	// if the pathList starts with the root directory, keep it, and go on processing
	// with the rest.
	if (not pathList.empty()) {
		if (pathList.front() == _sep) {
			result = boost::filesystem::path(_sep);
			pathList.pop_front();
		}
	}

	// if, after looking for a possible root directory at the beginning, there is
	// a _sep at the end, it means it must stay in the path, so transform it to "",
	// so that usual combination with /= adds it at the end.
	if (not pathList.empty()) {
		if (pathList.back() == _sep) {
			pathList.pop_back();
			pathList.push_back(""); // to make sure that a separator appears at the end.
		}

		std::list<std::string>::iterator it;
		for(it = pathList.begin(); it != pathList.end(); ++it) {
			if (*it == _sep) {
				continue; // skip middle separators
			}
			result /= *it;
		}

	}


	if (result.empty()) {
		result = boost::filesystem::path(_dot);
	}


	return result;
}

const boost::filesystem::path
relativePath(const boost::filesystem::path p, const boost::filesystem::path base) {
  using boost::filesystem::path;

  // check for emptiness
  if ((p.empty()) || (base.empty()))
      throw azucar::utils::BaseException("path or base was empty; couldn't generate relative path");


  // do not assert base is a directory, since it may happen that neither base nor p
  // actually exist.

  // convert p and base to complete normalized paths
  boost::filesystem::path pComplete = normalizePath(boost::filesystem::system_complete(p));
  boost::filesystem::path baseComplete = normalizePath(boost::filesystem::system_complete(base));

  if (pComplete == baseComplete)
      return _dot;
      /*!! this breaks stuff if path is a filename rather than a directory,
           which it most likely is... but then base shouldn't be a filename so... */

  // break into root and relative path
  if (pComplete.root_path() != baseComplete.root_path()) {
  	// root paths do not match, so pComplete (the system absolute path) is the right
  	// way to get to p from base.
  	return pComplete;
  }

  // now, only relative paths are needed.
  boost::filesystem::path pRelative = pComplete.relative_path();
  boost::filesystem::path baseRelative = baseComplete.relative_path();


  boost::filesystem::path output;

  boost::filesystem::path::iterator path_it = pRelative.begin(),    path_end = pRelative.end();
  boost::filesystem::path::iterator base_it = baseRelative.begin(), base_end = baseRelative.end();



  // iterate over path and base
  while (true) {

      // compare all elements so far of path and base to find greatest common root;
      // when elements of path and base differ, or run out:
      if ((path_it == path_end) || (base_it == base_end) || (*path_it != *base_it)) {

          // write to output, ../ times the number of remaining elements in base;
          // this is how far we've had to come down the tree from base to get to the common root
          for (; base_it != base_end; ++base_it) {
              if (*base_it == _dot) {
                  assert(false);
              }

              if (*base_it == _dots) {
              	assert(false);
              }

              output /= _dots;
          }

          // write to output, the remaining elements in path;
          // this is the path relative from the common root
          boost::filesystem::path::iterator path_it_start = path_it;
          for (; path_it != path_end; ++path_it) {

              if (*path_it == _dot)
                  continue;

              output /= *path_it;
          }

          break;
      }

      ++path_it, ++base_it;
  }

  return output;

}


void threadSleep(unsigned int seconds) {
  boost::xtime sleepTime;
  boost::xtime_get(&sleepTime, boost::TIME_UTC);
  sleepTime.sec += seconds;
  boost::thread::sleep(sleepTime);
}

GlobalRandomGeneratorType globalRandomGenerator;

UniformIntRandomGenerator createUniformIntRandomGenerator(int min, int max) {
  return UniformIntRandomGenerator(globalRandomGenerator,
            std::tr1::uniform_int<>(min, max));
}

UniformLongRandomGenerator createUniformLongRandomGenerator(long min, long max) {
  return UniformLongRandomGenerator(globalRandomGenerator,
            std::tr1::uniform_int<long>(min, max));
}


UniformRealRandomGenerator createUniformRealRandomGenerator(double min, double max) {
  return UniformRealRandomGenerator(globalRandomGenerator,
            std::tr1::uniform_real<>(min, max));
}    


const boost::filesystem::path getSelfPath() {
  const int bufSize = 2048;
  char programName[bufSize];
  
  int programNameSize = readlink("/proc/self/exe", programName, bufSize);

  if (programNameSize == -1) {
    throw new azucar::utils::BaseException("Could not obtain self path.");
  }
  
  std::string programStrName(programName, programNameSize);
  
  return boost::filesystem::path(programStrName);

}

ManagedArgv::ManagedArgv(int size) {
  args_.reset(new char*[size]);
  char **data = args_.get();
  for(int i=0; i < size; ++i) {
    data[i] = 0;
  }
  size_ = size;
  currentIndex_ = 0;
}

void ManagedArgv::append(const std::string &newArg) {
  assert(currentIndex_ < size_);

  const char* newArgStr = newArg.c_str();
  
  args_[currentIndex_] = new char[strlen(newArgStr)+1];
  strcpy(args_[currentIndex_], newArgStr);
  
  currentIndex_++;     
}

char **ManagedArgv::getData() {
  return args_.get();
}
  
ManagedArgv::~ManagedArgv() {
  for(int i=0; i < size_; ++i) {
    delete[] args_[i];
  }
}


ProgramOptionsResult 
processProgramOptions(po::variables_map &vm,
    const CheckOptionsFunc &checkF,
    const UsageFunc &usageF,
    const std::string &version_string,
    int argc, char **argv,
    const po::options_description &specificProgramOptions) {

  return processProgramOptions(vm, checkF, usageF, version_string,
  		argc, argv, specificProgramOptions,
  		po::positional_options_description());

}

ProgramOptionsResult
processProgramOptions(po::variables_map &vm,
    const CheckOptionsFunc &checkF,
    const UsageFunc &usageF,
    const std::string &version_string,
    int argc, char **argv,
    const po::options_description &namedProgramOptions,
    const po::positional_options_description &positionalOptions) {

  po::options_description desc("Generic options");

  desc.add_options()
    ("help", "produce help message")
    ("version", "program version")
    ;
    
  desc.add(namedProgramOptions);

  try {
    po::store(po::command_line_parser(argc, argv)
              .options(desc)
              .positional(positionalOptions).run(), vm);
    po::notify(vm);
  } catch (po::error &ex) {
    cerr << ex.what() << endl;
    usageF();
    cerr << desc << endl;
    return ERROR_EXECUTION;
  }
  
  if (vm.count("help")) {
    usageF();
    cout << desc << endl;
    return INFO_EXECUTION;
  }  
  
  if (vm.count("version")) {
    cout << "Version: " << version_string << endl;
    return INFO_EXECUTION;
  }
  
  if (not checkF(vm)) {
    usageF();
    cerr << desc << endl;
    return ERROR_EXECUTION;
  }
  
  return NORMAL_EXECUTION;

}


}
}
        
