/*! \file azucar/utils/xerces.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include "azucar/utils/xerces.hpp"

#include <cassert>

#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>

XERCES_CPP_NAMESPACE_USE


// private namespace
namespace {

/*! \brief Manage a Xerces-C++ string.
 *
 * This object takes care of releasing the assigned memory when
 * requested, or under destruction.
 *
 * Operation involves:
 * 1. Create a ManagedString object.
 * ManagedString myString;
 * 2. Pass the Xerces string to manage.
 * myString.manage(sourceStr);
 * 3. Convert to standard string.
 * const std::string myStdString = myString.toString();
 * 4. Release resources.
 * myString.release();
 *
 */
class ManagedString {
private:
	char* theChar;
	std::string theString;
public:
	/*! \brief Create an empty ManagedString.
	 *
	 */
	ManagedString(): theChar(0), theString() {};

	/*! \brief Manage the provided Xerces-C++ string.
	 *
	 * \pre No other Xerces-C++ string is currently managed by this object.
	 * \post The managed string will be released at destruction if no previous call
	 * to release is done before.
	 */
	void manage(const XMLCh *const sourceStr) {
		assert(not theChar);
		theChar = XMLString::transcode(sourceStr);
		theString = std::string(theChar);
	}

	/*! \brief Get the std::string version of the managed string.
	 *
	 */
	const std::string toString() const {
		return theString;
	}

	/*! \brief Releases the managed string.
	 *
	 */
	void release() {
		XMLString::release(&theChar);
		theString.clear();
	}

	/*! \brief Destroy this object, and release the managed string if needed.
	 *
	 */
	~ManagedString() {
		if (not theChar) {
			release();
		}
	}

private:
	// forbid copy constructor and assignment
	ManagedString(const ManagedString &);
	ManagedString & operator=(const ManagedString&);

};

}

namespace azucar {
namespace utils {

XercesLibrary::XercesLibrary() : initialized(false) {};

void XercesLibrary::init() {
	assert(not initialized);

	XMLPlatformUtils::Initialize();
	initialized = true;
}

void XercesLibrary::terminate() {
	assert(initialized);

	XMLPlatformUtils::Terminate();
	initialized = false;
}

XercesLibrary::~XercesLibrary() {
	if (initialized) {
		terminate();
	}
}

const std::string XMLChToString(const XMLCh *const sourceStr) {
	ManagedString transcoded;
	transcoded.manage(sourceStr); // transcoded is released at return time

	return transcoded.toString();
}


StdIstreamInputStream::StdIstreamInputStream(std::istream &is) : is_(is) {};

StdIstreamInputStream::~StdIstreamInputStream() {}

#if _XERCES_VERSION >= 30000

XMLFilePos StdIstreamInputStream::curPos () const {
	return static_cast<XMLFilePos> (is_.tellg ());
}

#else

unsigned int StdIstreamInputStream::curPos () const {
	return static_cast<unsigned int> (is_.tellg ());
}

#endif


#if _XERCES_VERSION >= 30000
 XMLSize_t
 StdIstreamInputStream::readBytes (XMLByte* const buf, const XMLSize_t size)
#else
 unsigned int
 StdIstreamInputStream::readBytes (XMLByte* const buf, const unsigned int size)
#endif
 {
	 // Some implementations don't clear gcount if you
	 // call read() on a stream that is in the eof state.
	 //
	 if (is_.eof ())
		 return 0;

	 // Unset the exception failbit while we are working
	 // with the stream.
	 //
	 std::ios_base::iostate old (is_.exceptions ());
	 is_.exceptions (old & ~std::ios_base::failbit);

	 is_.read (reinterpret_cast<char*> (buf),
			 static_cast<std::streamsize> (size));

	 // Clear the fail bit if it was caused by eof and restore
	 // the original exception state. If there are any pending
	 // errors then the exception will be thrown now.
	 //
	 if (is_.fail () && is_.eof ())
		 is_.clear (is_.rdstate () & ~std::ios_base::failbit);

	 is_.exceptions (old);

	 // Make sure that if we failed, readBytes won't be called
	 // again.
	 //
	 if (!is_.fail ()) {
#if _XERCES_VERSION >= 30000
		 return static_cast<XMLSize_t> (is_.gcount ());
#else
		 return static_cast<unsigned int> (is_.gcount ());
#endif
	 } else {
		 return 0;
   }
 }


#if _XERCES_VERSION >= 30000

const XMLCh* StdIstreamInputStream::getContentType () const {
	return 0;
}

#endif


StdIstreamInputSource::StdIstreamInputSource(std::istream &is) : is_(&is) {};

StdIstreamInputSource::~StdIstreamInputSource() { is_ = 0; }

BinInputStream* StdIstreamInputSource::makeStream () const
{
	assert(is_);

  std::istream& is (*is_);

  is_ = 0;

  return new StdIstreamInputStream(is);
}



}
}
