/*! \file azucar/utils/xmlInit.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <log4cpp/Category.hh>

#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/PlatformUtils.hpp>

#include "azucar/utils/xmlInit.hpp"
#include "azucar/utils/init.hpp"
#include "azucar/utils/exceptions.hpp"
#include "azucar/utils/baseXMLException.hpp"

using namespace std;

using azucar::utils::Initializer;

namespace {
  XERCES_CPP_NAMESPACE_USE

  const std::string category_name = "xmlInit_cat";

  log4cpp::Category &logger() {
    static log4cpp::Category &theLogger =   
      log4cpp::Category::getInstance(category_name);
    return theLogger;
  }

  /*! \brief Custom PanicHandler for XML fatal situations.
   * 
   * This handler logs the problem, and throws a BaseException.
   * 
   * Further use of XML library is an error.
   * 
   * \exception azucar::utils::BaseException Thrown always (as mandated
   *  by PanicHandler documentation).
   */
  class CustomXMLPanicHandler : public PanicHandler {
  public:
    void panic(const PanicHandler::PanicReasons reason) {
      // log error
      logger().errorStream() << "XML Panic: "
        << getPanicReasonString(reason)
        << log4cpp::CategoryStream::EOL;
      // throw custom exception
      throw azucar::utils::BaseXMLException("Panic XML");
    }
  };

  
  CustomXMLPanicHandler *customPanicHandler; //!< Ptr to 
  //!< the custom panic handler used by the XML module.
  //!< It is assigned during module initialization, and
  //!< deleted during finalization.
  
  /*! \brief Initialize Xerces XML Module.
   * 
   * If state of library is Initializer::NON_INITIALIZED,
   * allocate a new CustomXMLPanicHandler object,
   * and use it for initialization of XML library.
   * New state is Initializer::INITIALIZED.
   * 
   * In case of panic of the XML library, 
   * a azucar::utils::BaseXMLException is thrown during usage of the
   * XML service. Further use of XML library is an error.
   * 
   * \exception azucar::utils::BaseXMLException If an error happens during
   *  XML library initialization.
   */
  void init(Initializer& initObj) {
    if ((initObj.status == Initializer::NON_INITIALIZED) 
        or (initObj.status == Initializer::DESTROYED)) {
      customPanicHandler = new CustomXMLPanicHandler(); 
      XMLPlatformUtils::Initialize(XMLUni::fgXercescDefaultLocale,
        0, customPanicHandler);
      initObj.status = Initializer::INITIALIZED;
    }
  }
  
  /*! \brief Finalize Xerces XML Module.
   * 
   * If state of library is Initializer::INITIALIZED,
   * finalize XML library and deallocate CustomXMLPanicHandler object.
   * New state is Initializer::DESTROYED.
   */
  void destroy(Initializer& initObj) {
    if (initObj.status == Initializer::INITIALIZED) {
      XMLPlatformUtils::Terminate();
      delete customPanicHandler;
      initObj.status = Initializer::DESTROYED;
    }
  }
  
};

Initializer xmlInitializer(&init, &destroy);
