/*! \file test/azucar/utils/logger/test_logging.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <cassert>
#include <string>
#include <fstream>

#include <boost/test/unit_test.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include <azucar/utils/util.hpp>
#include <azucar/utils/logger/util.hpp>

//#include "test/test_logging.hpp"

using boost::unit_test::test_suite;
using boost::unit_test::test_case;
using namespace std;

BOOST_AUTO_TEST_SUITE(test_logging)

  const string property_file = "scratch/test_log.conf";
	const string new_property_file = "scratch/test_log_new.conf";
  const string logging_file = "scratch/test_logging.log";
  const string main_cat_str = "main_cat";

BOOST_AUTO_TEST_CASE(simple_test) {
    if (boost::filesystem::exists(logging_file)) {
      boost::filesystem::remove(logging_file);
    }
  

    // write configuration file
    ofstream log_conf(property_file.c_str());
    
    log_conf << "log4cpp.rootCategory=DEBUG, A1" << endl;
    log_conf << "log4cpp.appender.A1=log4cpp.RollingFileAppender"
      << endl;
    log_conf << "log4cpp.appender.A1.fileName=" << logging_file
      << endl;
    log_conf << "log4cpp.appender.A1.layout=log4cpp.PatternLayout"
      << endl;
    log_conf << "log4cpp.appender.A1.layout.ConversionPattern=%d{%a %d %b %H:%M:%S} %p [%c] : %m %n"
      << endl;
    log_conf << "log4cpp.category.main_cat=INFO" << endl;
    
    log_conf.close();  
    
    log4cpp::PropertyConfigurator::configure(property_file.c_str());
    log4cpp::Category &main_cat = log4cpp::Category::getInstance(main_cat_str.c_str());
    main_cat.info("This is some info");
    main_cat.debug("This debug message will fail to write");
    main_cat.alert("All hands abandon ship");

    main_cat.log(log4cpp::Priority::WARN, "This will be a logged warning");

    log4cpp::Category::shutdown();

    boost::filesystem::remove(property_file);   
    // now, open file and check results
    ifstream log_file(logging_file.c_str());
    string currentLine;
    getline(log_file, currentLine);
    boost::algorithm::trim(currentLine);
    BOOST_REQUIRE(boost::algorithm::ends_with(currentLine, "This is some info"));
    getline(log_file, currentLine);
    boost::algorithm::trim(currentLine);
    BOOST_REQUIRE(boost::algorithm::ends_with(currentLine, "All hands abandon ship"));
    getline(log_file, currentLine);
    boost::algorithm::trim(currentLine);
    BOOST_REQUIRE(boost::algorithm::ends_with(currentLine, "This will be a logged warning"));
    log_file.close();
    boost::filesystem::remove(logging_file);
}

BOOST_AUTO_TEST_CASE(createNewConfFile_substitutes_log_file_test) {
	// remove files left
	if (boost::filesystem::exists(property_file)) {
    boost::filesystem::remove(property_file);
  }

	if (boost::filesystem::exists(new_property_file)) {
    boost::filesystem::remove(new_property_file);
  }

  // create configuration file with a given log4cpp.appender.A1.fileName

  // write configuration file
  ofstream log_conf(property_file.c_str());

  log_conf << \
"# it accepts either starting with log4j or log4cpp" << endl <<
"# however, only log4cpp class names are allowed" << endl <<
"log4cpp.rootCategory=DEBUG, A1" << endl <<
"#log4cpp.appender.rootAppender=org.apache.log4cpp.ConsoleAppender" << endl <<
"#log4cpp.appender.rootAppender.layout=log4cpp.BasicLayout" << endl <<
"# Output configuration" << endl <<
"log4cpp.appender.A1=log4cpp.RollingFileAppender" << endl <<
"# file name must appear as property fileName and not File as it appears in" << endl <<
"# log4j" << endl <<
"# It is possible to put a full path name: /tmp/logging.log" << endl <<
"log4cpp.appender.A1.fileName=../test/scratch/global-test.log" << endl <<
"log4cpp.appender.A1.layout=log4cpp.PatternLayout" << endl <<
"# Print in following format:" << endl <<
"# Weekday Day Month HH:MM:SS priority [category] : message \n" << endl <<
"log4cpp.appender.A1.layout.ConversionPattern=%d{%a %d %b %H:%M:%S} %p [%c] : %m %n" << endl <<
"# individual log level for certain categories" << endl <<
"#log4cpp.category.chemmodel.ResidueAccessor=DEBUG" << endl <<
"log4cpp.category.main_cat=INFO" << endl <<
"log4cpp.category.alphasurfI_cat=DEBUG" << endl;
  log_conf.close();

  createNewConfFile(property_file,
      new_property_file,
      logging_file);

  // checking: new_property_file must be like property file, but logging_file is the
  // new value of log4cpp.appender.A1.fileName

  // first, check the file exists
  BOOST_CHECK(boost::filesystem::exists(new_property_file));

  // then, open and read property log4cpp.appender.A1.fileName, and chek the
  // value is logging_file
  azucar::utils::InputFile input(new_property_file);

  bool propertyFound = false;

  while(not azucar::utils::eof(input.stream)) {
    string line;
    getline(input.stream, line);

    boost::algorithm::trim(line);

    if (boost::algorithm::starts_with(line, "#")) {
    	continue;
    } else {
      // break line by = sign
      string::size_type equalPos = line.find('=');
      if (equalPos == string::npos) {
        continue;
      } else {
        string firstPart = line.substr(0, equalPos);
        string valuePart = line.substr(equalPos+1); // if no value, it will get
          // the empty string
        boost::algorithm::trim(firstPart);
        boost::algorithm::trim(valuePart);

        // Is it the property we are looking for?
        if (boost::algorithm::equals(firstPart, "log4cpp.appender.A1.fileName")) {
        	propertyFound = true;

        	BOOST_REQUIRE_EQUAL(logging_file, valuePart);

        	break;
        }
      }
    }
  }


  input.close();

  BOOST_CHECK(propertyFound);

  boost::filesystem::remove(property_file);
  boost::filesystem::remove(new_property_file);

}

BOOST_AUTO_TEST_SUITE_END()
