/*! \file test/azucar/utils/mail/test_mail.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <cassert>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>

#include <vmime/vmime.hpp>
#include <vmime/platforms/posix/posixHandler.hpp>

#include <loki/ScopeGuard.h>

#include <boost/utility.hpp>
#include <boost/date_time/gregorian/gregorian.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/xtime.hpp>

#include <boost/test/unit_test.hpp>

#include "azucar/utils/util.hpp"
#include "azucar/utils/mail/mailUtil.hpp"
#include "test/azucar/utils/mail/test_mail.hpp"
#include "test/azucar/utils/testCommon.hpp"
#include "azucar/utils/mail/mailCommon.hpp"

using boost::unit_test::test_suite;
using boost::unit_test::test_case;
using namespace std;


namespace {
  
  const string subject_string = "SMTP test";
  
  const string attachment_file = "scratch/test_attachment.txt";

  const std::vector<vmime::ref<const vmime::attachment> >::size_type
    expected_attachments = 1;
  
  string exchange_string;
  
MailSender mailSender;






void mailMessage(const string &recipient, const string &message,
      vmime::ref<vmime::fileAttachment> att = vmime::ref<vmime::fileAttachment>()) {
    vmime::ref<vmime::mailbox> destinationMailBox = 
      vmime::create<vmime::mailbox>(recipient);
    vmime::messageBuilder mb;
    
    mb.setSubject(vmime::text(subject_string));
    mb.setExpeditor(vmime::mailbox(MailConfiguration::mail_expeditor));
    mb.getRecipients().appendAddress
      (destinationMailBox);
    mb.getTextPart()->setCharset(vmime::charsets::UTF_8);
    mb.getTextPart()->setText(vmime::create<vmime::stringContentHandler>
      (message));

    if (att != vmime::ref<vmime::fileAttachment>()) {
      mb.appendAttachment(att);
    }

    vmime::ref<vmime::message> msg = mb.construct();
    
    mailSender.sendMail(msg);
}

//vmime::ref<vmime::net::message> retrieveMessage() {
void retrieveMessage() {
  azucar::utils::threadSleep(1);

    vmime::ref<vmime::net::session> sess = vmime::create<vmime::net::session>();
    vmime::utility::url url(test::settings.imap_url);

    vmime::ref<vmime::net::store> store
      = sess->getStore(url, vmime::create <interactiveAuthenticator>(test::settings.imap_user, test::settings.imap_key));
    store->setProperty("options.need-authentication", true);
    store->connect();
    // since LOKI_ON_BLOCK_EXIT will call the function before
    // tr is destroyed, we make sure we can do *tr.
    ON_BLOCK_EXIT(&vmime::net::store::disconnect, *store);
    vmime::ref<vmime::net::folder> inbox
      = store->getDefaultFolder();
    inbox->open(vmime::net::folder::MODE_READ_WRITE);
    std::vector<vmime::ref<vmime::net::message> > allMessages
      = inbox->getMessages();
    inbox->fetchMessages(allMessages,
      vmime::net::folder::FETCH_ENVELOPE);
    std::vector<vmime::ref<vmime::net::message> >::iterator msgIt;
    for(msgIt = allMessages.begin(); msgIt != allMessages.end();
        ++msgIt) {
      vmime::ref<vmime::net::message> msg = *msgIt;
      ostringstream subject;
      subject << msg->getHeader()->Subject()->getValue()->generate();
      
      if (subject.str() == subject_string) {
        vmime::ref<vmime::message> completeMsg
          = vmime::create<vmime::message>();
        string wholeMessage;
        vmime::utility::outputStreamStringAdapter os(wholeMessage);
        msg->extract(os);
        completeMsg->parse(wholeMessage);
        vmime::ref<vmime::body> body = completeMsg->getBody();
        string bodyText;
        vmime::utility::outputStreamStringAdapter os2(bodyText);
        body->getContents()->extract(os2);
        BOOST_REQUIRE(boost::algorithm::starts_with(bodyText, exchange_string));  
        // only gets here if BOOST_REQUIRE passes
        inbox->deleteMessage(msg->getNumber());
        inbox->expunge();
        return;
      }                
    }
    BOOST_FAIL("Could not find message");
}


void simple_test() {

  mailMessage(string(test::settings.imap_email),
    exchange_string);

  return;
}

void sendReceiveAttachment_test() {
  // write test file
  if (boost::filesystem::exists(attachment_file)) {
    boost::filesystem::remove(attachment_file);
  }
  
  ofstream attachmentFile(attachment_file.c_str());
  attachmentFile << exchange_string << endl;
  attachmentFile.close();
  
  boost::filesystem::path attachmentPath
    = boost::filesystem::system_complete(attachment_file);
  
  vmime::ref<vmime::fileAttachment> att
    = vmime::create<vmime::fileAttachment> (
      attachmentPath.string(),
      vmime::mediaType("text/plain"),
      vmime::text(attachment_file)
    );
  
  att->getFileInfo().setFilename(attachment_file);
  
  mailMessage(string(test::settings.imap_email),
    exchange_string, att);


  boost::filesystem::remove(attachment_file);
  

  // wait 2 seconds
  azucar::utils::threadSleep(2);
  
  // do reception
    vmime::ref<vmime::net::session> sess = vmime::create<vmime::net::session>();
    vmime::utility::url url(test::settings.imap_url);

    vmime::ref<vmime::net::store> store
      = sess->getStore(url, vmime::create <interactiveAuthenticator>(test::settings.imap_user, test::settings.imap_key));
    store->setProperty("options.need-authentication", true);
    store->connect();
    // since LOKI_ON_BLOCK_EXIT will call the function before
    // tr is destroyed, we make sure we can do *tr.
    ON_BLOCK_EXIT(&vmime::net::store::disconnect, *store);
    vmime::ref<vmime::net::folder> inbox
      = store->getDefaultFolder();
    inbox->open(vmime::net::folder::MODE_READ_WRITE);
    std::vector<vmime::ref<vmime::net::message> > allMessages
      = inbox->getMessages();
    inbox->fetchMessages(allMessages,
      vmime::net::folder::FETCH_ENVELOPE);
    std::vector<vmime::ref<vmime::net::message> >::iterator msgIt;
    for(msgIt = allMessages.begin(); msgIt != allMessages.end();
        ++msgIt) {
      vmime::ref<vmime::net::message> msg = *msgIt;
      ostringstream subject;
      subject << msg->getHeader()->Subject()->getValue()->generate();
      
      if (subject.str() == subject_string) {
        vmime::ref<vmime::message> completeMsg
          = vmime::create<vmime::message>();
        string wholeMessage;
        vmime::utility::outputStreamStringAdapter os(wholeMessage);
        msg->extract(os);
        completeMsg->parse(wholeMessage);
        
        const std::vector<vmime::ref<const vmime::attachment> > atts
          = vmime::attachmentHelper::findAttachmentsInMessage(completeMsg);
        
        BOOST_REQUIRE_EQUAL(atts.size(), expected_attachments);


        string description;
        vmime::utility::outputStreamStringAdapter descOs(description);
        atts[0]->getDescription().generate(descOs);        
        BOOST_CHECK_EQUAL(description, attachment_file);
        
        // check contents
        string attachmentData;
        vmime::utility::outputStreamStringAdapter dataOs(attachmentData);
        atts[0]->getData()->extract(dataOs);
        BOOST_CHECK(boost::algorithm::starts_with(attachmentData, exchange_string));        
        // check body contents
        vmime::ref<vmime::body> body = completeMsg->getBody()->getPartAt(0)->getBody();
        string bodyText;
        vmime::utility::outputStreamStringAdapter os2(bodyText);
        body->getContents()->extract(os2);
        BOOST_REQUIRE(boost::algorithm::starts_with(bodyText, exchange_string));  

        

        // find the attachment 


        // only gets here if BOOST_REQUIRE passes
        inbox->deleteMessage(msg->getNumber());
        inbox->expunge();
        return;
      }                
    }
    BOOST_FAIL("Could not find message");
  
  return;
}
  
} // end namespace

test_suite*
test_mail_init() {
  
    boost::posix_time::ptime status_time_ = boost::posix_time::second_clock::local_time();
    exchange_string = boost::gregorian::to_iso_extended_string(status_time_.date());
    exchange_string += boost::posix_time::to_simple_string(status_time_.time_of_day()); 

    test_suite* test= BOOST_TEST_SUITE(test_mail);

    // this example will pass cause we know ahead of time number of expected failures
    test_case *simpleTest = BOOST_TEST_CASE( &simple_test );
    test_case *retrieveTest = BOOST_TEST_CASE( &retrieveMessage);
    test_case *sendReceiveAttachmentTest = BOOST_TEST_CASE(&sendReceiveAttachment_test);
    retrieveTest->depends_on(simpleTest);
    sendReceiveAttachmentTest->depends_on(retrieveTest);
    test->add( simpleTest);
    test->add( retrieveTest, 0, 3);
    test->add(sendReceiveAttachmentTest, 0, 5);

    return test;
}
