/*! \file test/azucar/utils/testDOM.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 *
 * Code based on Apache source code (see below).
 */

/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Id$
 */

/*
 * This sample illustrates how you can create a DOM tree in memory.
 * It then prints the count of elements in the tree.
 */


// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------
#include <vector>
#include <functional>
#include <string>

#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>
#if defined(XERCES_NEW_IOSTREAMS)
#include <iostream>
#else
#include <iostream.h>
#endif
#include <xercesc/util/OutOfMemoryException.hpp>

#include <boost/foreach.hpp>
#include <boost/date_time/time_clock.hpp>
#include <boost/date_time/microsec_time_clock.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/shared_ptr.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/PropertyConfigurator.hh>
#include <log4cpp/NDC.hh>

#include <unistd.h>

#include "azucar/utils/init.hpp"
#include "azucar/utils/xmlInit.hpp"
#include "azucar/utils/xmlUtil.hpp"
#include "azucar/utils/exceptions.hpp"



XERCES_CPP_NAMESPACE_USE

using namespace std;
using azucar::utils::Initializer;


std::vector<Initializer *> initializers;

void globalInit(Initializer &initObj) {
  xmlInitializer.setConfigFile(initObj.getConfigFile());
  initializers.push_back(&xmlInitializer);
  
  for_each(initializers.begin(), initializers.end(),
    mem_fun(&Initializer::init));
}

void globalDestroy(Initializer &initObj) {
  for_each(initializers.rbegin(), initializers.rend(),
    mem_fun(&Initializer::destroy));
}

Initializer globalInitializer(&globalInit, &globalDestroy);


// ---------------------------------------------------------------------------
//  main
// ---------------------------------------------------------------------------

int main(int argC, char* /* argV[] */)
{

  typedef boost::date_time::microsec_clock<boost::posix_time::ptime> BoostClock;


  char programName[256];
  
  int programNameSize = readlink("/proc/self/exe", programName, 256);
  
  std::string programStrName(programName, programNameSize);
  
  cout << "Program name: '" << programStrName << "'" << endl;

  boost::shared_ptr<boost::posix_time::ptime> startTime;
  startTime.reset(new boost::posix_time::ptime(BoostClock::local_time()));
  

/*
  try {
//    string logFilename = vm[program_option_names[LOG_CONFIG_FILE]].as<string>();
    string logFilename = "log.conf";
    log4cpp::PropertyConfigurator::configure(logFilename);
  } catch (log4cpp::ConfigureFailure &ex) {
    // logging subsystem failing. Notify user of error. Give
    // some clue for debugging.
    cerr << "Logging subsystem failing." << endl;
    return 1;
  }
*/


//    globalInitializer.setConfigFile(configFilename);
    globalInitializer.init();



    // Watch for special case help request
    int errorCode = 0;

   {
       //  Nest entire test in an inner block.
       //  The tree we create below is the same that the XercesDOMParser would
       //  have created, except that no whitespace text nodes would be created.

       // <company>
       //     <product>Xerces-C</product>
       //     <category idea='great'>XML Parsing Tools</category>
       //     <developedBy>Apache Software Foundation</developedBy>
       // </company>

       DOMImplementation* impl =  DOMImplementationRegistry::getDOMImplementation(XMLSTR("LS"));

       if (impl != NULL)
       {
           try
           {
               DOMDocument* doc = impl->createDocument(
                           0,                    // root element namespace URI.
                           XMLSTR("provenance"),         // root element name
                           0);                   // document type object (DTD).

               DOMElement* rootElem = doc->getDocumentElement();

               DOMElement*  prodElem = doc->createElement(XMLSTR("applicationName"));
               rootElem->appendChild(prodElem);

               DOMText*    prodDataVal = doc->createTextNode(XMLSTR("myapplication"));
               prodElem->appendChild(prodDataVal);

               DOMElement*  catElem = doc->createElement(XMLSTR("execution"));
               rootElem->appendChild(catElem);

               catElem->setAttribute(XMLSTR("start"), XMLSTR("now"));

               DOMText*    catDataVal = doc->createTextNode(XMLSTR("duration"));
               catElem->appendChild(catDataVal);

               DOMElement*  devByElem = doc->createElement(XMLSTR("version"));
               rootElem->appendChild(devByElem);

               DOMText*    devByDataVal = doc->createTextNode(XMLSTR("1.0"));
               devByElem->appendChild(devByDataVal);

               //
               // Now count the number of elements in the above DOM tree.
               //

               unsigned int elementCount = doc->getElementsByTagName(XMLSTR("*"))->getLength();
               cout << "The tree just created contains: " << elementCount
                    << " elements." << XERCES_STD_QUALIFIER endl;


                DOMWriter *writer = ((DOMImplementationLS*)impl)->createDOMWriter();

                if (writer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true))
                 writer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true);

                
                XMLFormatTarget *writerTarget = new LocalFileFormatTarget("provenance.xml");
                
                
                writer->writeNode(writerTarget, *doc);


               doc->release();
           }
           catch (const OutOfMemoryException&)
           {
               cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
               throw new azucar::utils::BaseException("OutOfMemoryException");
           }
           catch (const DOMException& e)
           {
               cerr << "DOMException code is:  " << e.code << XERCES_STD_QUALIFIER endl;
               throw new azucar::utils::BaseException("DOMException");
           }
           catch (...)
           {
               cerr << "An error occurred creating the document" << XERCES_STD_QUALIFIER endl;
               throw new azucar::utils::BaseException("Error creating XML document");
           }
       }  // (inpl != NULL)
       else
       {
           cerr << "Requested implementation is not supported" << XERCES_STD_QUALIFIER endl;
           throw new azucar::utils::BaseException("XML implementation not supported");
       }
   }

  boost::posix_time::ptime endTime(BoostClock::local_time());

  boost::posix_time::time_period elapsed(*startTime, endTime);
  
  cout << "Start time: " << boost::posix_time::to_simple_string(*startTime) << endl;
  cout << "End time: " << boost::posix_time::to_simple_string(endTime) << endl;
  cout << "Elapsed: " << boost::posix_time::to_simple_string(elapsed.length()) << endl;


  globalInitializer.destroy();
   return errorCode;
}

