/*! \file test/azucar/utils/test_geometry.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <cmath>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>

#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>
#include <boost/shared_ptr.hpp>

#include "azucar/utils/geometry.hpp"

using boost::unit_test::test_suite;
using boost::unit_test::test_case;
using namespace std;

BOOST_AUTO_TEST_SUITE(geometry)

BOOST_AUTO_TEST_CASE(test_simple) {
    DoubleVector atomPoint(7.865,27.347,15.139);

    Box searchRange;
    searchRange.minPoint = atomPoint - DoubleVector(5.0, 5.0, 5.0);
    searchRange.maxPoint = atomPoint + DoubleVector(5.0, 5.0, 5.0);
    
    Box range2;
    range2.minPoint = atomPoint - DoubleVector(3.0, 3.0, 3.0);
    range2.maxPoint = atomPoint + DoubleVector(3.0, 3.0, 3.0);
    
    Box range3;
    range3.minPoint = atomPoint - DoubleVector(3.0, 3.0, 3.0);
    range3.maxPoint = atomPoint + DoubleVector(6.0, 6.0, 6.0);
    
    DoubleVector point2(7.865, 27.347, -200.0);
    
    BOOST_CHECK(contains(searchRange, atomPoint));
    BOOST_CHECK(contains(searchRange, searchRange));
    BOOST_CHECK(contains(searchRange, range2));
    BOOST_CHECK(not contains(searchRange, range3));
    BOOST_CHECK(intersects(searchRange, range3));
    BOOST_CHECK(intersects(searchRange, range2));
    BOOST_CHECK(not contains(range2, searchRange));
    BOOST_CHECK(contains(searchRange, atomPoint));
    BOOST_CHECK(not contains(searchRange, point2));
    
    Box range4;
    range4.minPoint = DoubleVector(-14.534,-2.089,10.714);
    range4.maxPoint = DoubleVector(5.166,37.075,49.822);
    
    BOOST_CHECK(not contains(searchRange, range4));
    BOOST_CHECK(intersects(searchRange, range4));
    BOOST_CHECK(not contains(searchRange, range4.minPoint));
    
    MPFBox mpf_searchRange;
    mpf_searchRange.minPoint.x = searchRange.minPoint.x;
    mpf_searchRange.minPoint.y = searchRange.minPoint.y;
    mpf_searchRange.minPoint.z = searchRange.minPoint.z;
    mpf_searchRange.maxPoint.x = searchRange.maxPoint.x;
    mpf_searchRange.maxPoint.y = searchRange.maxPoint.y;
    mpf_searchRange.maxPoint.z = searchRange.maxPoint.z;
    
    MPFBox mpf_range4;
    mpf_range4.minPoint.x = range4.minPoint.x;
    mpf_range4.minPoint.y = range4.minPoint.y;
    mpf_range4.minPoint.z = range4.minPoint.z;
    mpf_range4.maxPoint.x = range4.maxPoint.x;
    mpf_range4.maxPoint.y = range4.maxPoint.y;
    mpf_range4.maxPoint.z = range4.maxPoint.z;
    
    BOOST_CHECK(not contains(mpf_searchRange, mpf_range4));
    BOOST_CHECK(intersects(mpf_searchRange, mpf_range4));
    BOOST_CHECK(not contains(mpf_searchRange, mpf_range4.minPoint));
    
  }
  
BOOST_AUTO_TEST_CASE(test_position) {
//ATOM      3  C   LYS     2      14.156   0.065  32.948  1.00 31.90           C  
//ATOM     10  N   LYS     3      13.155   0.816  33.396  1.00 28.01           N  
//ATOM     11  CA  LYS     3      11.796   0.537  32.974  1.00 19.35           C  
//ATOM     14  CB  LYS     3      11.071  -0.337  34.002  1.00 19.06           C 

    DoubleVector natom;
    natom.x = 13.155;
    natom.y = 0.816;
    natom.z = 33.396;
    
    DoubleVector caatom;
    caatom.x = 11.796;
    caatom.y = 0.537;
    caatom.z = 32.974;
    
    DoubleVector catom;
    catom.x = 14.156;
    catom.y = 0.065;
    catom.z = 32.948;
    
    DoubleVector originalcbatom;
    originalcbatom.x = 11.071;
    originalcbatom.y = -0.337;
    originalcbatom.z = 34.002;
    
    double angle = getAngle(natom - caatom, originalcbatom - caatom);
    double distance = modulusVector(originalcbatom - caatom);
    double torsion = torsionAngle(catom, natom, caatom, originalcbatom);
    DoubleVector newcbatom = setPoint(catom, natom, caatom, distance, angle, torsion);
    
    BOOST_CHECK_CLOSE(newcbatom.x, originalcbatom.x, 0.0001);
    BOOST_CHECK_CLOSE(newcbatom.y, originalcbatom.y, 0.0001);
    BOOST_CHECK_CLOSE(newcbatom.z, originalcbatom.z, 0.0001);
    
  }
  
BOOST_AUTO_TEST_CASE(test_rmsd) {
    int n = 5;
    
    gsl_matrix *x = gsl_matrix_alloc(3, n);
    gsl_matrix *y = gsl_matrix_alloc(3, n);
    
    const gsl_rng_type * T;
    gsl_rng * rng;
   
    gsl_rng_env_setup();
   
    T = gsl_rng_default;
    rng = gsl_rng_alloc (T);
   
    for(int i=0; i< 3; ++i) {
      for(int j=0; j < n; ++j) {
        double u = gsl_rng_uniform(rng);
        double v = gsl_rng_uniform(rng);
        gsl_matrix_set(x, i, j, u);
        gsl_matrix_set(y, i, j, v);
      }
    }
   
    gsl_rng_free (rng);
    
    gsl_matrix *transform = gsl_matrix_alloc(4, 4);
    gsl_vector *quaternion = gsl_vector_alloc(4);
    
    double rmsd = doRMSD(x, y, transform, quaternion);
    
    gsl_matrix *originalX = gsl_matrix_alloc(4, n);
    // set last row to 1
    gsl_matrix_view subX = gsl_matrix_submatrix(originalX, 0, 0, 3, n);
    gsl_matrix_memcpy(&subX.matrix, x);
    gsl_vector_view lastRow = gsl_matrix_row(originalX, 3);
    gsl_vector_set_all(&lastRow.vector, 1.0);
      
    // new coordinates are transform * originalX
    
    gsl_matrix *finalX = gsl_matrix_alloc(4, n);
    gsl_matrix_set_zero(finalX);
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, transform, originalX, 0.0, finalX);
    gsl_matrix_view coordinates = gsl_matrix_submatrix(finalX, 0, 0, 3, n);
    

    BOOST_CHECK_CLOSE(RMSD(&coordinates.matrix, y), rmsd, 0.000001);
    
    gsl_matrix_free(finalX);
    gsl_matrix_free(originalX);
    gsl_vector_free(quaternion);
    gsl_matrix_free(transform);
    gsl_matrix_free(x);
    gsl_matrix_free(y);
   
    return;
    
  }

// end geometry
BOOST_AUTO_TEST_SUITE_END()

    
