/*! \file test/azucar/utils/test_kdtree.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <cassert>
#include <fstream>
#include <algorithm>
#include <string>

#include <boost/test/unit_test.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>


#include "azucar/utils/kdtree.hpp"
#include "azucar/utils/util.hpp"
#include "azucar/utils/geometry.hpp"

using boost::unit_test::test_suite;
using boost::unit_test::test_case;
using namespace std;

BOOST_AUTO_TEST_SUITE(kdtree)

class Object3D {
public:
	DoubleVector coordinates;

	Object3D() : coordinates() {};

	Object3D(const Object3D& other) : coordinates(other.coordinates) {
	};

	Object3D& operator=(const Object3D& other) {
		this->coordinates = other.coordinates;

		return *this;
	};

	DoubleVector getCoordinates() { return this->coordinates; };
};

  const std::string input_file = "data/test/common/input.csv";
BOOST_AUTO_TEST_CASE(test_simple) {
    
    boost::shared_ptr<Object3D> atom1(new Object3D());
    boost::shared_ptr<Object3D> atom2(new Object3D());
    boost::shared_ptr<Object3D> atom3(new Object3D());
    
    atom1->coordinates = DoubleVector(1.2, 3.5, 3.4);
    atom2->coordinates = DoubleVector(5.4, 2.9, 12.1);
    atom3->coordinates = DoubleVector(-3.5, 0.0, 12.1);
    
    KDTreeNode<Object3D>::ObjContainer sortX, sortY, sortZ;


    sortX.push_back(atom3);
    sortX.push_back(atom1);
    sortX.push_back(atom2);
    
    sortY.push_back(atom3);
    sortY.push_back(atom2);
    sortY.push_back(atom1);
    
    sortZ.push_back(atom1);
    sortZ.push_back(atom3);
    sortZ.push_back(atom2);
    
    Box searchRange;
    searchRange.minPoint = DoubleVector(0.0, 0.0, 0.0);
    searchRange.maxPoint = DoubleVector(5.0, 5.0, 5.0);
    
    KDTreeNode<Object3D>::ObjContainer result;
    
    
    KDTreeNode<Object3D>::NodePtr tree
      = KDTreeNode<Object3D>::buildTree(sortX, sortY, sortZ);

    tree->search(searchRange, result);

    BOOST_REQUIRE_EQUAL(result.size(), 1U);
    BOOST_CHECK_EQUAL(result[0].get(), atom1.get());
    
  }
  
  bool lessThanDimAtom(int dimension,
      const boost::shared_ptr<Object3D> v, const boost::shared_ptr<Object3D> w) {
    return lessThanDim(dimension,
      v->getCoordinates(), w->getCoordinates());
  }

  BOOST_AUTO_TEST_CASE(test_sphere_range_search) {

    KDTreeNode<Object3D>::ObjContainer sortX, sortY, sortZ;

    azucar::utils::InputFile inputF(input_file);

    std::string line;
    while(not azucar::utils::eof(inputF.stream)) {
    	std::getline(inputF.stream, line);
    	std::vector<std::string> coord;
    	coord = boost::algorithm::split(coord,
    			line, boost::algorithm::is_any_of(","));
    	boost::shared_ptr<Object3D> atom(new Object3D());
    	atom->coordinates = DoubleVector(azucar::utils::trimConvert<double>(coord[0]),
    			azucar::utils::trimConvert<double>(coord[1]),
    			azucar::utils::trimConvert<double>(coord[2]));

    	sortX.push_back(atom);
    	sortY.push_back(atom);
    	sortZ.push_back(atom);
    }

    // sort vectors
    sort(sortX.begin(), sortX.end(), boost::bind(lessThanDimAtom, 1, _1, _2));
    sort(sortY.begin(), sortY.end(), boost::bind(lessThanDimAtom, 2, _1, _2));
    sort(sortZ.begin(), sortZ.end(), boost::bind(lessThanDimAtom, 3, _1, _2));

    // print sorted data

    DoubleVector atomPoint(7.865,27.347,15.139);

    Box searchRange;
    searchRange.minPoint = atomPoint - DoubleVector(5.0, 5.0, 5.0);
    searchRange.maxPoint = atomPoint + DoubleVector(5.0, 5.0, 5.0);

    KDTreeNode<Object3D>::ObjContainer result;


    KDTreeNode<Object3D>::NodePtr tree
      = KDTreeNode<Object3D>::buildTree(sortX, sortY, sortZ);

    BOOST_REQUIRE(checkTree<Object3D>(tree));

    tree->search(searchRange, result);


    // get now only those in the 5.0 A range (25 if square distance)
    int totalAtoms = 0;
    KDTreeNode<Object3D>::ObjContainer::iterator resultIt;
    for(resultIt = result.begin(); resultIt != result.end(); ++resultIt) {
      if (squareDistance(atomPoint, (*resultIt)->getCoordinates()) <= 25) {
        totalAtoms++;
      }
    }

    BOOST_CHECK_EQUAL(totalAtoms, 8);


  }

// end kdtree
BOOST_AUTO_TEST_SUITE_END()
