/*! \file test/azucar/utils/test_network.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <cassert>
#include <string>

#include <boost/test/unit_test.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>
#include <boost/ref.hpp>

#include <boost/asio.hpp>

#include "azucar/utils/network.hpp"

#include "azucar/utils/util.hpp"
#include "test/azucar/utils/test_network.hpp"
//#include "test/test_util.hpp"

namespace {

using boost::unit_test::test_suite;
using boost::unit_test::test_case;
using namespace std;
using boost::asio::ip::tcp;

const string test_string1("Test string.");
const string test_string2("Done.");

const string test_host("localhost");
const int test_port = 5041; 

class Transmisor {
private:
  tcp::socket socket_;
public:
  Transmisor(boost::asio::io_service &io_service) : socket_(io_service) {
    tcp::resolver resolver(io_service);
    tcp::resolver::query query(test_host,
      boost::lexical_cast<string>(test_port),
      tcp::resolver::query::numeric_service);
    tcp::resolver::iterator endpoint_iterator 
      = resolver.resolve(query);
    tcp::resolver::iterator end;
    
    assert(endpoint_iterator != end);

    boost::system::error_code error = boost::asio::error::host_not_found;
    while(error && endpoint_iterator != end) {
    	socket_.close();
    	socket_.connect(*endpoint_iterator++, error);
    }
    if (error) {
    	throw boost::system::system_error(error);
    }
  }
  
  template<class T> void writeObject(const T& t) {
    network::write(socket_, t);
  }

  template<class T> const T readObject() {
    return network::read<T>(socket_);
  }
  
  void shutdownSendHalf() {
    // shutdown connection so that server knows we are finished
    socket_.shutdown(boost::asio::ip::tcp::socket::shutdown_send,
       boost::system::throws);
  }
  
  ~Transmisor() {
    socket_.close();
  }
};

class tcp_server
{
private:
  tcp::socket socket_;
  tcp::acceptor acceptor_;
public:
  tcp_server(boost::asio::io_service& io_service)
    : socket_(io_service),
      acceptor_(io_service, tcp::endpoint(tcp::v4(), 
                                    test_port))
  {
    acceptor_.accept(socket_);
  }

  template<class T> const T readObject() {
    return network::read<T>(socket_);
  }

  template<class T> void writeObject(const T& t) {
    network::write(socket_, t);
  }
  
  ~tcp_server() {
      acceptor_.close();
  }
};

string serverResult;
string clientResult;

void doServer(boost::asio::io_service &io_service) {
  tcp_server server(io_service);
  serverResult = server.readObject<string>();
  server.writeObject<string>(test_string2);
}

void doClient(boost::asio::io_service &io_service, network::NetAsyncService& asyncService) {
  Transmisor client(io_service);
  client.writeObject<string>(test_string1);
  client.shutdownSendHalf();

  clientResult = client.readObject<string>();
  
  asyncService.asyncTerminate();
}

void simple_test() {
  // create io_service
  boost::asio::io_service io_service;
  
  network::NetAsyncService asyncService(io_service);
  
  
  // create server
  boost::thread server(boost::bind(doServer, boost::ref(io_service)));

  azucar::utils::threadSleep(1);

  boost::thread client(boost::bind(doClient, boost::ref(io_service),
                                    boost::ref(asyncService)));


  asyncService.blockingRun();
  
  server.join();
  client.join();
  
  BOOST_CHECK_EQUAL(serverResult, string(test_string1));
  BOOST_CHECK_EQUAL(clientResult, string(test_string2));
  
}

void netAsyncService_test() {
  boost::asio::io_service io_service;
  network::NetAsyncService asyncService(io_service);
  boost::thread server(boost::bind(&network::NetAsyncService::blockingRun,
                                  &asyncService));

  azucar::utils::threadSleep(1);

  asyncService.asyncTerminate();
  server.join();                                  
} 

}

test_suite*
test_network_init() {
  
    test_suite* test= BOOST_TEST_SUITE(test_network);

    // this example will pass cause we know ahead of time number of expected failures
    test_case *simpleTest = BOOST_TEST_CASE( &simple_test );
    test_case *netAsyncServiceTest = BOOST_TEST_CASE( &netAsyncService_test);  
    simpleTest->depends_on(netAsyncServiceTest);
    test->add( netAsyncServiceTest, 0, 2);
    test->add( simpleTest, 0, 3);

    return test;
}
