/*! \file test/azucar/utils/test_process.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <cassert>
#include <string>
#include <vector>
#include <map>
#include <fstream>

#include <boost/test/unit_test.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include "azucar/utils/process.hpp"
#include "azucar/utils/util.hpp"
#include "test/azucar/utils/testCommon.hpp"

using boost::unit_test::test_suite;
using boost::unit_test::test_case;
using namespace std;

extern char **environ;

BOOST_AUTO_TEST_SUITE(test_process)

  const string temp_output = "scratch/tempProcess.txt";
  const string temp_output2 = "scratch/tempProcess2.txt";
  const string temp_input = "scratch/tempProcessInput.txt";
  
  BOOST_AUTO_TEST_CASE(test_processExecute_synchronous) {
    std::vector<std::string> args;
    args.push_back(string("/bin/date"));
    int resultCode = processExecute("/bin/date",
       args,
       convertEnvironToMap(environ),
       "/dev/null",
       "/dev/null");
    BOOST_CHECK_EQUAL(resultCode, 0);

    return;
  }
  
  BOOST_AUTO_TEST_CASE(test_processExecute_asynchronous) {
    std::vector<std::string> args;
    args.push_back(string("/bin/pwd"));
    int resultCode = processExecuteAsynchronous("/bin/pwd",
      args,
      convertEnvironToMap(environ),
      "/dev/null",
      temp_output);

    // wait for execution
    azucar::utils::threadSleep(1);
      
    BOOST_REQUIRE(resultCode != -1);
    ifstream inputFile(temp_output.c_str());
    BOOST_REQUIRE(not inputFile.fail());
    string directoryName;
    getline(inputFile, directoryName);
    BOOST_REQUIRE_EQUAL(boost::filesystem::path(directoryName).string(), 
      boost::filesystem::current_path().string());
  }
  
  BOOST_AUTO_TEST_CASE(test_processExecute_IPC) {
    boost::shared_array<double> ipcArgs(new double[10000]);
    boost::shared_array<double> results(new double[5000]);
    boost::shared_array<double> expectedResults(new double[5000]);

    boost::filesystem::path clientCompletePath
      = boost::filesystem::system_complete(test::settings.testIPCProcessClient_path);

    std::vector<std::string> args;
    args.push_back(clientCompletePath.string());

    for(int i = 0; i < 10000; ++i) {
      ipcArgs.get()[i] = i;
    }
    for(int i = 0; i < 5000; ++i) {
      results.get()[i] = 0;
      expectedResults.get()[i] = i*2;
    }




    processExecuteIPC(clientCompletePath.string(),
        reinterpret_cast<char *>(ipcArgs.get()),
        sizeof(double) * 10000,
        reinterpret_cast<char *>(results.get()),
        sizeof(double) * 5000, 5,
        args, convertEnvironToMap(environ),
        "/dev/null", "/dev/null");

    BOOST_CHECK_EQUAL_COLLECTIONS(results.get(), results.get()+5000,
      expectedResults.get(), expectedResults.get()+5000);
  }


  BOOST_AUTO_TEST_CASE(test_timedProcessExecute) {
    boost::shared_array<double> operandOne(new double[5000]);
    boost::shared_array<double> operandTwo(new double[5000]);
    boost::shared_array<double> results(new double[5000]);
    boost::shared_array<double> expectedResults(new double[5000]);

    boost::filesystem::path clientCompletePath
      = boost::filesystem::system_complete(test::settings.testIPCProcessClient_path);

    std::vector<std::string> args;
    args.push_back(clientCompletePath.string());
    args.push_back(temp_input);
    args.push_back(temp_output);
    args.push_back(boost::lexical_cast<string>(5000));

    for(int i = 0; i < 5000; ++i) {
      operandOne.get()[i] = i;
      operandTwo.get()[i] = i+1;
    }
    for(int i = 0; i < 5000; ++i) {
      results.get()[i] = 0;
      expectedResults.get()[i] = i*2+1;
    }



    boost::filesystem::remove(temp_input);
    boost::filesystem::remove(temp_output);

    // write input
    azucar::utils::OutputFile inputData(temp_input);

    for(int i = 0; i < 5000; ++i) {
      inputData.stream << operandOne.get()[i] << ","
        << operandTwo.get()[i] << "\n";
    }

    inputData.close();

    timedProcessExecute(clientCompletePath.string(),
        5, 1000000L,
        args, convertEnvironToMap(environ),
        "/dev/null", "/dev/null");

    // read output

    BOOST_REQUIRE(boost::filesystem::exists(temp_output));

    azucar::utils::InputFile outputData(temp_output);
    for(int i=0; i < 5000; ++i) {
      if (azucar::utils::eof(outputData.stream)) {
        BOOST_FAIL("Wrong output data format.");
      }
      string line;
      getline(outputData.stream, line);
      boost::algorithm::trim(line);
      results.get()[i] = boost::lexical_cast<double>(line);
    }

    if (not azucar::utils::eof(outputData.stream)) {
      BOOST_FAIL("Wrong output data format (too much data).");
    }

    outputData.close();

    BOOST_CHECK_EQUAL_COLLECTIONS(results.get(), results.get()+5000,
      expectedResults.get(), expectedResults.get()+5000);
  }

  void timedMultiProcessExecuteCore(bool runWithDelay) {
    boost::shared_array<double> operandOne(new double[5000]);
    boost::shared_array<double> operandTwo(new double[5000]);
    boost::shared_array<double> results(new double[5000]);
    boost::shared_array<double> expectedResults(new double[5000]);

    boost::filesystem::path clientCompletePath
      = boost::filesystem::system_complete(test::settings.testIPCProcessClient_path);

    std::vector<std::string> args;
    args.push_back(clientCompletePath.string());
    args.push_back(temp_input);
    args.push_back(temp_output);
    args.push_back(boost::lexical_cast<string>(5000));

    for(int i = 0; i < 5000; ++i) {
      operandOne.get()[i] = i;
      operandTwo.get()[i] = i+1;
    }
    for(int i = 0; i < 5000; ++i) {
      results.get()[i] = 0;
      expectedResults.get()[i] = i*2+1;
    }



    boost::filesystem::remove(temp_input);
    boost::filesystem::remove(temp_output);
    boost::filesystem::remove(temp_output2);

    // write input
    azucar::utils::OutputFile inputData(temp_input);

    for(int i = 0; i < 5000; ++i) {
      inputData.stream << operandOne.get()[i] << ","
        << operandTwo.get()[i] << "\n";
    }

    inputData.close();

    ProcessRequest request1(clientCompletePath.string(),
      args, convertEnvironToMap(environ), "/dev/null", "/dev/null", "");

    args.clear();
    args.push_back("/bin/date");
    ProcessRequest request2("/bin/date",
      args, convertEnvironToMap(environ), "/dev/null", temp_output2, "");

    std::vector<ProcessRequest> requests;
    requests.push_back(request1);
    requests.push_back(request2);

    bool ok = false;

    if (runWithDelay) {
    	ok = timedMultiProcessExecuteWithDelay(requests,
        5, 1000000L, 100000L);
    } else {
    	ok = timedMultiProcessExecute(requests,
        5, 1000000L);
    }

    BOOST_REQUIRE(ok);

    // read output

    BOOST_REQUIRE(boost::filesystem::exists(temp_output));
    BOOST_REQUIRE(boost::filesystem::exists(temp_output2));

    azucar::utils::InputFile outputData(temp_output);
    for(int i=0; i < 5000; ++i) {
      if (azucar::utils::eof(outputData.stream)) {
        BOOST_FAIL("Wrong output data format.");
      }
      string line;
      getline(outputData.stream, line);
      boost::algorithm::trim(line);
      results.get()[i] = boost::lexical_cast<double>(line);
    }

    if (not azucar::utils::eof(outputData.stream)) {
      BOOST_FAIL("Wrong output data format (too much data).");
    }

    outputData.close();

    BOOST_CHECK_EQUAL_COLLECTIONS(results.get(), results.get()+5000,
      expectedResults.get(), expectedResults.get()+5000);
  }
  
  BOOST_AUTO_TEST_CASE(test_timedMultiProcessExecute) {
  	timedMultiProcessExecuteCore(false);
  }

  BOOST_AUTO_TEST_CASE(test_timedMultiProcessExecuteWithDelay) {
  	timedMultiProcessExecuteCore(true);
  }

BOOST_AUTO_TEST_SUITE_END()

