/*! \file test/azucar/utils/test_propertyDBXMLReader.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <iostream>
#include <string>
#include <sstream>

#include <boost/test/unit_test.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/python.hpp>
#include <boost/format.hpp>

#include "azucar/utils/propertyDBXMLReader.hpp"
#include "azucar/utils/exceptions.hpp"
#include "azucar/utils/util.hpp"

using namespace std;

BOOST_AUTO_TEST_SUITE(global)

BOOST_AUTO_TEST_SUITE(test_azucar_utilities_propertyDBXMLReader)

BOOST_AUTO_TEST_CASE(testCreation) {
	// setup
	// create an empty string io
	std::istringstream input("<azucar_utils_propertyDB></azucar_utils_propertyDB>");

	// exercise
	boost::shared_ptr<azucar::utils::PropertyDB> propdb
		= azucar::utils::readPropertyDBXML(input);

	// check results
	BOOST_CHECK(propdb); // should not be null (though it will contain no properties).
}

BOOST_AUTO_TEST_CASE(testCreationFromFile) {
  // open temporary file
	// write text there
	// close temporary file

	// reopen temporary file for reading
	// create propertyDB


	const std::string propFileText("<azucar_utils_propertyDB></azucar_utils_propertyDB>");

	// temporary file is destroyed at the end
	azucar::utils::TempFile temporaryFile("/tmp/testXXXXXX");

	azucar::utils::OutputFile outFile(temporaryFile.getPath());

	outFile.stream << propFileText;

	outFile.close();

	// exercise
	boost::shared_ptr<azucar::utils::PropertyDB> propdb
		= azucar::utils::readPropertyDBXML(temporaryFile.getPath());

	BOOST_CHECK(propdb);


}

BOOST_AUTO_TEST_CASE(testReadProperty) {
	std::istringstream input("<azucar_utils_propertyDB><prop1>value1</prop1></azucar_utils_propertyDB>");

	// exercise
	boost::shared_ptr<azucar::utils::PropertyDB> propdb
		= azucar::utils::readPropertyDBXML(input);

	BOOST_CHECK_EQUAL(propdb->get("prop1"), "value1");
}

BOOST_AUTO_TEST_CASE(testReadPropertyTrimsSpaces) {
	std::istringstream input("<azucar_utils_propertyDB><prop1> value1 </prop1></azucar_utils_propertyDB>");

	// exercise
	boost::shared_ptr<azucar::utils::PropertyDB> propdb
		= azucar::utils::readPropertyDBXML(input);

	BOOST_CHECK_EQUAL(propdb->get("prop1"), "value1");
}

BOOST_AUTO_TEST_CASE(testReadProperties) {
	std::istringstream input("<azucar_utils_propertyDB>\n"
			"<prop1>value1</prop1>\n"
			"<prop2>value2</prop2>\n"
			"</azucar_utils_propertyDB>");

	// exercise
	boost::shared_ptr<azucar::utils::PropertyDB> propdb
		= azucar::utils::readPropertyDBXML(input);

	BOOST_CHECK_EQUAL(propdb->get("prop1"), "value1");
	BOOST_CHECK_EQUAL(propdb->get("prop2"), "value2");
}

BOOST_AUTO_TEST_CASE(testWrongRepeatedProperty) {
	std::istringstream input("<azucar_utils_propertyDB>\n"
			"<prop1>value1</prop1>\n"
			"<prop1>value2</prop1>\n"
			"</azucar_utils_propertyDB>");

	// exercise and check
	BOOST_CHECK_THROW(azucar::utils::readPropertyDBXML(input),
			azucar::utils::BaseException);

}

BOOST_AUTO_TEST_CASE(testWrongTextInBetweenProperties) {
	std::istringstream input("<azucar_utils_propertyDB>\n"
			"<prop1>value1</prop1>\n"
			"Wrong text\n"
			"<prop2>value2</prop2>\n"
			"</azucar_utils_propertyDB>");

	// exercise and check
	BOOST_CHECK_THROW(azucar::utils::readPropertyDBXML(input),
			azucar::utils::BaseException);

}

BOOST_AUTO_TEST_CASE(testWrongNestedProperty) {
	std::istringstream input("<azucar_utils_propertyDB>\n"
			"<prop1><prop2>value2</prop2></prop1>\n"
			"</azucar_utils_propertyDB>");

	// exercise and check
	BOOST_CHECK_THROW(azucar::utils::readPropertyDBXML(input),
			azucar::utils::BaseException);
}


// end test_azucar_utilities_propertyDBXMLReader
BOOST_AUTO_TEST_SUITE_END()

// end global
BOOST_AUTO_TEST_SUITE_END()
