/*! \file test/azucar/utils/test_util.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <cassert>
#include <string>

#include <boost/test/unit_test.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem.hpp>

#include "azucar/utils/util.hpp"
#include "azucar/utils/version.hpp"

using boost::unit_test::test_suite;
using boost::unit_test::test_case;
using namespace std;

using namespace azucar::utils;



BOOST_AUTO_TEST_SUITE(test_util)

  const string email1 = "abc@example.com";
  const string email2 = "Abc@example.com";
  const string email3 = "aBC@example.com";
  const string email4 = "abc.123@example.com";
  const string email5 = "1234567890@example.com";
  const string email6 = "_______@example.com";
  const string email7 = "abc+mailbox/department=shipping@example.com";
  const string email8 = "!#$%&'*+-/=?^_`.{|}~@example.com";
  const string fail1 = "Abc.example.com";
  const string fail2 = "Abc.@example.com";
  const string fail3 = "Abc..123@example.com";
  const string fail4 = "A@b@c@example.com";
  const string fail5 = "()[]\\;:,<>@example.com";
  

BOOST_AUTO_TEST_CASE(versionsStartAtZero_test) {
  // minimum allowable version is 0.0.0. This tests the
	// main logic in azucar::utils::Version class works.
	BOOST_CHECK(azucar::utils::Version::IsAtLeast(0,0,0));
}



BOOST_AUTO_TEST_CASE(simple_test) {
    BOOST_CHECK(hasEmailAddressFormat(email1));
    BOOST_CHECK(hasEmailAddressFormat(email2));
    BOOST_CHECK(hasEmailAddressFormat(email3));
    BOOST_CHECK(hasEmailAddressFormat(email4));
    BOOST_CHECK(hasEmailAddressFormat(email5));
    BOOST_CHECK(hasEmailAddressFormat(email6));
    BOOST_CHECK(hasEmailAddressFormat(email7));
    BOOST_CHECK(hasEmailAddressFormat(email8));
    BOOST_CHECK(not hasEmailAddressFormat(fail1));
    BOOST_CHECK(not hasEmailAddressFormat(fail2));
    BOOST_CHECK(not hasEmailAddressFormat(fail3));
    BOOST_CHECK(not hasEmailAddressFormat(fail4));
    BOOST_CHECK(not hasEmailAddressFormat(fail5));
}

BOOST_AUTO_TEST_CASE(tempFile_test) {
  boost::filesystem::path testFilePath;
  
  {
    TempFile testFile("/tmp/testXXXXXX");
    
    testFilePath = testFile.getPath();
  
    BOOST_CHECK(boost::filesystem::exists(testFilePath));
  }
  
  BOOST_CHECK(not boost::filesystem::exists(testFilePath));
}

BOOST_AUTO_TEST_CASE(normalizePath_test) {
	BOOST_CHECK_EQUAL(normalizePath("/some/path/./to/nowhere/.."),
			boost::filesystem::path("/some/path/to"));
	BOOST_CHECK_EQUAL(normalizePath("./some/path/./to/nowhere/.."),
			boost::filesystem::path("some/path/to"));
	BOOST_WARN_THROW(normalizePath("/.."), azucar::utils::BaseException); // should say it is an invalid path
	BOOST_WARN_EQUAL(normalizePath("C:\\"),
			boost::filesystem::path("C:/")); // should work in Windows
	BOOST_CHECK_EQUAL(normalizePath("/"),
			boost::filesystem::path("/"));
}

BOOST_AUTO_TEST_CASE(relativePath_test) {
	BOOST_CHECK_EQUAL(relativePath("/path/to/nowhere/with/relative/name.txt", "/path/to/nowhere"),
			boost::filesystem::path("with/relative/name.txt"));
	BOOST_CHECK_EQUAL(relativePath("path/common/different/path", "path/common/another/path"),
			boost::filesystem::path("../../different/path"));
	BOOST_CHECK_EQUAL(relativePath("..", "."),
			boost::filesystem::path(".."));
}

BOOST_AUTO_TEST_SUITE_END()
