/*! \file test/tester.cpp
 *
 * \author Jorge Estrada
 *
 * Copyright 2013-2014 Jorge Estrada.
 *
 * This file is part of the azucar.cppUtils project.
 * It is subject to the license terms in the
 * LICENSE file found in the top-level directory of this distribution.
 * No part of the azucar.cppUtils project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms contained
 * in the LICENSE file.
 */

#include <iostream>
#include <string.h>
#include <vector>
#include <string>

#include <boost/test/unit_test.hpp>
#include <boost/python.hpp>
#include <boost/filesystem.hpp>

#include "azucar/utils/exceptions.hpp"
#include "azucar/utils/xerces.hpp"
#include "azucar/utils/init.hpp"
#include "azucar/utils/util.hpp"
#include "azucar/utils/processInit.hpp"
#include "azucar/utils/xmlInit.hpp"
#include "azucar/utils/zip/zipInit.hpp"
#include "azucar/utils/mail/mailInit.hpp"
#include "azucar/utils/baseXMLException.hpp"

#include "test/azucar/utils/testCommon.hpp"
#include "test/azucar/utils/test_monitoredQueue.hpp"
#include "test/azucar/utils/test_network.hpp"
#include "test/azucar/utils/mail/test_mail.hpp"

using namespace std;

using azucar::utils::Initializer;

std::vector<Initializer *> initializers;

void globalInit(Initializer &initObj) {
  xmlInitializer.setConfigFile(initObj.getConfigFile());
  initializers.push_back(&xmlInitializer);
  zipInitializer.setConfigFile(initObj.getConfigFile());
  initializers.push_back(&zipInitializer);
  processInitializer.setConfigFile(initObj.getConfigFile());
  initializers.push_back(&processInitializer);
  mailInitializer.setConfigFile(initObj.getConfigFile());
  initializers.push_back(&mailInitializer);

  for_each(initializers.begin(), initializers.end(),
    mem_fun(&Initializer::init));
}

void globalDestroy(Initializer &initObj) {
  for_each(initializers.rbegin(), initializers.rend(),
    mem_fun(&Initializer::destroy));
}

Initializer globalInitializer(&globalInit, &globalDestroy);

ON_BLOCK_EXIT(&Initializer::destroy, globalInitializer);

namespace test {
	Settings settings = test::Settings();
} // end namespace

bool
init_unit_test() {
  // Set Master Test Suite name
  boost::unit_test::framework::master_test_suite().p_name.value = "tester";

  try {
  	const std::string &test_config_file = boost::unit_test::framework::master_test_suite().argv[1];
		test::settings.load(test_config_file);

		globalInitializer.setConfigFile(test_config_file);
		globalInitializer.init();
  } catch (azucar::utils::BaseException& ex) {
  	std::cerr << ex.getMessage() << std::endl;
    return false; // return error
  }

  boost::unit_test::framework::master_test_suite().add(test_monitoredQueue_init());
  boost::unit_test::framework::master_test_suite().add(test_network_init());
  boost::unit_test::framework::master_test_suite().add(test_mail_init());


  return true;
}

int
main( int argc, char* argv[] )
{
	azucar::utils::XercesLibrary xerces;
  try {
    xerces.init();
  } catch (const exception& toCatch) {
    // Do your failure processing here
    return 1;
  }

  try {
    if (not boost::filesystem::exists("scratch")) {
      throw azucar::utils::BaseException("scratch directory is missing, and it is required for the tests.");
    }

    if (not boost::filesystem::is_directory("scratch")) {
      throw azucar::utils::BaseException("scratch directory is missing, and it is required for the tests.");
    }
  } catch (azucar::utils::BaseException& ex) {
  	std::cerr << ex.getMessage() << std::endl;
    return 1; // return error
  }

	return boost::unit_test::unit_test_main( &init_unit_test, argc, argv );
}


